/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Class for action-embedded queries (AQs) in Aorta.
 */
public class AQ {
	
	/** The Aorta query processor that this AQ belongs to. */
	public Aorta aortaqp;
	
	/** SQL text of this AQ. */
	public String AQString;  
	
	/** Type of this AQ. */
	public byte type;
	
	/** ID of this AQ. */
	public byte id;
		      
	/** Name of this AQ. */
	public String name;
	
	/** First-level action embedded in this AQ. */
	public Action  action;
	/** Second-level action embedded in this AQ; maybe <code>null</code>. */  
	public Action action2;   
		
	/** Id of the triggering device. */
	String sid; 
	/** ID of the triggered device. */      
	String did;       
		
	/** TinyDB sensor query for accessing sensor tuples from a real mote network. */
	public AortaSensorQuery sq;	           
	/** Simulated sensor query for accessing sensor tuples a simulated sensor network. */
	public AortaSimulatedSensorQuery simsq;
	/** SQL text of the sensor query. */
	public String sensorQueryString;          
	
	/** Specification of the input parameters of the first-level action. */
	public String [] actionParams = new String [50];	
	/** Specification of the input parameters of the second-level action; may be <code>null</code>. */
	public String [] actionParams2 = new String [50];
	
	/** <code>true</code> if there exists a nested (second-level) action in this AQ; <code>false</code> otherwise. */
	public boolean actionNesting;
	
	/** A flag that indicates whether this AQ involves join and/or selection. */
	public byte flag;	           
	public static final byte JOIN_FLAG = 0x01;
	public static final byte SELECTION_FLAG = 0x02;
		
	String triggeringJoinAttrN = Aorta.NULL_STRING;  // join attribute for the triggering device (sensor) 
	String triggeredJoinAttrN = Aorta.NULL_STRING;   // join attribute for the triggered device
	String joinFuncName = Aorta.NULL_STRING;         // the join function name, operator and constant
	String joinArithOp = Aorta.NULL_STRING;
	String joinConstant = Aorta.NULL_STRING;  
	
	String selectionAttrN = Aorta.NULL_STRING;    // selection attribute and its value
	String selectionAttrV = Aorta.NULL_STRING;    
	
	// timer function data structure
	String whereTimerFuncName = null;
	String [] whereTimerFuncParams = new String [50];
	WhereTimerQC whereThread = null;
	
    String startTimerFuncName = null;
	String [] startTimerFuncParams = new String [50];
	StartClause startThread = null;
	
	String stopTimerFuncName = null;
	String [] stopTimerFuncParams = new String [50];
	StopClause stopThread = null;
		
	/* -1 means the query will run for ever. 
       0 means the query is a snapshot query that runs only once */
	public static final int FOR_EVER = -1;
	public static final int ONCE = 0;
	int lifeTime = FOR_EVER;   
	String lifeTimeUnit = null;
	long startTime = System.currentTimeMillis(); // when the query starts
	LifeTimeClause lifeTimeThread = null;
	
	/** <code>true</code> if this AQ is running; <code>false</code> otherwise. */
	public boolean start_or_stop = false;
	
	/* Indicate whether the query optimizer should optimize different types of operators
	   true: optimized; false: not optimized */
	public static final boolean START_OPTIMIZATION = true;
	public static final boolean STOP_OPTIMIZATION = true;
	public static final boolean WHERE_TIMER_OPTIMIZATION = true;
	public static final boolean SCAN_DUPLICATE_OPTIMIZATION = true;	
	public static final boolean SELECTION_OPTIMIZATION = true;
	public static final boolean JOIN_OPTIMIZATION = true;
	public static final boolean ACTION_OPTIMZATION = true;
	
	/* Operators in the execution plan of the query */
	ScanOperator scanOp = null;
	JoinOperator joinOp = null;
    SelOperator selOp = null;
	ActionOperator actionOp = null, actionOp2 = null;
	Vector OpVector = new Vector(10,5);
	
	/* Compute the reaction time for one-time query execution */
	long eventStartTime;
	long eventEndTime;
		
	
	/** <code>true</code> if this AQ has been dropped; <code>false</code> otherwise. */
	public boolean queryDropped = false;
	
	/* File for recording query result (action invocation) logs */
    String logFile; 
    PrintWriter logFileWriter; 
    
	
	public AQ (String s[],String AQString, byte id, Action action, Action action2, String actionParams2[], String AQName, byte flag, byte type, boolean actionNesting, Aorta aortaqp) {
		
		sensorQueryString = s[0].toLowerCase();		
		
		if ( (flag & JOIN_FLAG) != 0)
		  {
		    int start = s[1].indexOf(";");
		    triggeringJoinAttrN = s[1].substring(0,start);
		    
		    int stop = s[1].indexOf(";",start+1);
		    joinFuncName = s[1].substring(start+1,stop);
		    
		    start = s[1].indexOf(";",stop + 1);
		    joinArithOp = s[1].substring(stop+1,start);
		    
		    joinConstant = s[1].substring(start + 1);
		    
		    triggeredJoinAttrN = s[2];
		}
		
		if ( (flag & SELECTION_FLAG) != 0)
		  {
		  	selectionAttrN = s[3]; 
		    selectionAttrV = s[4];
		  }
		
		int i = 5;
		while ( s[i] != Aorta.DELIMITER_STRING)
		  {actionParams[i-5] = s[i]; i++;}
		i++;
		
		whereTimerFuncName = s[i++];
		int j = 0;
		while (!s[i].equals(Aorta.DELIMITER_STRING))
		  whereTimerFuncParams[j++] = s[i++];
		i++;
		
		startTimerFuncName = s[i++];
		j = 0;
		while (!s[i].equals(Aorta.DELIMITER_STRING))
		  startTimerFuncParams[j++] = s[i++];
		i++;      
        
        stopTimerFuncName = s[i++];
		j = 0;
		while (!s[i].equals(Aorta.DELIMITER_STRING))
		  stopTimerFuncParams[j++] = s[i++];
		i++; 
		
	    if (!s[i].equals(Aorta.NULL_STRING))
	      lifeTime = Integer.parseInt(s[i++]);
	    lifeTimeUnit = s[i++];
	    
	    this.AQString = AQString;
		this.id = id;
		this.action = action;
		this.action2= action2;
		this.actionParams2 = actionParams2;
        this.name = AQName;
        this.flag = flag;
        this.type = type;
        this.actionNesting = actionNesting;
		this.aortaqp = aortaqp;		
		
		if (action2 != null)
		  {
		  	i = 0;
		  	while (actionParams2[i] != null)
		  	  {
		  	  	int len = actionParams2[i].length() - 1;
		  	  	if (actionParams2[i].charAt(0) == '\"' && actionParams2[i].charAt(len) == '\"')
		  	  	  actionParams2[i] = actionParams2[i].substring(1,len);
		  	   	i++;
		  	  }		  	
		  }						
		
		logFile = "../logs/single_query_logs/Q" + id + ".log"; 
		try {logFileWriter = new PrintWriter ( new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}        
        
        if (id <= 0) 
          {
          	System.err.println("Invalid query id: " + id + ".");
          	return;
          }
                  
        initialize();
                				
	}
	
	
	/** Initialize this AQ. */
	public synchronized void initialize ()
	  {
	    int size = aortaqp.AQVector.size(), i;
	    
	    // create the join operator
	    if ( (flag & JOIN_FLAG) != 0)
          {
             i = size;
		  	 if (aortaqp.MULTI_QUERY_OPTIMIZATION)
		  	   for (i=0;i<size;i++)
		  	     if ( ((AQ)aortaqp.AQVector.elementAt(i)).triggeringJoinAttrN.equals(triggeringJoinAttrN)
		  	  	      && ((AQ)aortaqp.AQVector.elementAt(i)).triggeredJoinAttrN.equals(triggeredJoinAttrN) 
		  	  	      && ((AQ)aortaqp.AQVector.elementAt(i)).joinFuncName.equals(joinFuncName) 
		  	  	      && ((AQ)aortaqp.AQVector.elementAt(i)).joinArithOp.equals(joinArithOp) 
		  	  	      && ((AQ)aortaqp.AQVector.elementAt(i)).joinConstant.equals(joinConstant) )
		  	  	   {joinOp = ((AQ)aortaqp.AQVector.elementAt(i)).joinOp; break;}
		  			  	 
		  	 if (i == size) joinOp = new JoinOperator(this);
		  	 OpVector.addElement(joinOp);  
          } 
          
        // create the selection operator 
        if ( (flag & SELECTION_FLAG) != 0)
		  {
		  	 i = size;
		  	 if (aortaqp.MULTI_QUERY_OPTIMIZATION)
		  	   for (i=0;i<size;i++)
		  	     if ( ((AQ)aortaqp.AQVector.elementAt(i)).selectionAttrN.equals(selectionAttrN)
		  	  	      && ((AQ)aortaqp.AQVector.elementAt(i)).selectionAttrV.equals(selectionAttrV) )
		  	  	   {selOp = ((AQ)aortaqp.AQVector.elementAt(i)).selOp; break;}
		  			  	 
		  	 if (i == size) selOp = new SelOperator(this);
               OpVector.addElement(selOp);                 
          }
	    
	    // create the scan operator
	    if (type == Aorta.EVENT_DRIVEN_QUERY)
          {
		  	scanOp = new ScanOperator(this);
		  	OpVector.addElement(scanOp);
          }
		  	
		// create the action operator
		i = size;
		if (aortaqp.MULTI_QUERY_OPTIMIZATION)
		  for (i=0;i<size;i++)
		   	if ( ((AQ)aortaqp.AQVector.elementAt(i)).action.equals(action))
		      {actionOp = ((AQ)aortaqp.AQVector.elementAt(i)).actionOp; break;}
		  			  	 
		if (i == size)
          actionOp = new ActionOperator(this, action, actionParams);
        OpVector.addElement(actionOp);
        
        if (actionNesting)
          {
          	i = size;
		    if (aortaqp.MULTI_QUERY_OPTIMIZATION)
		      {
		      	for (i=0;i<size;i++)
		   	      if ( ((AQ)aortaqp.AQVector.elementAt(i)).action.equals(action2))
		            {actionOp2 = ((AQ)aortaqp.AQVector.elementAt(i)).actionOp; break;}
		  			  	 
		        if (i == size)
		          for (i=0;i<size;i++)
		            if (((AQ)aortaqp.AQVector.elementAt(i)).action2 != null)
		   	          if ( ((AQ)aortaqp.AQVector.elementAt(i)).action2.equals(action2))
		                {actionOp2 = ((AQ)aortaqp.AQVector.elementAt(i)).actionOp2; break;}
		      }		      
		    
		    if (i == size)
              actionOp2 = new ActionOperator(this, action2, actionParams2);
            OpVector.addElement(actionOp2);
          }  
        
        // connect the operators into a AQ execution plan 
        if (type == Aorta.EVENT_DRIVEN_QUERY) 
	      if (!actionNesting)
	        {
	          if (joinOp != null)
	           {
	      	     scanOp.outputOperator = joinOp;
		  	     scanOp.outputQueue = joinOp.inputQueue;
		  	     joinOp.outputOperator = actionOp;
                 joinOp.outputQueue = actionOp.inputQueue;	
		       }
		      else
		        {
		  	      scanOp.outputOperator = actionOp;
                  scanOp.outputQueue = actionOp.inputQueue;
                  selOp.outputOperator = actionOp;
                  selOp.outputQueue = actionOp.inputQueue2;   
		        }
		    }
		  else
		    {
		       scanOp.outputOperator = joinOp;
		  	   scanOp.outputQueue = joinOp.inputQueue;
		  	   joinOp.outputOperator = actionOp2;
               joinOp.outputQueue = actionOp2.inputQueue;
               
               actionOp2.outputOperator = actionOp;
               actionOp2.outputQueue = actionOp.inputQueue;
               selOp.outputOperator = actionOp;
               selOp.outputQueue = actionOp.inputQueue2; 
		    }                                      
          
		if (type == Aorta.TRADITIONAL_CONTINUOUS_QUERY)
		  {
		  	if (selOp != null)
		  	  if (joinOp != null)
		  	  	{
		  	  	  selOp.outputOperator = joinOp;
		  	  	  selOp.outputQueue = joinOp.inputQueue;
		  	  	  
		  	  	  joinOp.outputOperator = actionOp;
		  	  	  joinOp.outputQueue = actionOp.inputQueue;
		  	    }
		  	  else
		  	    {
		  	  	  selOp.outputOperator = actionOp;
		  	  	  selOp.outputQueue = actionOp.inputQueue;
		  	  	}		  	  
		  }   
	    
          
        // create the threads for START, STOP and LIFETIME clause, and the timer function in the WHERE clause
        if (!whereTimerFuncName.equals(Aorta.NULL_STRING))
	      if (WHERE_TIMER_OPTIMIZATION)
	        {
	       	  whereThread = new WhereTimerQC(this);	       	  
	       	  whereThread.start();	       	  
	        }
        
        if (stopTimerFuncName.equals(Aorta.NULL_STRING))
	      {}
	    else 
	      {
	      	stopThread = new StopClause(this);
	        stopThread.start();
	      }
	    	    
	    if (startTimerFuncName.equals(Aorta.NULL_STRING))
	      {
	      	start();
	      	notifyAll(); 
	      }
	    else 
	      {
	      	startThread = new StartClause(this);
	        startThread.start();
	      }    
	        
	    if ((lifeTime != FOR_EVER) && (lifeTime != ONCE))
	      {
	      	lifeTimeThread = new LifeTimeClause(this);
	      	lifeTimeThread.start();
	      }      
	     	      	
	  }
	  
	
	/** Start this AQ. */
	public void start () 
	  {
		
		int i;
		long startTime, endTime;
		
		if (start_or_stop == true) 
		  return;		
		start_or_stop = true;
		
		for (i=0;i<OpVector.size();i++)
		  ((Operator)OpVector.elementAt(i)).start();		      		
		
		if (type == Aorta.TRADITIONAL_CONTINUOUS_QUERY)
		  { 
		  	if (selOp != null)
		  	  {
		  	  	selOp.getInput();
		  	  	if (selOp.outputOperator == joinOp)
		  	  	  joinOp.getInput();
		  	  }		  	  
            
            actionOp.action.q = this;
            
            actionOp.getInput();            
                        
            stop();
		  }   		
		
	  }
	
	
	/** Stop this AQ. */
	public void stop() 
	  {
		
		if (start_or_stop == false) 
		  return;
		  
		start_or_stop = false;
	
		for (int i=0;i<OpVector.size();i++)
		  ((Operator)OpVector.elementAt(i)).stop();
	  }	
	
	
	/** Test the satisfaction of the timer function in the WHERE clause. */
	public boolean testTimerCondition ()
      {
      	if (whereTimerFuncName.equals(TimerFunctions.ATTIME))
          if (!TimerFunctions.atTime(Integer.parseInt(whereTimerFuncParams[0]), Integer.parseInt(whereTimerFuncParams[1]), Integer.parseInt(whereTimerFuncParams[2]) ))
	        return false;
	        
	    if (whereTimerFuncName.equals(TimerFunctions.ONDAY))
          if (!TimerFunctions.onDay(Integer.parseInt(whereTimerFuncParams[0]), Integer.parseInt(whereTimerFuncParams[1]), Integer.parseInt(whereTimerFuncParams[2]) ))
	        return false;
	        
	    if (whereTimerFuncName.equals(TimerFunctions.INTIMEINTERVAL))
          if (!TimerFunctions.inTimeInterval(Integer.parseInt(whereTimerFuncParams[0]), Integer.parseInt(whereTimerFuncParams[1]), Integer.parseInt(whereTimerFuncParams[2]), Integer.parseInt(whereTimerFuncParams[3]), Integer.parseInt(whereTimerFuncParams[4]), Integer.parseInt(whereTimerFuncParams[5]) ))
	        return false;
	    
	    if (whereTimerFuncName.equals(TimerFunctions.INDAYINTERVAL))
          if (!TimerFunctions.inDayInterval(Integer.parseInt(whereTimerFuncParams[0]), Integer.parseInt(whereTimerFuncParams[1]), Integer.parseInt(whereTimerFuncParams[2]), Integer.parseInt(whereTimerFuncParams[3]), Integer.parseInt(whereTimerFuncParams[4]), Integer.parseInt(whereTimerFuncParams[5]) ))
	        return false;
	       
	    if (whereTimerFuncName.equals(TimerFunctions.ONDAYOFWEEK))
          if (!TimerFunctions.onDayOfWeek(whereTimerFuncParams[0]))
	        return false;
	        
	    if (whereTimerFuncName.equals(TimerFunctions.INDAYOFWEEKINTERVAL))
          if (!TimerFunctions.inDayOfWeekInterval(whereTimerFuncParams[0], whereTimerFuncParams[1]))
	        return false;
	           
	    return true;
	    
      } 
      
            
    /** Destroy this AQ. */
    public void destroy()
      {
        stop();
           
        whereThread = null;
        startThread = null;
        stopThread = null;
        lifeTimeThread = null;
           
        scanOp = null;
        selOp = null;
        joinOp = null;
        actionOp = null;
        actionOp2 = null;
        actionParams = null;
        actionParams2 = null;
           
        logFileWriter.close();
           
        System.gc();
           
        queryDropped = true;           
      }         
	
} 