/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;

/** Action in Aorta. */
public abstract class Action {
	
	/** Name of the action. */
	public String name;
	
	/** Pathname of the action profile. */
	public String profile;
	
	/** 
	 * Input parameters of the action.	 
	 */
	public Parameter params [] = new Parameter [10];
	
	/** Number of action input parameters. */
	public int paramNum = 0;	
	
	/** Return type of the action. */
	public String returnType;
	
	/** 
	 * Information about the type of device that the action is executed on.
	 */	
	public DeviceInfo devInf; 
	
	/** 
	 * Top-level operation sequence of the action.
	 */
	public OperationSequence opSeq;
	
	/** The value of this flag is set to be <code>0</code> if this action is system-provided; <code>1</code> if it is user-defined. */
	public byte flag;	
	public static byte SYSTEM_BUILT_IN = 0;	
	public static byte USER_DEFINED = 1;
	
	/** The cost assigned to an atomic deterministic operation whose cost is currently unmeasured. */
	public static final double UNMEASURED_ESTIMATED_COST = 1;
	
	AQ q;		
	String profileString = "";
	int start,stop,end1,end2;
	int count;	  		
	
	
	public Action () 
	  {
		for (int i=0;i<10;i++)
		  params[i] = new Parameter ();
		
		devInf = new DeviceInfo();
		
		opSeq = new OperationSequence();		
	  }
	
	
	/** 
	 * Parse the action profile.
	 *
	 * @return <code>true</code> if the parsing is successful; 
     *         <code>false</code> otherwise.
     *
     * @see #parseOperationSequence
	 * @see #parseOperationSet
	 * @see #parseOperation
	 * @see #parseAtomicOperation 
	 */
	public boolean parseProfile ()  
	  {
	  	try 
	  	  {
	  		BufferedReader br = new BufferedReader (new FileReader(profile));
	  		String line;
	  		
	  		while ( (line=br.readLine()) != null)
	  		  profileString += line;
	  		br.close();
	  		  
	  		start = profileString.indexOf("<actionProfile>");
	  		if (start == -1)
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag <actionProfile> is missing.");
	  		  	return false;
	  		  }
	  		start += "<actionProfile>".length();
	  	    stop = profileString.indexOf("</actionProfile>",start);
	  	    if (stop == -1)
	  	      {
	  		  	System.err.println("Fail to parse the action profile. Tag </actionProfile> is missing.");
	  		  	return false;
	  		  }
	  		end1 = stop;
	  		  
	  		start = profileString.indexOf("<name>",start);
	  		if ( (start == -1) || (start >= end1) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag <name> is missing.");
	  		  	return false;
	  		  }
	  		start += "<name>".length();
	  		stop = profileString.indexOf("</name>",start);
	  		if ( (stop == -1) || (stop >= end1) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag </name> is missing.");
	  		  	return false;
	  		  }
	  		if (!name.equals(profileString.substring(start,stop).trim()))
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Incorrect action name.");
	  		  	return false;
	  		  }
	  		start = stop + "</name>".length();
	  		
	  		start = profileString.indexOf("<params>",start);
	  		if ( (start == -1) || (start >= end1) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag <params> is missing.");
	  		  	return false;
	  		  }
	  		start += "<params>".length();
	  		stop = profileString.indexOf("</params>",start);
	  		if ( (stop == -1) || (stop >= end1) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag </name> is missing.");
	  		  	return false;
	  		  } 
	  		end2 = stop;	  		
            count = 0;
     		while (true)
	  		  {
	  		  	start = profileString.indexOf("<param",start);
	  		  	if ( (start == -1) || (start >= end2) ) break;
	  		  	start = profileString.indexOf(">",start);
	  		    if ( (start == -1) || (start >= end2) ) break;
	  		    start ++;
	  		    stop = profileString.indexOf("</param",start);
	  		    if ( (stop == -1) || (stop >= end2) ) break;
	  		    params[count++].symbol = profileString.substring(start,stop).trim();
	  		    start = stop + "<param".length();		  	
	  		  }
	  		if (count != paramNum)
	  		  {
	  		  	System.err.print("Fail to parse the action profile. Invalid format within the <params></params> tag.");
	  		  	return false;
	  		  }
	  		  
	  		start = end2 + "</params>".length();
	  		start = profileString.indexOf("<returnType",start);
	  		if ( (start == -1) || (start >= end1) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag <returnType> is missing.");
	  		  	return false;
	  		  }
	  		start += "<returnType>".length(); 
	  		stop = profileString.indexOf("</returnType>",start);
	  		if ( (stop == -1) || (stop >= end1) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag </returnType> is missing.");
	  		  	return false;
	  		  }
	  	    returnType = profileString.substring(start,stop).trim();
	  	    start = stop + "</returnType>".length();
	  	    
	  	    start = profileString.indexOf("<device>",start);
	  	    if ( (start == -1) || (start >= end1) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag <device> is missing.");
	  		  	return false;
	  		  }
	  		start += "<device>".length();
	  		stop = profileString.indexOf("</device>",start);
	  		if ( (stop == -1) || (stop >= end1) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag </device> is missing.");
	  		  	return false;
	  		  }
	  		end2 = stop;
	  		start = profileString.indexOf("<type>",start);
	  		if ( (start == -1) || (start >= end2) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag <type> within tag <device></device> is missing.");
	  		  	return false;
	  		  }
	  		start += "<type>".length();
	  		stop = profileString.indexOf("</type>",start);
	  		if ( (stop == -1) || (stop >= end2) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag </type> within tag <device></device> is missing.");
	  		  	return false;
	  		  }
	  		devInf.type = profileString.substring(start, stop).trim();
	  		start = stop + "</type>".length();
	  		start = profileString.indexOf("<model>",start);
	  		if ( (start == -1) || (start >= end2) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag <model> within tag <device></device> is missing.");
	  		  	return false;
	  		  }
	  		start += "<model>".length();
	  		stop = profileString.indexOf("</model>",start);
	  		if ( (stop == -1) || (stop >= end2) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag </model> within tag <device></device> is missing.");
	  		  	return false;
	  		  }
	  		devInf.model = profileString.substring(start, stop).trim(); 
	  		start = stop + "</model>".length();
	  		start = profileString.indexOf("<physicalStatusInvolved>",start);
	  		if ( (start == -1) || (start >= end2) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag <physicalStatusInvolved> within tag <device></device> is missing.");
	  		  	return false;
	  		  }
	  		start += "<physicalStatusInvolved>".length();
	  		stop = profileString.indexOf("</physicalStatusInvolved>",start);
	  		if ( (stop == -1) || (stop >= end2) )
	  		  {
	  		  	System.err.println("Fail to parse the action profile. Tag </physicalStatusInvolved> within tag <device></device> is missing.");
	  		  	return false;
	  		  }
	  	    end2 = stop;
	  		count = 0;
     		while (true)
	  		  {
	  		  	start = profileString.indexOf("<attribute>",start);
	  		  	if ( (start == -1) || (start >= end2) ) break;
	  		  	start = profileString.indexOf("<name>",start);
	  		    if ( (start == -1) || (start >= end2) ) break;
	  		    start += "<name>".length();
	  		    stop = profileString.indexOf("</name>",start);
	  		    if ( (stop == -1) || (stop >= end2) ) break;
	  		    devInf.dps[count].attribute = profileString.substring(start,stop).trim();
	  		    start = stop + "</name>".length();
	  		    start = profileString.indexOf("<value>",start);
	  		    if ( (start == -1) || (start >= end2) ) break;
	  		    start += "<value>".length();
	  		    stop = profileString.indexOf("</value>",start);
	  		    if ( (stop == -1) || (stop >= end2) ) break;
	  		    devInf.dps[count].symbol = profileString.substring(start,stop).trim();
	  		    start = stop + "</value>".length(); 
	  		    count ++; 	
	  		  }
	  		devInf.dpsNum = count;  		
	  		  
	  		start = profileString.indexOf("<operationSequence>",start);
	  		if (start == -1)
	  		  {
	  		  	System.err.println("Fail to parse the action profile. The top-level tag <operationSequence> is missing.");
	  		  	return false;
	  		  }
	  		start += "<operationSequence>".length();
	  		if (!parseOperationSequence(opSeq))
	  		  return false;	  	
	  	    
	  	  }
	  	catch (Exception e) {e.printStackTrace();}
	  		
	  	return true;
	  } 
	  
	  
	/** 
	 * Parse the XML texts within an <code>operationSequence</code>
	 * element in the action profile. 
	 *
	 * @param opSeq an instance of {@link OperationSequence} that will be 
	 *              initialized correspondingly
	 *
	 * @return <code>true</code> if the parsing is successful; 
     *         <code>false</code> otherwise.
     *
     * @see OperationSequence
     * @see #parseProfile 
	 */ 
	public boolean parseOperationSequence(OperationSequence opSeq) 
	  {
		int start1, start2, start3, start_t, end;
		
		while (true)
		  {
		  	start1 = profileString.indexOf("<operation>",start);
		    if (start1 == -1) start1 = Integer.MAX_VALUE;
		    
		    start2 = profileString.indexOf("<operationSequence>",start);
		    if (start2 == -1) start2 = Integer.MAX_VALUE;
		    
		    start3 = profileString.indexOf("<operationSet>",start);
		    if (start3 == -1) start3 = Integer.MAX_VALUE;
		    
	 	    start_t = Math.min(start1,Math.min(start2,start3));
		    if (start_t == Integer.MAX_VALUE) break;
		    
		    end = profileString.indexOf("</operationSequence>",start_t);
		    if (end == -1) break;
		    else start = start_t;
		    
		    if (start == start1)
		     {
		  	   start += "<operation>".length();
		  	   Operation op = new Operation();
		  	   if (!parseOperation(op)) return false;
		  	   opSeq.op_units[opSeq.opUnitNum++] = op;
		     }
	        else
		      if (start == start2)
		        {
		          start += "<operationSequence>".length();
		          OperationSequence opSeq2 = new OperationSequence();
		          if (!parseOperationSequence(opSeq2)) return false;
		          opSeq.op_units[opSeq.opUnitNum++] = opSeq2;	      	
		        }
		      else
		        {
		          start += "<operationSet>".length();
		          OperationSet opSet = new OperationSet();
		          if (!parseOperationSet(opSet)) return false;
		          opSeq.op_units[opSeq.opUnitNum++] = opSet;
		        }
		  }		
		
		return true;		
      } 
    
    
    /** 
	 * Parse the XML texts within an <code>operationSet</code>
	 * element in the action profile. 
	 *
	 * @param opSet an instance of {@link OperationSet} that will be 
	 *              initialized correspondingly
	 *
	 * @return <code>true</code> if the parsing is successful; 
     *         <code>false</code> otherwise.
     *
     * @see OperationSet
     * @see #parseProfile 
	 */ 
    public boolean parseOperationSet(OperationSet opSet) 
      {
        int start1, start2, start3, start_t, end;
		
		while (true)
		  {
		  	start1 = profileString.indexOf("<operation>",start);
		    if (start1 == -1) start1 = Integer.MAX_VALUE;
		    
		    start2 = profileString.indexOf("<operationSequence>",start);
		    if (start2 == -1) start2 = Integer.MAX_VALUE;
		    
		    start3 = profileString.indexOf("<operationSet>",start);
		    if (start3 == -1) start3 = Integer.MAX_VALUE;
		    
		    start_t = Math.min(start1,Math.min(start2,start3));
		    if (start_t == Integer.MAX_VALUE) break;
		    
		    end = profileString.indexOf("</operationSet>",start_t);
		    if (end == -1) break;
		    else start = start_t;
		  
		    if (start == start1)
		      {
		  	    start += "<operation>".length();
		  	    Operation op = new Operation();
		  	    if (!parseOperation(op)) return false;
		  	    opSet.op_units[opSet.opUnitNum++] = op;
		      }
		    else
		      if (start == start2)
		        {
		          start += "<operationSequence>".length();
		          OperationSequence opSeq = new OperationSequence();
		          if (!parseOperationSequence(opSeq)) return false;
		          opSet.op_units[opSet.opUnitNum++] = opSeq;	      	
		        }
		      else
		        {
		          start += "<operationSet>".length();
		          OperationSet opSet2 = new OperationSet();
		          if (!parseOperationSet(opSet2)) return false;
		            opSet.op_units[opSet.opUnitNum++] = opSet2;
		        }
		  }		
		
		return true;      	
      }
    
    
    /** 
	 * Parse the XML texts within an <code>operation</code>
	 * element in the action profile. 
	 *
	 * @param op an instance of {@link Operation} that will be 
	 *           initialized correspondingly
	 *
	 * @return <code>true</code> if the parsing is successful; 
     *         <code>false</code> otherwise.
     *
     * @see Operation
     * @see #parseProfile 
	 */ 
    public boolean parseOperation(Operation op) 
      {
         start = profileString.indexOf("<atomicOperation>",start);
         if (start == -1)
	       {
	  	 	 System.err.println("Fail to parse the action profile. Tag <atomicOperation> within tag <operation></operation> is missing.");
	  	  	 return false;
	       }
	     start += "<atomicOperation>".length();
	     stop = profileString.indexOf("</atomicOperation>",start);
	     if (stop == -1)
	       {
	  	 	 System.err.println("Fail to parse the action profile. Tag </atomicOperation> within tag <operation></operation> is missing.");
	  	  	 return false;
	       }
	     op.atomOp.name = profileString.substring(start,stop).trim();
	     if (!parseAtomicOperation(op.atomOp)) return false;
         start = stop + "</atomicOperation>".length();
         start = profileString.indexOf("<number>",start);
         if (start == -1)
	       {
	  	 	 System.err.println("Fail to parse the action profile. Tag <number> within tag <operation></operation> is missing.");
	  	  	 return false;
	       }
	     start += "<number>".length();
	     stop = profileString.indexOf("</number>",start);
	     if (stop  == -1)
	       {
	  	 	 System.err.println("Fail to parse the action profile. Tag </number> within tag <operation></operation> is missing.");
	  	  	 return false;
	       }
	     op.number = profileString.substring(start,stop).trim();
      
         start = profileString.indexOf("</operation>",stop);
         if (start  == -1)
	       {
	  	 	 System.err.println("Fail to parse the action profile. Tag </operation> is missing.");
	  	  	 return false;
	       }
	       
         return true;      	
      }
	
	
	/** 
	 * Parse the XML texts within an <code>atomicOperation</code>
	 * element in the action profile. 
	 *
	 * @param atomOp an instance of {@link AtomicOperation} that will be 
	 *               initialized correspondingly
	 *
	 * @return <code>true</code> if the parsing is successful; 
     *         <code>false</code> otherwise.
     *
     * @see AtomicOperation
     * @see #parseOperation
	 */ 
	public boolean parseAtomicOperation (AtomicOperation atomOp) 
	  {
		try 
		  {
	  	    int start, stop;
	  
	        BufferedReader br;
	        String line,aocString = "";
	  
	        if (devInf.type.equals("sensor"))
	          br = new BufferedReader (new FileReader (Aorta.SENSOR_AOC_FILE) );
	        else
	          if (devInf.type.equals("camera"))
	            br = new BufferedReader (new FileReader (Aorta.CAMERA_AOC_FILE) );
	          else
	            if (devInf.type.equals("phone"))
	              br = new BufferedReader (new FileReader (Aorta.PHONE_AOC_FILE) );
	            else
	              if (devInf.type.equals("robot"))
	                br = new BufferedReader (new FileReader (Aorta.ROBOT_AOC_FILE) );
	              else
	                return false;
	        while ( (line=br.readLine()) != null)
	          aocString += line;
	        br.close();
	    
	        start = 0; 
	        while (true)
	          {
	            start = aocString.indexOf("<model>",start);
	            if (start  == -1)
	              {
	  	 	        System.err.println("Fail to parse the atomic operation configuration profile. No device of the specified model is found.");
	  	  	        return false;
	              }
	            start += "<model>".length();
	            stop = aocString.indexOf("<",start);
	            if (devInf.model.equals(aocString.substring(start,stop).trim()))
	              {
	      	        while (true)
	      	          {
	      	        	start = aocString.indexOf("<name>",start);
	      	            if (start == -1)
	      	              {
	      	      	        System.err.println("Fail to parse the atomic operation configuration profile. No atomic operation of the corresponding name " + atomOp.name + " is found.");
	  	  	                return false;
	      	              }
	      	            start += "<name>".length();
	      	            stop = aocString.indexOf("</name>",start);
	      	            if (stop == -1)
	      	              {
	      	      	        System.err.println("Fail to parse the atomic operation configuration profile. No atomic operation of the corresponding name " + atomOp.name + " is found.");
	  	  	                return false;
	      	              }
	      	            if (atomOp.name.equals(aocString.substring(start,stop).trim()))
	      	              {
	      	          	    start = stop + "</name>".length();	      	          	    
	      	                start = aocString.indexOf("<estimatedCost>",start);
	      	          	    if (start == -1)
	      	          	      {
	      	      	            System.err.println("Fail to parse the atomic operation configuration profile. Invalid format.");
	  	  	                    return false;
	      	                  }
	      	                start += "<estimatedCost>".length();
	      	                stop = aocString.indexOf("</estimatedCost>",start);
	      	                if (stop == -1)
	      	          	      {
	      	      	            System.err.println("Fail to parse the atomic operation configuration profile. Invalid format.");
	  	  	                    return false;
	      	                  }
	      	                if (aocString.substring(start,stop).trim().equals("N.A."))
	      	                  atomOp.estimatedCost = UNMEASURED_ESTIMATED_COST;
	      	                else
	      	                  atomOp.estimatedCost = Float.parseFloat(aocString.substring(start,stop).trim());
	      	          	    return true;
	      	              } 	      	        
	      	          } 	      	    
	              }
	          }	    
	  
	      }
	    catch (Exception e) {e.printStackTrace(); return false;}		
	  }
	  
	
	/**
	 * This function should be overloaded by each system built-in or user-defined action 
	 * class that inherits this super class. 
	 *
	 * @param s a string array that contains the instantiated values of the action input parameters
	 *
	 * @see Photo#doAction
	 * @see SendMsg#doAction
	 * @see SendPhoto#doAction
	 */
	public Object doAction(String [] s)  // the action that is executed on the device
	  {
	  	return new Object();
	  }  
		
}