/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.util.*;

/**
 * Generate a synthetic action workload with a set of user-specified parameters:
 * (1) The arrival pattern of the action requests ({@link AortaSchedulingSimulator#POISSON} or {@link AortaSchedulingSimulator#CONSTANT}) 
 * (2) The distribution of # candidate devices of the requests ({@link AortaSchedulingSimulator#UNIFORM} or {@link AortaSchedulingSimulator#ZIPFIAN})  
 * (3) The name of the action to be scheduled ({@link AortaSchedulingSimulator#PHOTO} or {@link AortaSchedulingSimulator#WARN})
 * (4) Total number of devices involved in the scheduling ({@link AortaSchedulingSimulator#deviceNum})
 * (5) Total number of requests arrived ({@link AortaSchedulingSimulator#requestNum})
 * (6) The arrival rate of the requests
 *
 * @see AortaSchedulingSimulator
 */
public class ActionWorkload {
	
    public static double MIN_PHOTO_COST;
    public static double MAX_PHOTO_COST;
    public static double AVG_PHOTO_COST;
    public static double MIN_WARN_COST;
    public static double MAX_WARN_COST; 
    public static double AVG_WARN_COST; 
    
    public static double MIN_PHOTO_DEADLINE;
    public static double MAX_PHOTO_DEADLINE;
    public static double MIN_WARN_DEADLINE;
    public static double MAX_WARN_DEADLINE;    
    public static double W = 2.0; // MAX_ACTION_DEADLINE is W times MIN_ACTION_DEADLINE
    
    double requestStartTime = 0;
    
    Random startTimeRan = new Random(0), deadlineRan = new Random(0), ran = new Random(0);
    
    double zipf_nums [];
    
    Vector usedZoom = new Vector(4);
      
    
    ActionWorkload () {
    	
    	CameraSimulator camSim = new CameraSimulator("0");
    	
        MIN_PHOTO_COST = (2*camSim.STANDARD_POLLING_TIME + camSim.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST)/1000.0;
        MAX_PHOTO_COST = (camSim.computeHeadMovementCost(camSim.MAX_PAN, camSim.MIN_TILT, camSim.MAX_ZOOM) + 2*camSim.STANDARD_POLLING_TIME + camSim.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST)/1000.0;
        AVG_PHOTO_COST = (MIN_PHOTO_COST + MAX_PHOTO_COST)/2.0;
        
        MIN_WARN_COST = (ER1Simulator.STANDARD_POLLING_TIME + ER1Simulator.DEFAULT_SOUND_COST)/1000.0;
        MAX_WARN_COST = (ER1Simulator.STANDARD_POLLING_TIME + ER1Simulator.DEFAULT_SOUND_COST + ER1Simulator.ROTATEPERDEG_ESTIMATED_COST*360 + Math.sqrt(Math.pow(ER1Simulator.MAX_LOC_X-ER1Simulator.MIN_LOC_X,2)+Math.pow(ER1Simulator.MAX_LOC_Y-ER1Simulator.MIN_LOC_Y,2))*100*ER1Simulator.MOVEPERCM_ESTIMATED_COST)/1000.0;
        AVG_WARN_COST = (MIN_WARN_COST + MAX_WARN_COST)/2.0;
        
        MIN_PHOTO_DEADLINE = AVG_PHOTO_COST;
        MAX_PHOTO_DEADLINE = W*MIN_PHOTO_DEADLINE;
        
        MIN_WARN_DEADLINE = AVG_WARN_COST;
        MAX_WARN_DEADLINE = W*MIN_WARN_DEADLINE;
        
        //printConstants();      
        
        if (AortaSchedulingSimulator.candidateDeviceDistribution == AortaSchedulingSimulator.ZIPFIAN)
          {
          	zipf_nums = new double [AortaSchedulingSimulator.deviceNum];
		
		    generateZipfianNums();
		  }	     
    } 
      
    	
	/** Generate the action workload. */
	public Vector generate () 
	  {
		Vector requests = new Vector (AortaSchedulingSimulator.requestNum,10);
		
		int i,j,k;
		
		if (AortaSchedulingSimulator.candidateDeviceDistribution == AortaSchedulingSimulator.UNIFORM)
		  for (i=0;i<AortaSchedulingSimulator.requestNum;i++)
		    {
		  	  double cost;
		  	  ActionRequest ar = new ActionRequest();
		  	  
		  	  ar.aid = String.valueOf(i+1);
		  	  
		  	  int devNum = 1 + Math.abs(ran.nextInt()) % AortaSchedulingSimulator.deviceNum;		  	  
		  	  	
		      for (j=0;j<devNum;j++)
		        {
		  	      ar.didVector.addElement(String.valueOf(j+1));
		  	
		  	      cost = ranActionCost(ar)/1000.0;
		  	      
		  	      ar.costVector.addElement(String.valueOf(cost));	  		  	
		        }
		        
		      ar.deadline = ranDeadline();
		      
		      ar.startTime = getNextStartTime();
		      
		      requests.addElement((ActionRequest)ar);
		    }		  	
		
		if (AortaSchedulingSimulator.candidateDeviceDistribution == AortaSchedulingSimulator.ZIPFIAN)
		  {
		  	int id = 1, size, mark;
			double singleRequestNum, singledeviceNum, totalRequestNum = 0;
			double cost, startTime;
			String aid;
			ActionRequest ar, ar2;		
			
			for (i=AortaSchedulingSimulator.deviceNum-1;i>=0;i--)
		  	  {
		  		singleRequestNum = zipf_nums[i]*AortaSchedulingSimulator.requestNum*(i+1);
		  		
		  		singleRequestNum = Math.round(singleRequestNum);
		  		
		  		if (singleRequestNum == 0)
		  	  	  singleRequestNum = 1;
		  		
		  		if (i==0)
		  	  	  singleRequestNum = AortaSchedulingSimulator.requestNum - totalRequestNum;
		  		else
		  	      totalRequestNum += singleRequestNum;
		  	  
		  		singledeviceNum = i+1;
		  		for (j=0;j<singleRequestNum;j++)
		  	  	  {
		  	  		ar = new ActionRequest ();
		  	  		
		  	    	ar.aid = String.valueOf(id++);
		  	    	
		  	    	for (k=0;k<singledeviceNum;k++)
		  	          {
		  	        	ar.didVector.addElement(String.valueOf(k+1));
		  	        			  	
		  	        	cost = ranActionCost(ar)/1000.0;
		  	        	
		  	        	ar.costVector.addElement(String.valueOf(cost));		  	      
		  	      	  }
		  	      	  
		  	   		ar.deadline = ranDeadline();
		  	   		
		  	   		ar.startTime = getNextStartTime();
		  	   		
		  	   		size = requests.size();
		  	   		
		  	   		if (size > 1)
		  	   		  mark = Math.abs(ran.nextInt())%size;
		  	   		else
		  	   		  mark = 0;		  	   		  
		  	   		
		  	   		if (size == 0)
		  	   		  requests.insertElementAt((ActionRequest)ar, mark);
		  	   		else 
		  	   		  {
		  	   		  	ar2 = (ActionRequest)requests.elementAt(mark);
		  	   		  	
		  	   		  	startTime = ar.startTime;
		  	   		  	
		  	   		  	aid = ar.aid;
		  	   		  	
		  	   		  	ar.startTime = ar2.startTime;
		  	   		  	
		  	   		  	ar.aid = ar2.aid;
		  	   		  	
		  	   		  	requests.removeElementAt(mark);
		  	   		  	
		  	   		  	requests.insertElementAt((ActionRequest)ar,mark);
		  	   		  	
		  	   		  	ar2.startTime = startTime;
		  	   		  	
		  	   		  	ar2.aid = aid;
		  	   		  	
		  	   		  	requests.addElement((ActionRequest)ar2);		  	   		  	
		  	   		  }		  	   		
		  	   			
		  	    	if (requests.size() == AortaSchedulingSimulator.requestNum) 
		  	      	  {
		  	      	  	//requests = polish(requests); 
		  	      	  	
		  	      	  	return requests;
		  	      	  }	
		  	  	  } 
		  	  
		  		for (j=0;j<=i;j++)
		  	  	  zipf_nums[j] -= zipf_nums[i]; 	
		  	  }	  	
		  }	   
	
		//requests = polish(requests);
	
		return requests;
			
	  }
	  
	  
	/*public Vector polish(Vector requests)
	  {
	  	double NUM = 5;
	  	
	  	int count = 0;
	  	int i;
	  	ActionRequest ar;
	  	
	  	while (count < requests.size())
	  	  {
	  	  	double minDeadline = Double.MAX_VALUE, maxDeadline = Double.MIN_VALUE;
	  	  	
	  	  	for (i=count;i<count+NUM;i++)
	  	  	  {
	  	  	  	ar = (ActionRequest)requests.elementAt(i);
	  	  	  	
	  	  	  	if (minDeadline > ar.deadline)
	  	  	  	  minDeadline = ar.deadline;
	  	  	  	  
	  	  	  	if (maxDeadline < ar.deadline)
	  	  	  	  maxDeadline = ar.deadline;	  	  	  	
	  	  	  }  	  	
	  	  	
	  	  	ar = (ActionRequest)requests.elementAt(count);
	  	  	
	  	  	Vector didVector = ar.didVector;
	  	  	
	  	  	Vector costVector = ar.costVector;
	  	  	
	  	  	Vector statusVector = ar.statusVector;
	  	  	
	  	  	for (i=count;i<count+NUM;i++)
	  	  	  {
	  	  	  	ar = (ActionRequest)requests.elementAt(i);
	  	  	  	
	  	  	  	ar.didVector = didVector;
	  	  	  	
	  	  	  	ar.costVector = costVector;
	  	  	  	
	  	  	  	ar.statusVector = statusVector;
	  	  	  	
	  	  	  	ar.deadline = maxDeadline - (i-count)*(maxDeadline-minDeadline)/(NUM-1.0);	  	  	  	
	  	  	  }
	  	  	
	  	  	count += NUM;
	  	  }
	  	
	  	return requests;	  	
	  }*/
	  
	  
	/*public Vector sort(Vector requests)
	  {
	  	byte MIN = 0x01;
	  	byte AVG = 0x02;
	  	byte MAX = 0x03;
	  	byte sortTag = MIN;
	  	
	  	double PERCENT = 1.0; 
	  	
	  	int i,j, size=requests.size(), id=requests.size();
	  	double cost;
	  	Tuple t = null;
	  	double startTime [] = new double[requests.size()];
	  	Vector sortedRequests = new Vector(requests.size());	  		  	
	  	ActionRequest ar, sel_ar = null;
	  		
	  	for (i=0;i<requests.size();i++)
		  startTime[i]=((ActionRequest)requests.elementAt(i)).startTime;
	  	
	  	if (sortTag == MIN)  // sort the minimum cost
	  	  {
	  	  	while (requests.size() - (int)(size*(1-PERCENT)) > 0)
	  	  	  {
	  	  	  	double minCost = Double.MAX_VALUE;
	  	  	  		  	  	  	
	  	  	  	for (i=0;i<requests.size();i++)
	  	  	      {
	  	  	  	    ar = (ActionRequest)requests.elementAt(i);
	  	  	  	
	  	  	  	    for (j=0;j<ar.costVector.size();j++)
	  	  	  	      {
	  	  	  	  	    cost = Double.parseDouble((String)ar.costVector.elementAt(j));
	  	  	  	  	    
	  	  	  	  	    if (minCost > cost)
	  	  	  	  	      {minCost = cost; t = (Tuple)ar.statusVector.elementAt(j); sel_ar = ar;}
	  	  	  	      }	  	  	  	
	  	  	      }
	  	  	     
	  	  	     requests.removeElement((ActionRequest)sel_ar);
	  	  	     
	  	  	     sel_ar.aid = String.valueOf(id);
	  	  	     id--;
	  	  	     
	  	  	     Vector costVector = new Vector(sel_ar.costVector.size(),10);
	  	  	     
	  	  	     Vector statusVector = new Vector(sel_ar.statusVector.size(),10);
	  	  	     
	  	  	     for (i=0;i<sel_ar.costVector.size();i++)
	  	  	       costVector.addElement(String.valueOf(minCost));
	  	  	       
	  	  	     for (i=0;i<sel_ar.statusVector.size();i++)
	  	  	       statusVector.addElement((Tuple)t);
	  	  	     
	  	  	     sel_ar.costVector = costVector;
	  	  	     
	  	  	     sel_ar.statusVector = statusVector;
	  	  	     
	  	  	     sel_ar.startTime = startTime[id];
	  	  	    	  	  	     
	  	  	     sortedRequests.insertElementAt((ActionRequest)sel_ar,0);
	  	  	   }
	  	  }
	  	else
	  	  if (sortTag == AVG)  // sort the average cost
	  	    {
	  	      while (requests.size() - (int)(size*(1-PERCENT)) > 0)
	  	  	    {
	  	  	  	  double minAvgCost = Double.MAX_VALUE;	  	      
	  	  	  	  
	  	  	  	  for (i=0;i<requests.size();i++)
	  	  	        {
	  	  	  	      ar = (ActionRequest)requests.elementAt(i);
	  	  	  	    
	  	  	  	      double avgCost = 0.0;
	  	  	  	    
	  	  	  	      for (j=0;j<ar.costVector.size();j++)
	  	  	  	        {
	  	  	  	  	      cost = Double.parseDouble((String)ar.costVector.elementAt(j));
	  	  	  	  	      
	  	  	  	  	      avgCost += cost;
	  	  	  	        }
	  	  	  	        
	  	  	  	      avgCost /= ar.costVector.size();	
	  	  	  	      
	  	  	  	      if (minAvgCost > avgCost)
	  	  	  	        {minAvgCost = avgCost; sel_ar = ar;}  	  	  	
	  	  	        }
	  	  	     
	  	  	       requests.removeElement((ActionRequest)sel_ar);
	  	  	     
	  	  	       sel_ar.aid = String.valueOf(id);
	  	  	       id--;
	  	  	     
	  	  	       sel_ar.startTime = startTime[id];
	  	  	       
	  	  	       sortedRequests.insertElementAt((ActionRequest)sel_ar,0);
	  	  	     }		
	  	    }
	  	  else	// sort the maximum cost
	  	    {
	  	      while (requests.size() - (int)(size*(1-PERCENT)) > 0)
	  	  	    {
	  	  	  	  double minMaxCost = Double.MAX_VALUE;
	  	      
	  	          for (i=0;i<requests.size();i++)
	  	  	        {
	  	  	  	      ar = (ActionRequest)requests.elementAt(i);
	  	  	  	    
	  	  	  	      double maxCost = Double.MIN_VALUE;
	  	  	  	    
	  	  	  	      for (j=0;j<ar.costVector.size();j++)
	  	  	  	        {
	  	  	  	  	      cost = Double.parseDouble((String)ar.costVector.elementAt(j));
	  	  	  	  	      
	  	  	  	  	      if (maxCost < cost)
	  	  	  	  	        {maxCost = cost; t = (Tuple)ar.statusVector.elementAt(j);}
	  	  	  	        }	
	  	  	  	      
	  	  	  	      if (minMaxCost > maxCost)
	  	  	  	        {minMaxCost = maxCost; sel_ar = ar;}  	  	  	
	  	  	        }
	  	  	     
	  	  	       requests.removeElement((ActionRequest)sel_ar);
	  	  	     
	  	  	       sel_ar.aid = String.valueOf(id);
	  	  	       id--;
	  	  	       
	  	  	       Vector costVector = new Vector(sel_ar.costVector.size(),10);
	  	  	       
	  	  	       Vector statusVector = new Vector(sel_ar.statusVector.size(),10);
	  	  	       
	  	  	       for (i=0;i<sel_ar.costVector.size();i++)
	  	  	         costVector.addElement(String.valueOf(minMaxCost));
	  	  	         
	  	  	       for (i=0;i<sel_ar.statusVector.size();i++)
	  	  	         statusVector.addElement((Tuple)t);
	  	  	     
	  	  	       sel_ar.costVector = costVector;
	  	  	     
	  	  	       sel_ar.statusVector = statusVector; 	  	    
	  	  	     
	  	  	       sel_ar.startTime = startTime[id];
	  	  	    	  	  	     
	  	  	       sortedRequests.insertElementAt((ActionRequest)sel_ar,0);
	  	  	     }	
	  	    }	  	    	  	
	  
	  	for (i=size;i>=0;i--)
	  	  {
	  	  	sel_ar = (ActionRequest)requests.elementAt(requests.size()-1);
	  	  	
	  	  	requests.removeElement((ActionRequest)sel_ar);
	  	  	     
	  	  	sel_ar.aid = String.valueOf(id);
	  	    id--;
	  	  	     
	  	    sel_ar.startTime = startTime[id];
	  	  	    	  	  	     
	  	  	sortedRequests.insertElementAt((ActionRequest)sel_ar,0);	  	  	
	  	  }
	  	
	  	Vector sortedRequests_backup = new Vector(sortedRequests.size());
	  	
	    id = 1;
	    for (i=sortedRequests.size()-1;i>=0;i--)	
	      {
	      	ar = (ActionRequest)sortedRequests.elementAt(i);
	      	
	      	ar.aid = String.valueOf(id);	      	
	      	
	      	ar.startTime = startTime[id-1];
	      	
	      	id ++;
	      	
	      	sortedRequests_backup.addElement(ar);
	      }
	      
	    sortedRequests = sortedRequests_backup;
	  	
	  	return sortedRequests;
	  }*/
	  
	  
	/** Generate a set of real numbers in (0,1) following zipfian distribution whose sum is equal to 1. */
	public void generateZipfianNums ()
	  {
	  	zipf_nums[0] = Math.log(Math.E)/Math.log(1.78*AortaSchedulingSimulator.deviceNum);
	  	
	  	for (int i=1;i<AortaSchedulingSimulator.deviceNum;i++)
	  	  zipf_nums[i] = zipf_nums[0]/(i+1);
	  }	
	  
    
    /** Get start time of next action request. */
    public double getNextStartTime()
      {
      	if (AortaSchedulingSimulator.requestArrivalPattern == AortaSchedulingSimulator.POISSON)
      	  return getNextStartTimePoisson();
      	else
      	  return getNextStartTimeConstant();
      }
      
      
    /** Get start time of next action request with a poisson arrival pattern. */
    public double getNextStartTimePoisson()
      {
      	double currentRequestStartTime = requestStartTime;
      	
      	requestStartTime += -Math.log(startTimeRan.nextDouble())*AortaSchedulingSimulator.requestArrivalRate;
      	
      	return currentRequestStartTime;    	
      }
      
      
    /** Get start time of next action request with a constant arrival pattern. */
    public double getNextStartTimeConstant()
      {
      	double currentRequestStartTime = requestStartTime;
      	
      	requestStartTime += AortaSchedulingSimulator.requestArrivalRate;
      	
      	return currentRequestStartTime;    	
      }   	  
    
    
    /** Randomly generate the deadline of an action request. */
    public double ranDeadline()
 	  {
	  	if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
	  	  if (W == 1)
	  	    return MIN_PHOTO_DEADLINE;
	  	  else
	  	    return MIN_PHOTO_DEADLINE + Math.abs(deadlineRan.nextInt()) % (MAX_PHOTO_DEADLINE - MIN_PHOTO_DEADLINE);
	  	else
	  	  if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.WARN)
	  	    if (W == 1)
	  	        return MIN_WARN_DEADLINE;
	  	      else
	  	        return MIN_WARN_DEADLINE + Math.abs(deadlineRan.nextInt()) % (MAX_WARN_DEADLINE - MIN_WARN_DEADLINE);
	  	  else
	  	    {
	  	      System.err.println("Unsupported action to be scheduled. Must be \"photo\" or \"warn\".");
              
              System.exit(1);
        
              return Double.MAX_VALUE;
            }
	  } 
	     
    
    /** Randomly generate the cost of an action request. */
    public double ranActionCost(ActionRequest ar)
      {
      	if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
      	  return ranPhotoCost(ar);
      	else
      	  if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.WARN)
      	    return ranWarnCost(ar);
      	  else
      	    {
      	      System.err.println("Unsupported action to be scheduled. Must be \"photo\" or \"warn\".");
              
              System.exit(1);              
            }
            
        return Double.MAX_VALUE;
      }
      
    
    /** Randomly generate the cost of a {@link Photo} action request. */
    public double ranPhotoCost(ActionRequest ar)
     {
       CameraSimulator camSim = new CameraSimulator("0");
              
       int pan = ranPan();
        
       int tilt = ranTilt();
       
       int zoom = ranZoom();       
                   
       Vector v = new Vector(10,5);
       
       v.addElement(String.valueOf(pan));
       
       v.addElement(String.valueOf(tilt));
       
       v.addElement(String.valueOf(zoom));
       
       Tuple t = new Tuple(v);
       
       ar.statusVector.addElement((Tuple)t);
       
       double headMovementCost = camSim.computeHeadMovementCost(pan, tilt, zoom);
       
       return 2*camSim.STANDARD_POLLING_TIME + headMovementCost + camSim.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST; 
     }	
     
     
    /** Randomly pick a target <code>pan</code> for the camera head. */
    public int ranPan()
	  {
	  	return CameraSimulator.MIN_PAN + Math.abs(ran.nextInt()) % (CameraSimulator.MAX_PAN - CameraSimulator.MIN_PAN + 1);
	  }
	  
	  
	/** Randomly pick a target <code>tilt</code> for the camera head. */
    public int ranTilt()
 	  {
	  	return CameraSimulator.MIN_TILT + Math.abs(ran.nextInt()) % (CameraSimulator.MAX_TILT - CameraSimulator.MIN_TILT + 1);
	  }
	  
	  
    /** Randomly pick a target <code>zoom</code> for the camera head. */
    public int ranZoom()
	  {
	  	Integer num = new Integer(Math.abs(ran.nextInt()) % 4);
	  	
	  	/*if (num > 1) 
	  	  num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num > 2)
	  	  num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num == 0) return 1;
	  	else if (num == 1) return 2501; 
	         else if (num == 2) return 5001;
	              else return 9999;*/
	              
	    while (usedZoom.contains(num))
	      num = new Integer(Math.abs(ran.nextInt()) % 4);	    
	    
	    usedZoom.addElement((Integer)num);
	    
	    if (usedZoom.size() == 4)
	      usedZoom.removeAllElements();	    
	    
	    if ( num.intValue() == 0)
	      return 1;
	    else if (num.intValue() == 1)
	           return 2501;
	         else if (num.intValue() == 2)
	                return 5001;
	              else
	                return 9999;	      	
	  }	 
	  
	  
   /** Randomly generate the cost of a {@link Warn} action request. */
    public double ranWarnCost(ActionRequest ar)
     {
       Point loc = ranLoc();
                         
       Vector v = new Vector(10,5);
       
       v.addElement(String.valueOf(loc.x));
       
       v.addElement(String.valueOf(loc.y));
       
       v.addElement(String.valueOf(loc.z));  
            
       Tuple t = new Tuple(v);
       
       ar.statusVector.addElement((Tuple)t);
       
       double actionCost = 0;
       
       actionCost += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;
       
       actionCost += Math.sqrt(Math.pow(loc.x-ER1Simulator.HOME_LOC_X,2)+Math.pow(loc.y-ER1Simulator.HOME_LOC_Y,2))*100*ER1Simulator.MOVEPERCM_ESTIMATED_COST;
       
       double currentX = ER1Simulator.HOME_LOC_X, currentY = ER1Simulator.HOME_LOC_Y;
       double targetX = loc.x, targetY = loc.y;  	    
       double deltaAngle; 
       
       if ((currentX == targetX) && (currentY == targetY))
	      deltaAngle = 0; 
	   else
	      {
	     	if (currentX == targetX)
	     	  if (currentY < targetY)
	     	    deltaAngle = 90;
	     	  else
	     	    deltaAngle = -90;
	     	else
	     	  {
	     	  	 deltaAngle = Math.atan((targetY - currentY)/(targetX-currentX))/Math.PI*180;
	     	  	  	
	     	  	 if ((deltaAngle > 0) && (targetY < currentY))
	     	  	   deltaAngle -= 180;
	     	  	  	
	     	  	 if ((deltaAngle < 0) && (targetY > currentY))
	     	  	   deltaAngle += 180;
	     	  }     	    	
	      }
	      	
	   actionCost += Math.abs(ER1Simulator.ROTATEPERDEG_ESTIMATED_COST*deltaAngle);
	   
       actionCost += ER1Simulator.DEFAULT_SOUND_COST;
       
       return actionCost; 
     }	
     
     
   /** Randomly pick a target location for the robot. */
   public Point ranLoc()
     {
       Point loc = new Point();
       
       loc.x = ER1Simulator.MIN_LOC_X + Math.abs(ran.nextInt()) % (ER1Simulator.MAX_LOC_X - ER1Simulator.MIN_LOC_X + 0.01);
       
       loc.y = ER1Simulator.MIN_LOC_Y + Math.abs(ran.nextInt()) % (ER1Simulator.MAX_LOC_Y - ER1Simulator.MIN_LOC_Y + 0.01);
       
       if (loc.x > ER1Simulator.MAX_LOC_X)
         loc.x = ER1Simulator.MAX_LOC_X;
       
       if (loc.y > ER1Simulator.MAX_LOC_Y)
         loc.y = ER1Simulator.MAX_LOC_Y;
       
       loc.z = ER1Simulator.HOME_LOC_Z;
       
       return loc;	
     }
     
     
   /*public void printConstants()
      {
      	System.out.println("Minimum photo() cost = " + MIN_PHOTO_COST);
        System.out.println("Maximum photo() cost = " + MAX_PHOTO_COST);
        System.out.println("Average photo() cost = " + AVG_PHOTO_COST);
        System.out.println();
        
        System.out.println("Minimum warn() cost = " + MIN_WARN_COST);
        System.out.println("Maximum warn() cost = " + MAX_WARN_COST);
        System.out.println("Average warn() cost = " + AVG_WARN_COST);
        System.out.println();
        
      	System.out.println("Minimum photo() deadline = " + MIN_PHOTO_DEADLINE);
        System.out.println("Maximum photo() deadline = " + MAX_PHOTO_DEADLINE);
        System.out.println();
        
        System.out.println("Minimum warn() deadline = " + MIN_WARN_DEADLINE);
        System.out.println("Maximum warn() deadline = " + MAX_WARN_DEADLINE);
        System.out.println();
        
        System.exit(0);
      } */   
} 