/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;
import net.tinyos.tinydb.*;  // TinyDB classes for acquiring sensory data from Crossbow motes

/**
 * This class implements the backbone of our Aorta, an action-oriented
 * query processor for pervasive computing.
 *                                                    
 * @author Wenwei Xue (wwxue@cs.ust.hk) 
 * @version 0.4                                           
 */
public class Aorta {		
	
	/* Query type, a query can be any combination of (1), (2), and (3). 
	 * Only action-embedded queries (AQs) are supported in current prototype implementaion of Aorta.
	 */
	public static final byte EVENT_DRIVEN_QUERY = 0x01;              // event-driven continuous query with a system-controlled sample period
	public static final byte TRADITIONAL_CONTINUOUS_QUERY = 0x02;    // traditional continuous query with a user-specified regular sample period
	public static final byte SNAPSHOT_QUERY = 0x04;                  // traditional snapshot query, only be executed once
	
	
	/* Aorta global data structure */
	Action actions [] = new Action [50];
	int actionNum = 0;
	Vector AQVector = new Vector(100,50);
						
	/* Private variables */
	private Action action;
	private Action action2;
	private byte flag;
	private byte queryType;
	private String s [] = new String [200];
	private String actionParams2 [];
		
	private byte AQID  = 0;  // id of the newly registered query, system-provided
	private String AQName;   // name of the newly registered query, user-defined
	
	private boolean actionNesting;  // indicate whether there exists nested (second-level) actions
	
					
	/** Sensor catalog file. */
	public static final String SENSOR_CATALOG = "../UCLayer/device_profiles/catalogs/sensor.xml";
	/** Camera catalog file. */
	public static final String CAMERA_CATALOG = "../UCLayer/device_profiles/catalogs/camera.xml";
    /** Phone catalog file. */
    public static final String PHONE_CATALOG = "../UCLayer/device_profiles/catalogs/phone.xml";
    /** Robot catalog file. */
    public static final String ROBOT_CATALOG = "../UCLayer/device_profiles/catalogs/robot.xml";
       
      
    /** Sensor atomic operation configuration file. */ 
    public static final String SENSOR_AOC_FILE = "../UCLayer/device_profiles/atomic_operation_configuration_files/sensor.xml";
	/** Camera atomic operation configuration file. */
	public static final String CAMERA_AOC_FILE = "../UCLayer/device_profiles/atomic_operation_configuration_files/camera.xml";
    /** Phone atomic operation configuration file. */
    public static final String PHONE_AOC_FILE = "../UCLayer/device_profiles/atomic_operation_configuration_files/phone.xml";
    /** Robot atomic operation configuration file. */
    public static final String ROBOT_AOC_FILE = "../UCLayer/device_profiles/atomic_operation_configuration_files/robot.xml";
    
    /** Sensor base station, implemented as an XML file. */ 
    public static final String SENSOR_BASESTATION = "../UCLayer/base_stations/sensor.xml";
	/** Camera base station, implemented as an XML file. */
	public static final String CAMERA_BASESTATION = "../UCLayer/base_stations/camera.xml";
	/** Phone base station, implemented as an XML file. */
    public static final String PHONE_BASESTATION = "../UCLayer/base_stations/phone.xml";
    /** Robot base station, implemented as an XML file. */
    public static final String ROBOT_BASESTATION = "../UCLayer/base_stations/robot.xml";
               
    
    /* Name of the sensor virtual table. */
	public String SENSOR_TABLE;
	/* Name of the camera virtual table. */
	public String CAMERA_TABLE;
	/* Name of the phone virtual table. */
	public String PHONE_TABLE;
	/* Name of the robot virtual table. */
	public String ROBOT_TABLE;
	
	
	/** Shorthands for the sensor virtual table */
	public static final String SENSOR_SHORT = "s.";
	/** Shorthands for the camera virtual table. */
	public static final String CAMERA_SHORT = "c.";
	/** Shorthands for the phone virtual table. */
	public static final String PHONE_SHORT = "p.";
	/** Shorthands for the robot virtual table. */
	public static final String ROBOT_SHORT = "r";
	
	
    /** Default value of the system-controlled sample period (in milliseconds) for triggering devices. */
    public static final int DEFAULT_SAMPLE_PERIOD = 1024;  
    
    
    /* Special string constants added when parsing AQs */
    public static final String DELIMITER_STRING = "@@";
    public static final String NULL_STRING = "";        
    

    /** Indicate whether the optimizer should seek for multi-query optimization. */
    public static boolean MULTI_QUERY_OPTIMIZATION = false;
    /** System-defined scheduling (grouping) time interval. */
    long seqNo = 0;
    
    /* Flags for the scheduling heuristic adopted */
    public static final byte WEIGHTED_SUM = 0x01;
    public static final byte MULTIPLICATION = 0x02;
    public static final byte PRIORITY = 0x03;   
    static byte schedulingAlgorithm = WEIGHTED_SUM;  // default value
    
    
    /** Indicate whether Aorta should use real sensors or the sensor simulator.  */             	
	public static boolean SENSOR_SIMULATION;
	
	/** Indicate whether Aorta should use real cameras or the camera simulator. */
    public static boolean CAMERA_SIMULATION;
    
    /** Indicate whether Aorta should use real phones or the phone emulator. */
    public static boolean PHONE_EMULATION;	  
    
    /** Indicate whether Aorta should use real robots or the robot simulator. */
    public static boolean ROBOT_SIMULATION;	 
    
    /** System-controlled retry number when failing to polling a device, executing an action, and so on. */
    public static int RETRY_NUM = 3;
    
    /* Constants and variables for integrating with the J2ME phone emulator */
    public static final String MESSAGE_FILE = "../messages/message.txt";  // initial warning message
    public static final String PHOTO_FILE = "C:\\WTK21\\apps\\AortaPhotoAlbum\\res\\New\\photo.jpg";  // the photo for display in the simulator
      
    long messageId = 1000000000000000L;  // 16 bit
    int serverPort = 80;  // default web server port
	boolean firstTime = true;
	WebServer ws;

	
	/* Data structure for the locking mechanism implementation */
	public static final boolean lockingEnabled = true;  // indicate whether the locking mechanism is enabled
	
	public Vector lockCameraVector = new Vector (10,5);
	public Vector lockPhoneVector = new Vector (10,5);
	public Vector lockSensorVector = new Vector (10,5);
	public Vector lockRobotVector = new Vector (10,5);
	
	
	/**
	 * Constructor of the Aorta query processor.
	 * <p>
	 * It has the following two roles:
	 * <ul>
	 * <li> Initialize the system built-in actions.
	 * <li> Get the name of the virtual device tables from the device catalogs.
	 * </ul>
	 */
	public Aorta () 
	  {
	  	initializeSystemBuiltInActions();
	  	
	  	/* Get the names of virtual device tables from device catalogs */
	  	getVirtualDeviceTableNames();	  	
	  }
		
		
	private void initializeSystemBuiltInActions ()
	  {	   
	     /* In the current implementation of Aorta, we only support three system built-in actions: photo(), sendmsg() and sendphoto() */
	     actions [0] = new Photo (this) ;
	     actions [1] = new SendMsg(this);
	     actions [2] = new SendPhoto(this);
	     actions [3] = new Warn(this);
	     
	     actionNum = 4;
	  }
	
	
	private void getVirtualDeviceTableNames ()
	{
		BufferedReader catalogReader; 
		String line;
		int start,stop;
		
		try {
		catalogReader =  new BufferedReader (new FileReader(SENSOR_CATALOG));
		while ( (line = catalogReader.readLine()) != null)
	      {
	      	if ( (start = line.indexOf("<virtualTableName>")) != -1)
	      	  {
	      	  	start += "<virtualTableName>".length();
	      	  	if ( (stop = line.indexOf("</virtualTableName>",start)) != -1)
	      	  	  SENSOR_TABLE = line.substring(start,stop).trim();	      	  	  
	      	  }
	      }	      
	    catalogReader.close();
		
		catalogReader =  new BufferedReader (new FileReader(CAMERA_CATALOG));
		while ( (line = catalogReader.readLine()) != null)
	      {
	      	if ( (start = line.indexOf("<virtualTableName>")) != -1)
	      	  {
	      	  	start += "<virtualTableName>".length();
	      	  	if ( (stop = line.indexOf("</virtualTableName>",start)) != -1)
	      	  	  CAMERA_TABLE = line.substring(start,stop).trim();	      	  	  
	      	  }
	      }	      
	    catalogReader.close();
	    
	    catalogReader =  new BufferedReader (new FileReader(PHONE_CATALOG));
		while ( (line = catalogReader.readLine()) != null)
	      {
	      	if ( (start = line.indexOf("<virtualTableName>")) != -1)
	      	  {
	      	  	start += "<virtualTableName>".length();
	      	  	if ( (stop = line.indexOf("</virtualTableName>",start)) != -1)
	      	  	  PHONE_TABLE = line.substring(start,stop).trim();	      	  	  
	      	  }
	      }	      
	    catalogReader.close();
	    
	    catalogReader =  new BufferedReader (new FileReader(ROBOT_CATALOG));
		while ( (line = catalogReader.readLine()) != null)
	      {
	      	if ( (start = line.indexOf("<virtualTableName>")) != -1)
	      	  {
	      	  	start += "<virtualTableName>".length();
	      	  	if ( (stop = line.indexOf("</virtualTableName>",start)) != -1)
	      	  	  ROBOT_TABLE = line.substring(start,stop).trim();	      	  	  
	      	  }
	      }	      
	    catalogReader.close();	    
	   }
	   catch (Exception e) {e.printStackTrace();}	   
	}
	
	
	/** 
	 * Install a new action-embedded query (AQ).
	 *
	 * @param AQString the SQL text of an action-embedded query
	 * 
	 * @see AQ
	 * @see #parseAQ
	 **/
	public void addAQ (String AQString) {
		
	   // initialization
	   action = null;
	   action2 = null;
	   actionParams2 = null;	   	   
	   queryType = EVENT_DRIVEN_QUERY;
	      
	   int i = 0;
	   while (s[i] != null)
	     s[i++] = null;	
	   for (i=0;i<5;i++)
	     s[i] = NULL_STRING;     
	   	   
	   boolean success = parseAQ(AQString);	   
	   if (!success) return;
	   else System.out.println("Successfully install Query " + (AQID+1) + ".\n");  // the query id in Aorta starts from 1 not 0
	   
	   AQ newAQ = new AQ (s, AQString, ++AQID, action, action2, actionParams2, AQName, flag, queryType, actionNesting, this);
	   AQVector.addElement((AQ)newAQ); 	   		
	}
	
		
	/**
	 * A prelimenary query parser of Aorta. 
	 * <p>
	 * The format of the outputted (parsed) string array s[] is as follows:
	 * <ul>
	 * <li> s[0] is the sensor query string for TinyDB.
	 * <li> s[1] and s[2] are the join attributes of two virtual relations. s[1] also include the join function name, operator and constant value.
	 * <li> s[3] and s[4] are name and value of the selection attribute.
	 * <li> The action parameters begin at s[5], followed by the name and parameters of the timer functions/variables in the WHERE,START,and STOP clauses.
	 *      The system-defined delimeter string {@link #DELIMITER_STRING} seperates the boundary of these variable-length items. 
	 * <li> Finally, there are the value and time unit in the LIFETIME clause. 
	 * <li> Any item above that is not involved in the AQ will be assigned with a system-defined null (empty) string {@link #NULL_STRING} (not the null pointer in Java).
	 * </ul>
	 *
	 * @param AQString the SQL text of an action-embedded query
	 *
	 * @return <code>true</code> if the parsing if successful;
	 *         <code>false</code> otherwise         
     */		
	public boolean parseAQ (String AQString)
          {
            
            int start,stop,end;
            int i;
            String str;
            actionNesting = false;
                                
            
            // CREATE keyword
            start = AQString.indexOf("create ");
            if (start == -1)
              {
              	System.err.println("Invalid query syntax: CREATE keyword is missing.");
              	return false;
              }
            start += 6;
            
                        
            // AQ keyword
            start = AQString.indexOf(" aq ",start);
            if (start == -1)
             {
               System.err.println("Invalid query syntax: AQ keyword is missing.");
               return false;
             }
            start += 3;
            
            
            // AS keyword and Query name
            stop = AQString.indexOf(" as ",start);
            if (stop == -1)
              {
              	System.err.println("Invalid query syntax: AS keyword is missing.");
              	return false;
              }
            AQName = AQString.substring(start,stop).trim();
            for (i=0;i<AQVector.size();i++)
              if (AQName.equals(((AQ)AQVector.elementAt(i)).name))
                {
                	System.err.println("An action-embedded query with the same name has already been registered. Please rename the new query or remove the old one first.");
                	return false;
                }        
           
            
            // SELECT Keyword and action name
            start = AQString.indexOf(" select ");
            if (start == -1)
              {
              	System.err.println("Invalid query syntax: SELECT clause is missing.");
              	return false;
              }
            start += 7; 
            stop = AQString.indexOf("(",start);
            str = AQString.substring(start,stop).trim();
            i=0; 
            while (actions[i] != null)
              { 
                if (str.equals(actions[i].name) )
                  {
                    action = actions[i];
                    break;
                  }
                i++;
              }
            if (action == null)
              {
              	System.err.println("Unrecognized action: " + str + "().");
              	return false;
              }    
            
                
            // FROM keyword and input parameters of the first-level action
            start = stop + 1;
            stop = AQString.indexOf(" from ");
            if (stop == -1)
             {
              	System.err.println("Invalid query syntax: FROM clause is missing.");
              	return false;
              }             
            stop ++;
            end = AQString.indexOf(")",start);
            while ( ( (i=AQString.indexOf(")",end+1)) != -1) && (i < stop)) 
              end = i;
            
            /* end is the ")" of the first-level action. Aorta only supports one level of action nesting now. */ 
            i = 5;  // action parameters begins from s[7]
            while (true)
              {
              	stop = AQString.indexOf(",",start);
              	if ( (stop == -1) || (stop > end))
              	  { 
              	    str = AQString.substring(start,end).trim();
              	    s[i] = str;
              	    if (s[i].charAt(0) == '\"' && s[i].charAt(s[i].length()-1) == '\"')
              	      s[i] = s[i].substring(1,s[i].length()-1);
              	    i++;
              	    break;
              	  }
              	str = AQString.substring(start,stop).trim();
              	s[i] = str;  	// e.g., s[i] = "p.no"
              	if (s[i].charAt(0) == '\"' && s[i].charAt(s[i].length()-1) == '\"')
              	  s[i] = s[i].substring(1,s[i].length()-1);
              
              	/* The input is the output of a second-level action, e.g., photo(c.ip, "photos/") */
              	if ( (s[i].indexOf("(")) != -1 )
              	  {
              	    actionNesting = true;
              	    start = AQString.indexOf(")",stop) + 1;
              	    s[i] = s[i] + AQString.substring(stop, start).trim();
              	    stop = start;
              	    
              	    /* Extract out the query constraint related to the second-level action */              	       
              	    int t1, t2;
              	    t1 = AQString.indexOf(CAMERA_SHORT,start);
              	    if (t1 == -1)
                      {
              	        System.err.println("Invalid query syntax: prefix \"" + CAMERA_SHORT + "\" for camera virtual table is required in the current version of AortaSQL.");
              	        return false;
                      }  
              	    t2 = Math.max(AQString.lastIndexOf(" and ",t1), AQString.lastIndexOf(" or ",t1));
              	    if (t2 == -1) 
              	      t2 = AQString.lastIndexOf(" where ",t1);
              	    if (t2 == -1)
              	      {
              	      	System.err.println("Invalid query syntax in the WHERE clause. Fail to parse the second-level action.");
              	      	return false;
              	      }
              	    t1 = t2 + 1;
              	    if (AQString.charAt(t1) == 'a') t1 += 3;  // AND
              	    else if (AQString.charAt(t1) == 'o') t1 += 2;  // OR
              	         else t1 += 5;   // WHERE
              	         
              	    /* t1 is the beginning of the constraint */
              	    t2 = AQString.indexOf(" or ",t1);
              	    int t;
              	    if ( (t2 == -1) || (t = AQString.indexOf(" and ",t1)) != -1 && t < t2)
              	      t2 = AQString.indexOf(" and ",t1);
              	    if (t2 == -1)
              	      t2 = AQString.indexOf(" start ",t1);
              	    if (t2 == -1)
              	      t2 = AQString.indexOf(" stop ",t1);
              	    if (t2 == -1)
              	      t2 = AQString.indexOf(" lifetime ",t1);
              	    if (t2 == -1)
              	      t2 = AQString.length() - 1;
              	    String ts = AQString.substring(t1,t2);
              	    ts = ts.trim();
              	    
              	    AQString = AQString.substring(0,t1) + AQString.substring(t2);
              	    if ( (t2 = AQString.lastIndexOf(" and ", t1)) != -1)
              	      AQString = AQString.substring(0,t2) + AQString.substring(t2 + 4);
              	    else
              	      AQString = AQString.replaceFirst(" and "," ");
              	      
              	    /* Remove the second-level action and the corresponding query constraint */         	    
              	    t1 = ts.indexOf("(");  // extract function name
              	    s[1] += ";" + ts.substring(0,t1).trim();
              	               	    
              	    t1 = ts.indexOf(SENSOR_SHORT);  // extract the attribute for the triggering device (sensor)
              	    if (t1 == -1)
                     {
              	       System.err.println("Invalid query syntax: prefix \"" + SENSOR_SHORT + "\" for sensor virtual table is required in current version of AortaSQL.");
              	       return false;
                     } 
                    t2 = ts.indexOf(",",t1);
              	    if (t2 == -1)
              	      t2 = ts.indexOf(")",t1);
              	    if (t2 == -1)
              	      {
              	      	System.err.println("Invalid query syntax in the WHERE clause. Fail to parse the embedded action.");
              	      	return false;
              	      }
              	    s[1]= ts.substring(t1,t2).trim() + s[1];
              	    
              	    t1 = ts.indexOf(CAMERA_SHORT);  // extract the attribute for the triggered device (camera)
              	    if (t1 == -1)
                     {
              	       System.err.println("Invalid query syntax: prefix \"" + CAMERA_SHORT + "\" for camera virtual table is required in current version of AortaSQL.");
              	       return false;
                     }  
              	    t2 = ts.indexOf(",",t1);
              	    if (t2 == -1)
              	      t2 = ts.indexOf(")", t1);
              	    if (t2 == -1)
              	      {
              	      	System.err.println("Invalid query syntax in the WHERE clause. Fail to parse the second-level action.");
              	      	return false;
              	      }
              	    s[2]+= ts.substring(t1,t2).trim();              	    
              	    
              	    t1 = ts.indexOf(">", t2);  // extract the operator
              	    if (t1 == -1) 
              	      t1 = ts.indexOf("<", t2);
              	    if (t1 == -1)
              	      t1 = ts.indexOf("=",t2);
              	    if (t1 == -1)
              	      t1 = ts.indexOf("!=", t2);
              	    if (t1 == -1)
              	      {
              	      	System.err.println("Invalid arithmetic operator in the WHERE clause. Fail to parse the second-level action.");
              	      	return false;
              	      }
              	    t2 = ts.indexOf(" ",t1);
              	    if (t2 == -1)
              	      {
              	      	System.err.println("Invalid arithmetic operator in the WHERE clause. Fail to parse the second-level action.");
              	      	return false;
              	      }
              	    String s_t = ts.substring(t1,t2).trim();
              	    if ( s_t.equals(">") || s_t.equals(">=") || s_t.equals("<") || s_t.equals("<=") || s_t.equals("==") || s_t.equals("!=") || s_t.equals("<>") )
              	      s[1] += ";" + s_t;
                    else 
                      {
              	      	System.err.println("Invalid arithmetic operator in the WHERE clause. Fail to parse the second-level action.");
              	      	return false;
              	      }              	      
              	    
              	    while (ts.charAt(t2) == ' ') t2++;  // extract the constant value
              	    t1 = ts.indexOf(" ", t2);
              	    if (t1 == -1)
              	      t1 = ts.length();
              	    s_t = ts.substring(t2,t1).trim();
              	    int len = s_t.length();
              	    if (s_t.charAt(len-1) == ')')
              	      s_t = s_t.substring(0,len-1);
              	    try 
              	      {
              	      	start = (int)Double.parseDouble(s_t);
              	      } 
              	    catch (Exception e) 
              	      {
              	      	System.err.println("Invalid constant value for the distance function in the WHERE clause. Fail to parse the second-level action."); 
              	      	return false;
              	      }
              	    s[1] += ";" + s_t;              	                	                 	                  	                 	    
              	  } 
              	                 
              	i++;
              	if (stop == end) break;
              	start = stop + 1;
              	
              }
                       
            
            /* Decide the type of the query */
            stop = AQString.indexOf(TimerFunctions.EVERY,start);
            if (stop != -1)
              {
              	stop = AQString.lastIndexOf(" sensor ss ",stop);
              	if (stop == -1)
              	  queryType = TRADITIONAL_CONTINUOUS_QUERY;
              }
           	
           	stop = AQString.indexOf("lifetime",start);
            if (stop != -1)
              stop = AQString.indexOf(TimerFunctions.$ONCE,stop);
            if (stop != -1)
              queryType = (byte)(queryType | SNAPSHOT_QUERY);
             
            if (queryType == TRADITIONAL_CONTINUOUS_QUERY)
              {
              	stop = AQString.indexOf(" or ");
              	if (stop != -1)
              	  queryType = TRADITIONAL_CONTINUOUS_QUERY | EVENT_DRIVEN_QUERY;
              }
              
            if ( (queryType & SNAPSHOT_QUERY) != 0)
              {
              	
              	start = AQString.indexOf(" where ");
              	if (start == -1)
              	  {
              	  	queryType = SNAPSHOT_QUERY | TRADITIONAL_CONTINUOUS_QUERY;
              	  	start = AQString.indexOf(" lifetime ");
              	  	AQString = AQString.substring(0,start) + " where s.null = null " + AQString.substring(start);
                  }
                else
                  {
                  	stop = AQString.indexOf(PHONE_SHORT,start);
                  	if (stop == -1)
                  	  {
                  	  	stop = AQString.indexOf("=",start);
                  	  	if (stop != -1)
                  	  	  queryType = TRADITIONAL_CONTINUOUS_QUERY | SNAPSHOT_QUERY;
                  	  }
                  }
              }
            
            
            // WHERE keyword
            start = AQString.indexOf(" where ",end);
            if (start == -1)
              {
              	System.err.println("Invalid query syntax: WHERE clause is missing.");
              	return false;
              } 
            
  
            /* Decide the class of the query: selection or join.
               Selection: select some particular devices to take the action, the devices picked are usually the same for each action invocation, e.g., p.owner = "admin".
               Join: join the triggering device table with the triggered device table, the devices picked are usually different for each action invocation, e.g., dist(s.loc, c.loc) < 10. */
            start = AQString.indexOf(CAMERA_SHORT,start);
            if (start == -1)
              flag = AQ.SELECTION_FLAG;  // selection query
            else flag = AQ.JOIN_FLAG;  // join query            
            
            // FROM keyword
            start = AQString.indexOf(" from ",end);
            if (start == -1)
              {
              	System.err.println("Invalid query syntax: FROM clause is missing.");
              	return false;
              }           
            
                                  
            /* Use the Java API provided by TinyDB to acquire attributes from the sensor virtual table */
            start = AQString.indexOf(" from ");
            start = AQString.indexOf("robot ",start);
            if (start != -1)
              AQString += " and r.id = $any ";            
            
            start = AQString.indexOf(" where ", end);
            start += 6; 
            if ( (queryType & EVENT_DRIVEN_QUERY) != 0)
              {
                stop = AQString.indexOf(" and ",start);
                if (stop == -1)
                  {
              	    System.err.println("Invalid query syntax: AND Keyword is missing.");
              	    return false;
                  }
            
                end = AQString.lastIndexOf(SENSOR_SHORT,stop);
                if (end == -1) 
                  {
              	    end = AQString.lastIndexOf(SENSOR_TABLE,stop);
              	    if (end < start) end = -1;
                  }
                if (end != -1)
                  { 
              	    int j = AQString.lastIndexOf(CAMERA_SHORT,stop);
              	    if ( j < start || j== -1 ) 
              	      {
              	  	    j = AQString.lastIndexOf(" or ", end);
              	      	if (j != -1) start = j+3;
              	     	str = AQString.substring(start,stop);
              	     	j = str.indexOf(" or ");
              	    	if (j != -1)
              	  	      str = str.substring(0,j);
                      }
                     else 
                       end = -1;
                  }
            if (end == -1)  
              {
                end = AQString.indexOf(SENSOR_SHORT,stop);
                if (end == -1)
                  end = AQString.indexOf(SENSOR_TABLE,stop);
                if (end == -1)
                 {
              	   System.err.println("Invalid query syntax: prefix \"" + SENSOR_SHORT + "\" for sensor virtual table is required in current version of AortaSQL.");
              	   return false;
                 }  
                stop += 4;
                end = AQString.indexOf(" or ",stop);
                if (end == -1)
                  end = AQString.indexOf(" and ",stop);
                if (end == -1)
                  end = AQString.indexOf(" start ",stop);
                if (end == -1)
                  end = AQString.indexOf(" stop ",stop);
                if (end == -1)
                  end = AQString.indexOf(" lifetime ",stop);
                if (end == -1) 
                  end = AQString.length() -1;
                str = AQString.substring(stop,end);	
              }
            str = str.trim();
            str = str.replace('.','@');
            str = str.replaceAll(SENSOR_SHORT.substring(0,SENSOR_SHORT.length()-1) + "@","");
            
            /* str is the query constraint on sensor virtual table in the WHERE clause */
            start = str.indexOf(SENSOR_TABLE);           
            
            if (start == -1)  // non-aggregate query
              {
              	str = str.replace('(', ' ');
                str = str.replace(')', ' ');                
                s[0] = "select nodeid from " + SENSOR_TABLE + "s" + " where " + str + " sample period " + String.valueOf(DEFAULT_SAMPLE_PERIOD);
              }
            else //  a nested aggregate query
              {
              	String samplePeriod;
              	
              	start = str.indexOf("(");
              	str += " ";
              	start += 1;
              	stop = str.indexOf(")") + 1;
               	s[0] = str.substring(start,stop).trim() + " from " + SENSOR_TABLE + " sample period ";
               	start = str.indexOf(TimerFunctions.EVERY);
               	if (start == -1)
               	  str += (samplePeriod = String.valueOf(DEFAULT_SAMPLE_PERIOD));
               	else
               	  {
               	  	start += TimerFunctions.EVERY.length();
               	  	start = str.indexOf("(",start);
               	  	start ++;
               	  	stop = str.indexOf(",",start);
               	  	if (stop == -1)
               	  	  {
               	  	  	System.err.println("Invalid timer function syntax in the nested query.");
               	  	  	return false;
               	  	  }
               	  	int value;
               	  	try {value = Integer.parseInt(str.substring(start,stop).trim());}
               	  	catch (Exception e) {System.err.println("Invalid timer function syntax in the nested query."); return false;}
               	  	
               	  	start = stop + 1;
               	  	stop = str.indexOf(")",start);
               	  	if (stop == -1)
               	  	  {
               	  	  	System.err.println("Invalid timer function syntax in the nested query.");
               	  	  	return false;
               	  	  }
               	  	String unit = str.substring(start,stop).trim();
               	  	
               	  	if (unit.equals("second"))
               	  	  value *= 1000;
               	  	else if (unit.equals("minute"))
               	  	       value *= 1000*60;
               	  	     else if (unit.equals("hour"))
               	  	            value *= 1000*60*60;
               	  	          else if (unit.equals("day"))
               	  	                 value *= 1000*60*60*24;
               	  	               else if (unit.equals("week"))
               	  	                      value *= 1000*60*60*24*7;
               	  	                    else if (unit.equals("month"))
               	  	                           value *= 1000*60*60*24*7*30;
               	  	                         else if (unit.equals("year"))
               	  	                                value *= 1000*60*60*24*7*30*12;
               	  	                              else
               	  	                                {
               	  	                                  System.err.println("Invalid timer function syntax in the nested query.");
               	  	  	                              return false;
               	  	  	                            }
               	  	  	                            
               	    s[0] += (samplePeriod = String.valueOf(value));               	  	  
               	  }
                s[0] += ";";
                
                // covert the input format of the temporal aggregate
                start = s[0].indexOf("(");
                if (start == -1)
                  {
                  	System.err.println("Invalid timer function syntax in the nested query.");
               	  	return false;
               	  }
               	start ++;
               	stop = s[0].indexOf(",",start);
               	if (stop == -1)
                  {
                  	System.err.println("Invalid timer function syntax in the nested query.");
               	  	return false;
               	  }
               	String attribute = s[0].substring(start,stop).trim();
               	start = stop + 1;
               	stop = s[0].indexOf(",",start);
               	if (stop == -1)
                  {
                  	System.err.println("Invalid timer function syntax in the nested query.");
               	  	return false;
               	  }
               	int aggWin;
               	try {aggWin = Integer.parseInt(s[0].substring(start,stop).trim());}
               	catch (Exception e) {System.err.println("Invalid timer function syntax in the nested query."); return false;}
                start = stop + 1;
                stop = s[0].indexOf(",",start);
               	if (stop == -1)
                  {
                  	System.err.println("Invalid timer function syntax in the nested query.");
               	  	return false;
               	  }
               	int slideWin;
               	try {slideWin = Integer.parseInt(s[0].substring(start,stop).trim());}
               	catch (Exception e) {System.err.println("Invalid timer function syntax in the nested query."); return false;}
                start = stop + 1;
                stop = s[0].indexOf(")",start);
               	if (stop == -1)
                  {
                  	System.err.println("Invalid timer function syntax in the nested query.");
               	  	return false;
               	  }
               	String unit = s[0].substring(start, stop).trim();
               	if (unit.equals("second"))
               	  	  aggWin *= 1000;
               	  	else if (unit.equals("minute"))
               	  	       aggWin *= 1000*60;
               	  	     else if (unit.equals("hour"))
               	  	            aggWin *= 1000*60*60;
               	  	          else if (unit.equals("day"))
               	  	                 aggWin *= 1000*60*60*24;
               	  	               else if (unit.equals("week"))
               	  	                      aggWin *= 1000*60*60*24*7;
               	  	                    else if (unit.equals("month"))
               	  	                           aggWin *= 1000*60*60*24*7*30;
               	  	                         else if (unit.equals("year"))
               	  	                                aggWin *= 1000*60*60*24*7*30*12;
               	  	                              else
               	  	                                {
               	  	                                  System.err.println("Invalid timer function syntax in the nested query.");
               	  	  	                              return false;
               	  	  	                            }
                if (unit.equals("second"))
               	  	  slideWin *= 1000;
               	  	else if (unit.equals("minute"))
               	  	       slideWin *= 1000*60;
               	  	     else if (unit.equals("hour"))
               	  	            slideWin *= 1000*60*60;
               	  	          else if (unit.equals("day"))
               	  	                 slideWin *= 1000*60*60*24;
               	  	               else if (unit.equals("week"))
               	  	                      slideWin *= 1000*60*60*24*7;
               	  	                    else if (unit.equals("month"))
               	  	                           slideWin *= 1000*60*60*24*7*30;
               	  	                         else if (unit.equals("year"))
               	  	                                slideWin *= 1000*60*60*24*7*30*12;
               	  	                              else
               	  	                                {
               	  	                                  System.err.println("Invalid timer function syntax in the nested query.");
               	  	  	                              return false;
               	  	  	                            }
                aggWin /= Integer.parseInt(samplePeriod);
                slideWin /= Integer.parseInt(samplePeriod);
                start = s[0].indexOf("(");
                stop = s[0].indexOf(")");
                s[0] = s[0].substring(0,start+1) + String.valueOf(aggWin) + "," + String.valueOf(slideWin) + "," + attribute + s[0].substring(stop);               	              
                               
                start = str.indexOf(">");
                if (start == -1)
                  start = str.indexOf("<");
                if (start == -1)
                  start = str.indexOf("=");
                if (start == -1)
                  start = str.indexOf("!=");
                if (start == -1)
                  {
                  	System.err.println("Invalid arithmetic operator in the WHERE clause.");
              	    return false;
              	  }
              	String aggValue = str.substring(0,start).trim(); 
              	stop = str.indexOf("(");
              	String aggOp = str.substring(start,stop).trim();
              	if (aggOp.equals(">")) 
              	  aggOp = "<";
              	else if (aggOp.equals("<"))
              	       aggOp = ">";
              	     else if (aggOp.equals(">="))
              	            aggOp = "<=";
              	          else if (aggOp.equals("<="))
              	                 aggOp = ">=";
              	               else if (aggOp.equals("==") || aggOp.equals("!=") || aggOp.equals("<>")) 
              	                      {}
              	                    else 
              	                      {
              	                      	System.err.println("Invalid arithmetic operator in the WHERE clause.");
              	                      	return false;
              	                      }
                s[0] += aggOp;
                s[0] += ";";
                s[0] += aggValue;                
                
                /* s[0] = select agg(?) from sensors;aggOp;aggValue  */               
              }            
            } 
            else 
             {
             	s[0] = NULL_STRING; 
             	flag = AQ.SELECTION_FLAG;
             }
            
                         
            /* Parse the other query constraints for other devices */            
            start = AQString.indexOf(" where ") + 1;
            int mark = AQString.indexOf(CAMERA_SHORT, start);
            if (mark == -1)
              mark = AQString.indexOf(PHONE_SHORT, start);
            if (mark == -1)
              mark = AQString.indexOf(SENSOR_SHORT,start);
            if (mark == -1)
              mark = AQString.indexOf(ROBOT_SHORT,start);
            if (mark == -1) 
              {
              	if (queryType == TRADITIONAL_CONTINUOUS_QUERY)
              	  s[3] = s[4] = NULL_STRING;
              	else
              	  {
              	  	System.err.println("Invalid query syntax: prefix \"" + SENSOR_SHORT + "\" for sensor or \"" + CAMERA_SHORT + "\" for camera or \"" + PHONE_SHORT + "\" for phone virtual table is required in current version of AortaSQL.");
                    return false;
                  }
                
              }
            else 
            {
            start = AQString.lastIndexOf(" or ",mark);
            if (start == -1)
              start = AQString.lastIndexOf(" and ", mark);
            if (start == -1)
              start = AQString.lastIndexOf(" where ",mark);
            if (start == -1)
                {
              	  System.err.println("Invalid query syntax in the WHERE clause.");
              	  return false;
                }
            start ++;
            if (AQString.charAt(start) == 'a')
              start += 3;
            else if (AQString.charAt(start) == 'w')
                   start += 5;
                 else if (AQString.charAt(start) == 'o')
                        start += 2;
            mark = start;
                                     
            if ( (flag & AQ.SELECTION_FLAG) != 0) // selection query, currently only for cell phones
              {
              	start = AQString.indexOf(PHONE_SHORT,mark);
              	if (start == -1)
              	  start = AQString.indexOf(ROBOT_SHORT,mark);              	
              	if (start == -1)
              	  {
              	  	start = AQString.indexOf(SENSOR_SHORT,mark);
              	  	if (start == -1)
                      {
              	        System.err.println("Invalid query syntax: prefix \"" + SENSOR_SHORT + "\" for sensor or \"" + CAMERA_SHORT + "\" for camera or \"" + PHONE_SHORT + "\" for phone virtual table is required in current version of AortaSQL.");
                        return false;
                      }                     
              	  	if (AQString.lastIndexOf(SENSOR_SHORT,start-1) != -1)
              	  	  {start = AQString.lastIndexOf(SENSOR_SHORT,start-1); flag |= AQ.JOIN_FLAG;}
              	  	else
              	  	  if (AQString.indexOf(SENSOR_SHORT,start+2) != -1)
              	  	    {start = AQString.indexOf(SENSOR_SHORT,start+2); flag |= AQ.JOIN_FLAG;}
              	  }              	
                stop = start;
                while ( AQString.charAt(stop) != ' ' )
                  stop ++;
                str = AQString.substring(start,stop);
                str = str.trim();                
                s[3] = str;                
                
                start = AQString.indexOf("=",start);
                start ++;
                while ( AQString.charAt(start) == ' ' )
                  start++;
                stop=start; 
                while ( AQString.charAt(stop) != ' ' )
                  stop ++;  
                str = AQString.substring(start,stop);
                str = str.trim();
                if ( (str.charAt(0) == '"') && (str.charAt(str.length() - 1) == '"') )
                  str = str.substring(1,str.length()-1);
                s[4] = str;                                                                                                                                           
              }
              if ( (flag & AQ.JOIN_FLAG) != 0)  // join query, currently only for network cameras
              {
              	start = AQString.indexOf("(", mark);
                s[1] = ";" + AQString.substring(mark,start).trim();                
              	               	    
              	start = AQString.indexOf(SENSOR_SHORT,mark);
                if (start == -1)
                 {
              	   System.err.println("Invalid query syntax: prefix \"" + SENSOR_SHORT + "\" for sensor virtual table is required in current version of AortaSQL.");
              	   return false;
                 }  
                end = AQString.indexOf(" or ",start);
                if (end == -1)
                  end = AQString.indexOf(" and ", start);
                if (end == -1)
                  end = AQString.indexOf(" start ",start);
                if (end == -1)
                  end = AQString.indexOf(" stop ",start);
                if (end == -1)
                  end = AQString.indexOf(" lifetime ",start);
                if (end == -1)
                  end = AQString.length() - 1;
                  
              	stop = AQString.indexOf(",",start);
              	if (stop == -1 || stop > end)
              	  stop = AQString.indexOf(")",start);
              	if (stop == -1)
              	  {
              	    System.err.println("Invalid query syntax in the WHERE clause.");
              	    return false;
              	  }
              	s[1] = AQString.substring(start,stop).trim() + s[1];              	
              	    
                start = AQString.indexOf(CAMERA_SHORT,mark);  
                if (start == -1)
                 {
              	   System.err.println("Invalid query syntax: prefix \"" + CAMERA_SHORT + "\" for camera virtual table is required in current version of AortaSQL.");
              	   return false;
                 }  
              	stop = AQString.indexOf(",",start);
              	if (stop == -1  || stop > end)
              	  stop = AQString.indexOf(")",start);
              	if (stop == -1)
              	  {
              	    System.err.println("Invalid query syntax in the WHERE clause.");
              	    return false;
              	  }
              	s[2] = AQString.substring(start,stop).trim();              	    
              	
              	start = AQString.indexOf(">", stop);
              	if (start == -1 || start > end) 
              	  start = AQString.indexOf("<", stop);
              	if (start == -1  || start > end)
              	  start = AQString.indexOf("=",stop);
              	if (start == -1 || start > end)
              	  start = AQString.indexOf("!=", stop);
              	if (start == -1 || start > end)
              	  {
              	 	System.err.println("Invalid arithmetic operator in the WHERE clause.");
              	   	return false;
              	  }
              	stop = AQString.indexOf(" ",start);
              	String s_t = AQString.substring(start,stop).trim();
              	if ( s_t.equals(">") || s_t.equals(">=") || s_t.equals("<") || s_t.equals("<=") || s_t.equals("==") || s_t.equals("!=") || s_t.equals("<>") )
              	      s[1] += ";" + s_t;
                else 
                  {
              	    System.err.println("Invalid arithmetic operator in the WHERE clause.");
              	    return false;
              	  }              	      
                
                start = stop;
                stop = AQString.indexOf(" or ", start);
                if (stop == -1)
                  stop = AQString.indexOf(" and ", start);
                if (stop == -1)
                  stop = AQString.indexOf(" start ",start);
                if (stop == -1)
                  stop = AQString.indexOf(" stop ",start);
                if (stop == -1)
                  stop = AQString.indexOf(" lifetime ",start);
                if (stop == -1)
                  stop = AQString.length() - 1;
                s_t = AQString.substring(start, stop).trim();
                try 
                  {
                  	start = (int)Double.parseDouble(s_t);
                  } 
              	catch (Exception e) 
              	  {
              	  	System.err.println("Invalid constant value for the join function in the WHERE clause."); 
              	  	return false;
              	  }
              	s[1] += ";" + s_t;
              	/* e.g., s[1] = loc;dist;<;10  s[2] = loc */              	              
              }                               
             } 
            
            if (actionNesting)
              {
              	flag = AQ.SELECTION_FLAG | AQ.JOIN_FLAG;
              	if (!parseNestedAction()) return false;
              }
                                      
            i = 0;
            while (s[i] != null)
              i++; 
            s[i] = DELIMITER_STRING; i++; 
            
            
            /* Parse the timer function query constraint */            
            start = AQString.indexOf(" where ") + 6;
            stop = AQString.indexOf(" start ",start);
            if (stop == -1)
              stop = AQString.indexOf(" stop ",start);
            if (stop == -1)
              stop = AQString.indexOf(" lifetime ",start);
            if (stop == -1)
              stop = AQString.length() - 1;
            str = AQString.substring(start,stop).trim();
            str = " " + str;
            
            start = str.indexOf(" " + TimerFunctions.ATTIME);
            if (start == -1)
              start = str.indexOf(TimerFunctions.EVERY);
            if (start == -1)
              start = str.indexOf(" " + TimerFunctions.INDAYINTERVAL);
            if (start == -1)
              start = str.indexOf(" " + TimerFunctions.INDAYOFWEEKINTERVAL);
            if (start == -1)
              start = str.indexOf(" " + TimerFunctions.INTIMEINTERVAL);
            if (start == -1)
              start = str.indexOf(" " + TimerFunctions.ONDAYOFWEEK);
            if (start == -1)
              start = str.indexOf(" " + TimerFunctions.ONDAY);
            if ( (start != -1) && (str.lastIndexOf("sensor ss ") != -1) )
              start = -1;
            if (start != -1) {
            stop = str.indexOf("(",start);
            if (stop == -1)
              {
              	System.err.println("Invalid timer function syntax in the WHERE clause.");
              	return false;
              }
            s[i] = str.substring(start,stop).trim();
            if ( (s[i].equals(TimerFunctions.EVERY)) || (s[i].equals(TimerFunctions.ATTIME)) || (s[i].equals(TimerFunctions.ONDAY)) || (s[i].equals(TimerFunctions.INTIMEINTERVAL)) || (s[i].equals(TimerFunctions.INDAYINTERVAL)) || (s[i].equals(TimerFunctions.ONDAYOFWEEK)) || (s[i].equals(TimerFunctions.INDAYOFWEEKINTERVAL)) )
              {i++;}
            else 
             {
               System.err.println("Invalid timer function syntax in the WHERE clause.");
               return false;
             }
            start = stop + 1;
            end = str.indexOf(")",start);
            if (end == -1)
              {
              	System.err.println("Invalid timer function syntax in the WHERE clause.");
               	return false;
           	  }
           	if ( s[i-1].equals(TimerFunctions.EVERY) )
           	  {
           	  	stop = str.indexOf(",",start);
           	  	if (stop == -1)
           	  	   {
              	     System.err.println("Invalid timer function syntax in the WHERE clause.");
               	     return false;
           	       }
           	    s[i] = str.substring(start,stop).trim();
           	    int j;
           	    try {j = Integer.parseInt(s[i]);}
               	catch (Exception e) {System.err.println("Invalid timer function syntax in the WHERE clause."); return false;}
               	i++;
               	start = stop + 1;
               	stop = end;
               	s[i]= str.substring(start,stop).trim();
               	j = s[i].length() - 1;
               	if (s[i].charAt(0) == '\"')
               	  s[i] = s[i].substring(1,j);
               	j =0;
                while (TimerFunctions.unitArray[j] != null)
                  {
                  	if (s[i].equals(TimerFunctions.unitArray[j])) break;
                  	j++;
                  }
                
                if (TimerFunctions.unitArray[j] == null)
                  {
                  	System.err.println("Unsupported time unit for the timer function every().");
                   	return false;
                  }
                i++;
                
                s[i++] = DELIMITER_STRING;
           	  }
           	else { // not every()
           	String FuncName = s[i-1];
            stop = str.indexOf(",",start);
            if ( (stop == -1 || stop > end) && (start != end) )
              stop = end;
            while (stop != -1 && stop <= end)
              {
              	s[i] = str.substring(start,stop).trim();
              	if (FuncName.equals(TimerFunctions.ONDAYOFWEEK) || FuncName.equals(TimerFunctions.INDAYOFWEEKINTERVAL) )
              	  {
              	  	int len = s[i].length() - 1;
              	  	if (s[i].charAt(0) == '\"')
              	  	  s[i] = s[i].substring(1,len);
              	  	
              	  	int day_of_week = TimerFunctions.testDayOfWeek(s[i]);
              	  	if (day_of_week == -1)
              	  	 {
               	  	    System.err.println("Invalid timer function syntax in the WHERE clause."); 
               	  	    return false;
               	     }
              	  }
              	else 
              	  {
              	  	if (s[i].equals(TimerFunctions.$ANY))
              	      s[i] = String.valueOf(TimerFunctions.$ANY_VAL);
               	    try 
               	      {
               	      	int j = Integer.parseInt(s[i]);
               	      }
               	    catch (Exception e) 
               	      {
               	  	    System.err.println("Invalid timer function syntax in the WHERE clause."); 
               	  	    return false;
               	      }
               	  } 
               	i++;
               	start = stop + 1;
               	if (start > end) break;
               	stop = str.indexOf(",",start);
               	if (stop == -1) stop = end;
               }
            
            
            try {  // check the validity of the timer function parameters
            if (FuncName.equals(TimerFunctions.ATTIME) || FuncName.equals(TimerFunctions.INTIMEINTERVAL))
              if (Integer.parseInt(s[i-1]) < -1 || Integer.parseInt(s[i-1]) > 59 || Integer.parseInt(s[i-2]) < -1 || Integer.parseInt(s[i-2]) > 59 || Integer.parseInt(s[i-3]) < -1 || Integer.parseInt(s[i-3]) > 23)
                {
                	System.err.println("Invalid parameters for the timer function in the WHERE clause.");
                	return false;
                }
                
            if (FuncName.equals(TimerFunctions.INTIMEINTERVAL))
              if (Integer.parseInt(s[i-4]) < -1 || Integer.parseInt(s[i-4]) > 59 || Integer.parseInt(s[i-5]) < -1 || Integer.parseInt(s[i-5]) > 59 || Integer.parseInt(s[i-6]) < -1 || Integer.parseInt(s[i-6]) > 23
               || Integer.parseInt(s[i-3]) < Integer.parseInt(s[i-6]) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) < Integer.parseInt(s[i-5])) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) == Integer.parseInt(s[i-5]))&&(Integer.parseInt(s[i-1]) <= Integer.parseInt(s[i-4])) 
               || !s[i-8].equals(DELIMITER_STRING))
                {
                  System.err.println("Invalid parameters for the timer function in the WHERE clause.");
                  return false;
                }
                
            if (FuncName.equals(TimerFunctions.ONDAY) || FuncName.equals(TimerFunctions.INDAYINTERVAL))
              if (Integer.parseInt(s[i-1]) < -1 || Integer.parseInt(s[i-1]) > 31 || Integer.parseInt(s[i-2]) < -1 || Integer.parseInt(s[i-2]) > 12 
               || Integer.parseInt(s[i-1]) == 0 || Integer.parseInt(s[i-2]) == 0 || Integer.parseInt(s[i-3]) < 0)
               { 
                 System.err.println("Invalid parameters for the timer function in the WHERE clause.");
                 return false;
               }
                  
            if (FuncName.equals(TimerFunctions.INDAYINTERVAL))
              if (Integer.parseInt(s[i-4]) < -1 || Integer.parseInt(s[i-4]) > 31 || Integer.parseInt(s[i-5]) < -1 || Integer.parseInt(s[i-5]) > 12
               || Integer.parseInt(s[i-4]) == 0 || Integer.parseInt(s[i-5]) == 0 || Integer.parseInt(s[i-6]) < 0
               || Integer.parseInt(s[i-3]) < Integer.parseInt(s[i-6]) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) < Integer.parseInt(s[i-5])) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) == Integer.parseInt(s[i-5]))&&(Integer.parseInt(s[i-1]) <= Integer.parseInt(s[i-4])) 
               || !s[i-8].equals(DELIMITER_STRING) )
               { 
                 System.err.println("Invalid parameters for the timer function in the WHERE clause.");
                 return false;
               }
              
            if (FuncName.equals(TimerFunctions.INDAYOFWEEKINTERVAL))
              if (!s[i-4].equals(DELIMITER_STRING))
               {
                 System.err.println("Invalid parameters for the timer function in the WHERE clause.");
                 return false;
               }
                  
             if (FuncName.equals(TimerFunctions.ATTIME) || FuncName.equals(TimerFunctions.ONDAY) )
               if (!s[i-5].equals(DELIMITER_STRING))
                 {
                    System.err.println("Invalid parameters for the timer function in the WHERE clause.");
                    return false;
                 }
                 
             if (FuncName.equals(TimerFunctions.EVERY))
               if (!s[i-4].equals(DELIMITER_STRING))
                 {
                    System.err.println("Invalid parameters for the timer function in the WHERE clause.");
                    return false;
                 }
                 
             if (FuncName.equals(TimerFunctions.ONDAYOFWEEK))
               if (!s[i-3].equals(DELIMITER_STRING))
                 {
                    System.err.println("Invalid parameters for the timer function in the WHERE clause.");
                    return false;
                 }                 
                 
            } 
            catch (Exception e)
              {
              	System.err.println("Invalid parameters for the timer function in the WHERE clause.");
                return false;
              }
                   
            s[i++] = DELIMITER_STRING; } }
            else 
              {
              	s[i++] = NULL_STRING; 
              	s[i++] = DELIMITER_STRING;
              }
            
            
            /* Parse the start and stop timer clauses, they are both optional */
            
            // START clause
            start = AQString.indexOf(" start ");
            if (start != -1)
              {
              	start += 6;
              	stop = AQString.indexOf(" " + TimerFunctions.$NOW + " ",start);
              	if (stop != -1)
              	  {
              	  	s[i++] = TimerFunctions.$NOW;
              	  	s[i++] = DELIMITER_STRING;
              	  }
              	else {
              	stop = AQString.indexOf("(",start);
              	if (stop == -1)
              	  {
              	  	System.err.println("Invalid timer function syntax in the START clause.");
              	  	return false;
              	  }
              	s[i] = AQString.substring(start,stop).trim();
              	if ( (s[i].equals(TimerFunctions.EVERY)) || (s[i].equals(TimerFunctions.ATTIME)) || (s[i].equals(TimerFunctions.ONDAY)) || (s[i].equals(TimerFunctions.INTIMEINTERVAL)) || (s[i].equals(TimerFunctions.INDAYINTERVAL)) || (s[i].equals(TimerFunctions.ONDAYOFWEEK)) || (s[i].equals(TimerFunctions.INDAYOFWEEKINTERVAL)) )
              	  {i++;}
              	else 
              	  {
              	  	System.err.println("Invalid timer function syntax in the START clause.");
              	  	return false;
              	  }
              	start = stop + 1;
              	stop = AQString.indexOf(",",start);
              	end = AQString.indexOf(")",start);
              	if (end == -1)
              	  {
              	  	System.err.println("Invalid timer function syntax in the START clause.");
              	  	return false;
              	  }
              	String FuncName = s[i-1];
              	if ( (stop == -1 || stop > end) && (start != end) )
                  stop = end;
              	while (stop != -1 && stop <=end)
              	  {
              	  	s[i] = AQString.substring(start,stop).trim();
              	  	if (FuncName.equals(TimerFunctions.ONDAYOFWEEK) || FuncName.equals(TimerFunctions.INDAYOFWEEKINTERVAL) )
              	      {
              	  	    int len = s[i].length() - 1;
              	  	    if (s[i].charAt(0) == '\"')
              	  	    s[i] = s[i].substring(1,len);
              	  	    int day_of_week = TimerFunctions.testDayOfWeek(s[i]);
              	  	    if (day_of_week == -1)
              	  	     {
               	  	        System.err.println("Invalid timer function syntax in the START clause."); 
               	  	        return false;
               	          }
              	      }
              	    else 
              	      {
              	      	if (s[i].equals(TimerFunctions.$ANY))
              	          s[i] = String.valueOf(TimerFunctions.$ANY_VAL);
              	  	    try 
              	  	      {
              	  	      	int j = Integer.parseInt(s[i]);
              	  	      }
              	  	    catch (Exception e) 
              	  	      {
              	  	      	System.err.println("Invalid timer function syntax in the START clause."); 
              	  	      	return false;
              	  	      }
              	  	  }
              	  	i++;
              	  	start = stop + 1;
              	  	if (start > end) break;
              	  	stop = AQString.indexOf(",",start);
              	  	if (stop > end || stop == -1) stop = end;
              	  }
              	              	  
              	 try {
              	 
              	 if (FuncName.equals(TimerFunctions.ATTIME) || FuncName.equals(TimerFunctions.ONDAY) )
                   if (!s[i-5].equals(DELIMITER_STRING))
                     {
                        System.err.println("Invalid parameters for the timer function in the START clause.");
                        return false;
                     }
                 
                 if (FuncName.equals(TimerFunctions.EVERY))
                   {
                     System.err.println("Invalid parameters for the timer function in the START clause.");
                     return false;
                   }
                 
              	 if (FuncName.equals(TimerFunctions.ATTIME) || FuncName.equals(TimerFunctions.INTIMEINTERVAL))
                   if (Integer.parseInt(s[i-1]) < -1 || Integer.parseInt(s[i-1]) > 59 || Integer.parseInt(s[i-2]) < -1 || Integer.parseInt(s[i-2]) > 59 || Integer.parseInt(s[i-3]) < -1 || Integer.parseInt(s[i-3]) > 23)
                    {
                	  System.err.println("Invalid parameters for the timer function in the START clause.");
                	  return false;
                    }
                 
                 if (FuncName.equals(TimerFunctions.INTIMEINTERVAL))
                   if (Integer.parseInt(s[i-4]) < -1 || Integer.parseInt(s[i-4]) > 59 || Integer.parseInt(s[i-5]) < -1 || Integer.parseInt(s[i-5]) > 59 || Integer.parseInt(s[i-6]) < -1 || Integer.parseInt(s[i-6]) > 23
                    || Integer.parseInt(s[i-3]) < Integer.parseInt(s[i-6]) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) < Integer.parseInt(s[i-5])) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) == Integer.parseInt(s[i-5]))&&(Integer.parseInt(s[i-1]) <= Integer.parseInt(s[i-4])) 
                    || !s[i-8].equals(DELIMITER_STRING))
                    { 
                	  System.err.println("Invalid parameters for the timer function in the START clause.");
                      return false;
                    }
                
                 if (FuncName.equals(TimerFunctions.ONDAY) || FuncName.equals(TimerFunctions.INDAYINTERVAL))
                   if (Integer.parseInt(s[i-1]) < -1 || Integer.parseInt(s[i-1]) > 31 || Integer.parseInt(s[i-2]) < -1 || Integer.parseInt(s[i-2]) > 12 
                    || Integer.parseInt(s[i-1]) == 0 || Integer.parseInt(s[i-2]) == 0 || Integer.parseInt(s[i-3]) < 0 )
                    {
                	   System.err.println("Invalid parameters for the timer function in the START clause.");
                	   return false;
                    }
                  
                 if (FuncName.equals(TimerFunctions.INDAYINTERVAL))
                   if (Integer.parseInt(s[i-4]) < -1 || Integer.parseInt(s[i-4]) > 31 || Integer.parseInt(s[i-5]) < -1 || Integer.parseInt(s[i-5]) > 12
                    || Integer.parseInt(s[i-4]) == 0 || Integer.parseInt(s[i-5]) == 0 || Integer.parseInt(s[i-6]) < 0
                    || Integer.parseInt(s[i-3]) < Integer.parseInt(s[i-6]) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) < Integer.parseInt(s[i-5])) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) == Integer.parseInt(s[i-5]))&&(Integer.parseInt(s[i-1]) <= Integer.parseInt(s[i-4])) 
                    || !s[i-8].equals(DELIMITER_STRING))
                    {  
                	   System.err.println("Invalid parameters for the timer function in the START clause.");
                	   return false;
                    }
              
                 if (FuncName.equals(TimerFunctions.INDAYOFWEEKINTERVAL))
                   if (!s[i-4].equals(DELIMITER_STRING))
                    {
                      System.err.println("Invalid parameters for the timer function in the START clause.");
                      return false;
                    } 
                    
                 if (FuncName.equals(TimerFunctions.ONDAYOFWEEK))
                   if (!s[i-3].equals(DELIMITER_STRING))
                     {
                        System.err.println("Invalid parameters for the timer function in the START clause.");
                        return false;
                     }
                    
                 } 
                 catch (Exception e) 
                  {
                  	System.err.println("Invalid parameters for the timer function in the START clause.");
                    return false;
                  }
                    
              	 s[i++] = DELIMITER_STRING; }
              	
              } 
              else 
                {
                  s[i++] = NULL_STRING;
                  s[i++] = DELIMITER_STRING;
                }  // end of parsing the start clause
                
            
            // STOP clause        
            start = AQString.indexOf(" stop ");
            if (start != -1)
              {
              	start += 5;
              	stop = AQString.indexOf(" " + TimerFunctions.$NOW + " ",start);
              	if (stop != -1)
              	  {
              	  	System.err.println("Invalid timer function syntax in the STOP clause. $now is not permitted");
              	  	return false;
              	  }
              	else {
              	stop = AQString.indexOf("(",start);
              	if (stop == -1)
              	  {
              	  	System.err.println("Invalid timer function syntax in the STOP clause.");
              	  	return false;
              	  }
              	s[i] = AQString.substring(start,stop).trim();
              	if ( (s[i].equals(TimerFunctions.EVERY)) || (s[i].equals(TimerFunctions.ATTIME)) || (s[i].equals(TimerFunctions.ONDAY)) || (s[i].equals(TimerFunctions.INTIMEINTERVAL)) || (s[i].equals(TimerFunctions.INDAYINTERVAL)) || (s[i].equals(TimerFunctions.ONDAYOFWEEK)) || (s[i].equals(TimerFunctions.INDAYOFWEEKINTERVAL)) )
              	  {i++;}
              	else 
              	  {
              	  	System.err.println("Invalid timer function syntax in the STOP clause.");
              	  	return false;
              	  }
              	start = stop + 1;
              	stop = AQString.indexOf(",",start);
              	end = AQString.indexOf(")",start);
              	if (end == -1)
              	  {
              	  	System.err.println("Invalid timer function syntax in the STOP clause.");
              	  	return false;
              	  }
              	String FuncName = s[i-1];
              	if ( (stop == -1 || stop > end) && (start != end) )
                  stop = end;
              	while (stop != -1 && stop <=end )
              	  {
              	  	s[i] = AQString.substring(start,stop).trim();
                    if (FuncName.equals(TimerFunctions.ONDAYOFWEEK) || FuncName.equals(TimerFunctions.INDAYOFWEEKINTERVAL) )
              	      {
              	  	    int len = s[i].length() - 1;
              	  	    if (s[i].charAt(0) == '\"')
              	  	    s[i] = s[i].substring(1,len);
              	  	    int day_of_week = TimerFunctions.testDayOfWeek(s[i]);
              	  	    if (day_of_week == -1)
              	  	     {
               	  	        System.err.println("Invalid timer function syntax in the STOP clause."); 
               	  	        return false;
               	          }
              	      }
              	    else 
              	      {
              	      	if (s[i].equals(TimerFunctions.$ANY))
              	          s[i] = String.valueOf(TimerFunctions.$ANY_VAL);
              	  	    try 
              	  	      {
              	  	      	int j = Integer.parseInt(s[i]);
              	  	      }
              	  	    catch (Exception e) 
              	  	      {
              	  	      	System.err.println("Invalid timer function syntax in the STOP clause."); 
              	  	      	return false;
              	  	      }
              	  	  } // else
              	  	i++;
              	  	start = stop + 1;
              	  	if (start > end) break;
              	  	stop = AQString.indexOf(",",start);
              	  	if (stop == -1) stop = end;
              	  }
              	  
              	try {
              	 
              	 if (FuncName.equals(TimerFunctions.ATTIME) || FuncName.equals(TimerFunctions.ONDAY) )
                   if (!s[i-5].equals(DELIMITER_STRING))
                     {
                        System.err.println("Invalid parameters for the timer function in the STOP clause.");
                        return false;
                     }
                 
                 if (FuncName.equals(TimerFunctions.EVERY))
                   {
                     System.err.println("Invalid parameters for the timer function in the STOP clause.");
                     return false;
                   }
                 
              	 if (FuncName.equals(TimerFunctions.ATTIME) || FuncName.equals(TimerFunctions.INTIMEINTERVAL))
                   if (Integer.parseInt(s[i-1]) < -1 || Integer.parseInt(s[i-1]) > 59 || Integer.parseInt(s[i-2]) < -1 || Integer.parseInt(s[i-2]) > 59 || Integer.parseInt(s[i-3]) < -1 || Integer.parseInt(s[i-3]) > 23)
                    {
                	  System.err.println("Invalid parameters for the timer function in the STOP clause.");
                	  return false;
                    }
                
                 if (FuncName.equals(TimerFunctions.INTIMEINTERVAL))
                   if (Integer.parseInt(s[i-4]) < -1 || Integer.parseInt(s[i-4]) > 59 || Integer.parseInt(s[i-5]) < -1 || Integer.parseInt(s[i-5]) > 59 || Integer.parseInt(s[i-6]) < -1 || Integer.parseInt(s[i-6]) > 23
                    || Integer.parseInt(s[i-3]) < Integer.parseInt(s[i-6]) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) < Integer.parseInt(s[i-5])) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) == Integer.parseInt(s[i-5]))&&(Integer.parseInt(s[i-1]) <= Integer.parseInt(s[i-4])) 
                    || !s[i-8].equals(DELIMITER_STRING))
                     {
                	    System.err.println("Invalid parameters for the timer function in the STOP clause.");
                	    return false;
                     }
                
                 if (FuncName.equals(TimerFunctions.ONDAY) || FuncName.equals(TimerFunctions.INDAYINTERVAL))
                   if (Integer.parseInt(s[i-1]) < -1 || Integer.parseInt(s[i-1]) > 31 || Integer.parseInt(s[i-2]) < -1 || Integer.parseInt(s[i-2]) > 12 
                    || Integer.parseInt(s[i-1]) == 0 || Integer.parseInt(s[i-2]) == 0 || Integer.parseInt(s[i-3]) < 0)
                    {
                	  System.err.println("Invalid parameters for the timer function in the STOP clause.");
                	  return false;
                    }
                  
                 if (FuncName.equals(TimerFunctions.INDAYINTERVAL))
                   if (Integer.parseInt(s[i-4]) < -1 || Integer.parseInt(s[i-4]) > 31 || Integer.parseInt(s[i-5]) < -1 || Integer.parseInt(s[i-5]) > 12
                    || Integer.parseInt(s[i-4]) == 0 || Integer.parseInt(s[i-5]) == 0 || Integer.parseInt(s[i-6]) < 0
                    || Integer.parseInt(s[i-3]) < Integer.parseInt(s[i-6]) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) < Integer.parseInt(s[i-5])) || (Integer.parseInt(s[i-3]) == Integer.parseInt(s[i-6]))&&(Integer.parseInt(s[i-2]) == Integer.parseInt(s[i-5]))&&(Integer.parseInt(s[i-1]) <= Integer.parseInt(s[i-4])) 
                    || !s[i-8].equals(DELIMITER_STRING))
                    {
                	    System.err.println("Invalid parameters for the timer function in the STOP clause.");
                	    return false;
                    }
              
                 if (FuncName.equals(TimerFunctions.INDAYOFWEEKINTERVAL))
                   if (!s[i-4].equals(DELIMITER_STRING))
                    {
                       System.err.println("Invalid parameters for the timer function in the STOP clause.");
                       return false;
                    } 
                  
                 if (FuncName.equals(TimerFunctions.ONDAYOFWEEK))
                   if (!s[i-3].equals(DELIMITER_STRING))
                     {
                        System.err.println("Invalid parameters for the timer function in the STOP clause.");
                        return false;
                     }
                    
                 } 
                 catch (Exception e) 
                  {
                  	System.err.println("Invalid parameters for the timer function in the STOP clause.");
                    return false;
                  }
              	              	
              	s[i++] = DELIMITER_STRING; }
              	
              }
              else 
                {
                	s[i++] = NULL_STRING;
                	s[i++] = DELIMITER_STRING;
                }  // end of parsing the stop clause     
            
            
            // LIFETIME clause
            start = AQString.indexOf(" lifetime ");
            if (start != -1)
              {
              	start += 9;
              	str = AQString.substring(start).trim();
                              	
              	// extract the lifetime interval
              	start = str.indexOf(" ");
              	if (start == -1)
              	  start = str.indexOf(TimerFunctions.$ONCE);
              	if (start == -1)
                  {
                  	System.err.println("Invalid query syntax in the LIFETIME clause.");
                  	return false;
                  }
                s[i] = str.substring(0,start).trim();
                
                if (str.equals(TimerFunctions.$ONCE))
                  { s[i] = String.valueOf(AQ.ONCE); i++;}
                else
                  {
                  	try 
                  	  { 
                  	     stop = Integer.parseInt(s[i]); 
                  	  }
                    catch (Exception e) 
                      { 
                         System.err.println("Invalid LIFETIME value. It must be a digit."); 
                         return false;
                      }
                    i++;
                
                    // extract the lifetime unit
                    s[i] = str.substring(start).trim();
                    if (stop > 1) 
                      {
                  	    int len = s[i].length();
                  	    if (!s[i].substring(len-1).equals(TimerFunctions.PLURAL))
                  	      {
                  	  	    System.err.println("Invalid LIFETIME unit.");
                  	  	    return false;
                          }
                        s[i] = s[i].substring(0,len-1);
                       }
                    int j =0;
                    while (TimerFunctions.unitArray[j] != null)
                      {
                  	    if (s[i].equals(TimerFunctions.unitArray[j])) break;
                  	    j++;
                      }
                
                    if (TimerFunctions.unitArray[j] == null)
                      {
                  	    System.err.println("Invalid LIFETIME unit.");
                   	    return false;
                      }
                    i++;
                 } 
              }  
              else 
                s[i++] = NULL_STRING; // end of parsing the LIFETIME clause 
                              
             /* for (i=0;i<20;i++)
               System.out.println(s[i]);
             System.exit(0); */
                        
             /* Successfully parse the action-embedded query */
             return true;                            
          }
          
          
    /**
     * Parse the input parameters for the nested (second-level) action.
     *
     * @return <code>true</code> if the parsing if successful;
	 *         <code>false</code> otherwise
     */
    public boolean parseNestedAction() {
    	
    	int i, nested, start, stop;
    	String str;
    	
    	actionParams2 = new String [20];    	
    	
    	i = 5;
    	nested = -1;
    	while (s[i] != null)
    	  if ( (s[i].indexOf("(") != -1) && (s[i].indexOf(")") != -1) )
    	    {nested = i; break;}
    	  else i++;
    	if (nested == -1)
    	  {
    	  	System.err.println("Fail to parse the second-level action.");
    	  	return false;
    	  }
        
        // parse the action name        
        start = s[nested].indexOf("(");
        str = s[nested].substring(0,start).trim();        
        i=0; 
        while (actions[i] != null)
          { 
             if (str.equals(actions[i].name) )
               {action2 = actions[i]; break;}
             i++;
          }
        if (action2 == null)
          {
             System.err.println("Unrecognized second-level action: " + str + "().");
             return false;
          }    	
        
        // parse the action parameters
        i = 0;
        start += 1;
        while (true)
          {
          	stop = s[nested].indexOf(",",start);
          	if (stop == -1)
          	  {actionParams2[i] = s[nested].substring(start,s[nested].length()-1).trim(); break;}
          	else {actionParams2[i] = s[nested].substring(start,stop).trim(); start = stop +1;}
          	if (actionParams2[i].charAt(0) == '\"' && actionParams2[i].charAt(actionParams2[i].length()-1) == '\"')
          	  actionParams2[i] = actionParams2[i].substring(1,actionParams2[i].length()-1);
          	i++;
          }
    	  
    	return true;   	
    	
    }
    
    
    /**
     * Compute the distance between two points.
     *
     * @param loc1 the 3D Cartesian coordinate of the first point
     * @param loc2 the 3D Cartesian coordinate of the second point
     *
     * @return the distance between the given points
     *
     * @see Point 
     */
    public static double dist (Point loc1, Point loc2)
	  {
	  	return Math.sqrt(Math.pow(Math.abs(loc1.x - loc2.x), 2) + Math.pow(Math.abs(loc1.y - loc2.y), 2) + Math.pow(Math.abs(loc1.z - loc2.z), 2)); 
	  }    
     
     
    
    /**
     * Read the system-level configuration file and start the Aorta query processor.
     *
     * @param argv argv[0] gives the path name of the configuration file 
     */
    public static void main(String argv[]) {
		
	   try {	   	
	   	
	    String AQString;
        String AQFile = "../conf/queries.txt";
        int AQNum = 0;
        int numQueries = 0;                      
                	
		if (argv.length != 1)
	        {
	           System.err.println("USAGE: java Aorta <configFile>");
	           System.exit(1);	
	        }
	        
	    String configFile = argv[0];
	    BufferedReader configFileReader = new BufferedReader(new FileReader(configFile));
	    String line;
	    int mark;
	    BufferedReader queryFileReader = null;
	    
	    
	    /* Read the system configuration information from the configuration file */
	    while ( (line = configFileReader.readLine()) != null)
	      {
	      	
	      	if ( (mark = line.indexOf("queryFileName=")) != -1)
	      	  {
	      	  	mark += "queryFileName=".length();
	      	  	AQFile = line.substring(mark);
	      	  	queryFileReader = new BufferedReader(new FileReader(AQFile));
	      	  }
	      	  
	      	  
	      	if ( (mark = line.indexOf("numQueries=")) != -1)
	      	  {
	      	  	mark += "numQueries=".length();
	      	  	numQueries = new Integer(line.substring(mark)).intValue();	      	  	
	      	  }
	      	  
	      	if ( (mark = line.indexOf("sensorSimulation=")) != -1)
	      	  {
	      	  	mark += "sensorSimulation=".length();
	      	  	SENSOR_SIMULATION = new Boolean(line.substring(mark)).booleanValue();
	      	  }
	      	  
	      	if ( (mark = line.indexOf("cameraSimulation=")) != -1)
	      	  {
	      	  	mark += "cameraSimulation=".length();
	      	  	CAMERA_SIMULATION = new Boolean(line.substring(mark)).booleanValue();
	      	  }
	      	  
	      	if ( (mark = line.indexOf("phoneEmulation=")) != -1)
	      	  {
	      	  	mark += "phoneEmulation=".length();
	      	  	PHONE_EMULATION = new Boolean(line.substring(mark)).booleanValue();
	      	  } 
	      	  
	      	if ( (mark = line.indexOf("robotSimulation=")) != -1)
	      	  {
	      	  	mark += "robotSimulation=".length();
	      	  	ROBOT_SIMULATION = new Boolean(line.substring(mark)).booleanValue();	      	  	
	      	  }        	
	      	
	      }		
			   	
        System.out.println("");
	    System.out.println("\t\t\tAorta Query Processor\tVersion 0.1");
	    System.out.println("");
	        
	    Aorta aortaqp = new Aorta();
        
                      
        /* Initialize TinyDB, suppress TinyDB's initialization messages to System.out and System.err */
	    if (SENSOR_SIMULATION == false)
	      { 
	      	PrintStream out_t1, out_t2, err_t;
	        out_t1 = System.out;
	        err_t = System.err;
	        File df = new File(".temp.out_err");
	        out_t2 = new PrintStream (new FileOutputStream(df));
	        System.setOut(out_t2);
	        System.setErr(out_t2);
	        
	        TinyDBMain.initMain();
	        
	        // restore System.out and System.err
	        System.setOut(out_t1); 
	        System.setErr(err_t);
	        out_t2.close();
	        df.delete();
	      }
	      
	                    
        /* Simple command line user interface for Aorta.
           Read and parse the commands (CREATE, DROP, ALTER ...) inputted by users.
           a GUI may be used in later version of Aorta after development. */
        CmdLineUserIF cmdUI = new CmdLineUserIF (aortaqp);
        cmdUI.start();
        
             	
      	/* When Aorta starts, it reads a query files containing a number of action-embedded query specifications (plain SQL texts)
      	   and install (register) them to the system. */
      	System.out.println("Installing queries from the query file " + AQFile + " ... "); 
		System.out.println();		
      	
      	while ( ((AQString = queryFileReader.readLine()) != null) && (AQNum < numQueries) )
		  {
		    if (AQString.trim().equals(""))
		      continue;
		      
		    System.out.println("Query " + String.valueOf(AQNum+1) + ": " + AQString);
		    System.out.println("");
		    AQString.trim();
		    AQString = AQString + " ";
		    AQString = AQString.toLowerCase();  // convert the query string to lower case
		    AQString.replace('\t',' ');
		    AQString.replace('\n',' ');
		    AQString.replace('\r',' ');
		    AQString.replaceAll("  ", " ");
		    
		    aortaqp.addAQ(AQString);
		    
		    // Thread.currentThread().sleep(2000); // add delays between multiple query registration 
		    
		    AQNum ++;
		    
		  } 
		  
		  System.out.println("Finish installing the queries."); 
		  System.out.println();
		  
		  // show the command line prompt
		  System.out.print("Aorta>");
		  		  
	   } 
	   catch (Exception e) {e.printStackTrace();}
	
    }  
    
}