/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * A simulator that simulates the action scheduling process within 
 * a shared action operator in Aorta.                
 *                                                         
 * It is developed for the sake of convenient performance evaluation  
 * of the proposed heuristic action scheduling algorithm and the  
 * comparison between our algorithm and different previous heuristic 
 * scheduling algorithms in the literature.
 *
 * @see ActionWorkload
 * @see ActionRequest
 * @see DHASA
 * @see SRF
 * @see EDF
 * @see LFD_LFR
 * @see RANDOM
 */ 
public class AortaSchedulingSimulator {
	
	// change these values to evalute the performance of the algorithms under different parameter settings
	public static final byte POISSON = 0x01;
	public static final byte CONSTANT = 0x02;
	public static byte requestArrivalPattern = POISSON;
	
	public static final byte UNIFORM = 0x01;
	public static final byte ZIPFIAN = 0x02;
    public static byte candidateDeviceDistribution = UNIFORM;
	
	public static final byte PHOTO = 0x01;
	public static final byte WARN = 0x02;
	public static byte actionName = WARN;
	
	/*public static final byte DHASA = 0x01; // our dynamic, heuristic action scheduling algorithm
	public static final byte EDF = 0x02; // earliest deadline first
	public static final byte SRF = 0x03; // shortest request first
	public static final byte LFD_LFR = 0x04; // least flexible machine + least flexible job
	public static final byte RANDOM = 0x05; // randomly schedule a request on the free device
	public static final byte MCT = 0x06;  // minimum completion time
	public static final byte MET = 0x07;  // minimum execution time
	public static byte schedulingAlgorithm = DHASA;*/
		
	/*public static final byte WEIGHTED_SUM = 0x01;
	public static final byte MULTIPLICATION = 0x02;
	public static final byte PRIORITY = 0x03;
	public static byte heuristic = WEIGHTED_SUM; // currently adopted heuristic*/
	
	public static final byte COST_BASIS = 0x01;
	public static final byte DEADLINE_BASIS = 0x02;
	public static final byte CDN_BASIS = 0x03;
	public static byte basis = COST_BASIS;	
	
	public static final double MAX_DEV_FAIL_PROB = 0.0;
	public static boolean WRE = false;
	
	public static int deviceNum = 20;  // total number of devices involved in the scheduling
	
	public static int requestNum = 1000;  // total number of requests examined in a workload
	
	public static double requestArrivalRate = 1.0/1.0;  // time units (seconds) between two consecutive requests on average
		
	
	/**
	 * Generate a synthetic action workload and then execute various scheduling
	 * algorithms on the workload.
	 */
	public static void schedule ()
	  {	
		long startTime, endTime;
		String logFile;
		Vector requests = null, requests_backup = null;
		PrintWriter logFileWriter = null;
		
 					
		System.out.println("\t\t     Aorta Action Workload Scheduling Simulator");
		System.out.println("A grouping time interval begins.");
		
		ActionWorkload aw = new ActionWorkload();
		
		requests = aw.generate();
		
		double failProbs [] = generateFailProbs();                         
	                 
	    //printRequests(requests);	    
	    
	    //printRequestsStatus(requests);		
	    
	    //System.exit(0);
	    
	    
	    logFile = "../logs/scheduling_logs/settings.txt";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("#requests = " + requestNum);
		logFileWriter.println("#devices = " + deviceNum);
		logFileWriter.println("request arrival rate = " + (1.0/requestArrivalRate));
		logFileWriter.print("Distribution of candidate device numbers of the requests = ");
		if (candidateDeviceDistribution == UNIFORM)
		  logFileWriter.println("UNIFORM");
		else
		  logFileWriter.println("ZIPFIAN");
		logFileWriter.print("action = ");
		if (actionName == PHOTO)
		  logFileWriter.println("photo()");
		else
		  logFileWriter.println("warn()");
		logFileWriter.println("W = " + ActionWorkload.W);
		logFileWriter.println();
		logFileWriter.flush();
		logFileWriter.close();
		
		
		if (MAX_DEV_FAIL_PROB > 0)
		  {
		    logFile = "../logs/scheduling_logs/deviceFailureProbabilities.txt";
			try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
			catch (Exception e) {e.printStackTrace();}
			for (int i=0;i<deviceNum;i++)
	  	      logFileWriter.println("Device " + (i+1) + ": " + failProbs[i]);
	  	    logFileWriter.println();
	  	    logFileWriter.println("Maximum device failure rate: " + MAX_DEV_FAIL_PROB);
			logFileWriter.println();
			logFileWriter.flush();
			logFileWriter.close();		  	
		  }
	    
	                       
	    // DHASA + Weighted Sum
	    DHASA dhasa;
		System.out.println("DHASA + Weighted Sum (Cost Basis) ... ");
		logFile = "../logs/scheduling_logs/DHASA_WS_COST_BASIS.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
 	  				
		requests_backup = new Vector(requestNum,10);
		backupRequests(requests,requests_backup);
		dhasa = new DHASA(requests,logFileWriter);
		initialize(dhasa.schdevs,logFileWriter,failProbs);
		dhasa.schedule();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();
		
		
		// DHASA + Weighted Sum	+ without reliability estimation   
	    if (WRE)
	      {
	      	System.out.println("DHASA + Weighted Sum (Cost Basis) ... ");
			logFile = "../logs/scheduling_logs/DHASA_WS_COST_BASIS_WRE.log";
			try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
			catch (Exception e) {e.printStackTrace();}
			logFileWriter.println("A grouping time interval begins.");
			logFileWriter.println();
 	  				
			requests_backup = new Vector(requestNum,10);
			backupRequests(requests,requests_backup);
			dhasa = new DHASA(requests,logFileWriter);
			initialize(dhasa.schdevs,logFileWriter,failProbs);
			dhasa.schedule();
			requests = requests_backup;
		
			logFileWriter.println();
			logFileWriter.println("The grouping time interval ends.");
			logFileWriter.flush();
			logFileWriter.close();		
		
		 	System.exit(0);
		  }
		
		
		// DHASA + Weighted Sum + deadline basis
		basis = DEADLINE_BASIS;	    
		System.out.println("DHASA + Weighted Sum (Deadline Basis) ... ");
		logFile = "../logs/scheduling_logs/DHASA_WS_DEADLINE_BASIS.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
 	  				
		requests_backup = new Vector(requestNum,10);
		backupRequests(requests,requests_backup);
		dhasa = new DHASA(requests,logFileWriter);
		initialize(dhasa.schdevs,logFileWriter,failProbs);
		dhasa.schedule();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();
		
		
		// DHASA + Weighted Sum + candidate device number (cdn) basis
		basis = CDN_BASIS;	    
		System.out.println("DHASA + Weighted Sum (Candidate Device Number Basis) ... ");
		logFile = "../logs/scheduling_logs/DHASA_WS_CDN_BASIS.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
 	  				
		requests_backup = new Vector(requestNum,10);
		backupRequests(requests,requests_backup);
		dhasa = new DHASA(requests,logFileWriter);
		initialize(dhasa.schdevs,logFileWriter,failProbs);
		dhasa.schedule();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();
		basis = COST_BASIS;
		
		
		// SRF
		System.out.println("SRF ... ");
		logFile = "../logs/scheduling_logs/SRF.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(requestNum,10);
		backupRequests(requests, requests_backup);
		SRF srf = new SRF(requests,logFileWriter);
		initialize(srf.schdevs,logFileWriter,failProbs);
		srf.schedule();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();			
			
		
		// EDF
		System.out.println("EDF ... ");
		logFile = "../logs/scheduling_logs/EDF.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(requestNum,10);
		backupRequests(requests, requests_backup);
		EDF edf = new EDF(requests,logFileWriter);
		initialize(edf.schdevs,logFileWriter,failProbs);
		edf.schedule();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();
		
		
		// LFD-LFR
		System.out.println("LFD-LFR ... ");
		logFile = "../logs/scheduling_logs/LFD-LFR.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(requestNum,10);
		backupRequests(requests, requests_backup);
		LFD_LFR lfd_lfr = new LFD_LFR(requests,logFileWriter);
		initialize(lfd_lfr.schdevs,logFileWriter,failProbs);
		lfd_lfr.schedule();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();	
		
		
		// RANDOM
		System.out.println("RANDOM ... ");
		logFile = "../logs/scheduling_logs/RANDOM.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(requestNum,10);
		backupRequests(requests, requests_backup);
		RANDOM random = new RANDOM(requests,logFileWriter);
		initialize(random.schdevs,logFileWriter,failProbs);
		random.schedule();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();
			
		
        System.out.println("The grouping time interval ends.");       
      }  
    
    
    /**
	 * Generate the average action failure probabilities on the devices. 
	 */
	public static double [] generateFailProbs()
	  {
	  	int i;
	  	
	  	double failProbs [] = new double [deviceNum];
	  	
	  	Random ran = new Random(deviceNum);
	  	
	  	for (i=0;i<deviceNum;i++)
	  	  failProbs[i] = ran.nextDouble()*MAX_DEV_FAIL_PROB;
	  	  
	  	/*for (i=0;i<deviceNum;i++)
	  	  System.out.println("Device " + (i+1) + ": " + failProbs[i]);
	  	
	  	System.out.println("Maximum device failure rate: " + MAX_DEV_FAIL_PROB);
	  	
	  	System.exit(1);*/
	  	  
	  	return failProbs;
	  }	  
	  
      
    /** 
     * Backup the generated synthetic workload so that the same
     * workload can be used for all scheduling algorithm we compare.
     * 
     * @param requests the action workload for backup
     * @param requests_backup a backup of the workload
     */
    public static void backupRequests (Vector requests, Vector requests_backup)
      {
      	int i,j,k;
      	ActionRequest ar, ar_backup;
      	Tuple t, tt;
      	Vector v, vv;      	      	
      	
      	for (i=0;i<requests.size();i++)
		  {
		  	ar = (ActionRequest)requests.elementAt(i);
		  	
		  	ar_backup = new ActionRequest();
		  	
		  	ar_backup.aid = ar.aid;
		  	
		  	ar_backup.deadline = ar.deadline;
		  	
		  	ar_backup.startTime = ar.startTime;
		  	
		  	ar_backup.didVector = new Vector(ar.didVector.size());
		  	for (j=0;j<ar.didVector.size();j++)
		  	  {
		  	    String did = new String((String)ar.didVector.elementAt(j));
		  	    
		  	    ar_backup.didVector.addElement((String)did);
		  	  }
		  	  
		  	ar_backup.costVector = new Vector(ar.costVector.size());
		  	for (j=0;j<ar.costVector.size();j++)
		  	  {
		  	    String cost = new String((String)ar.costVector.elementAt(j));
		  	    
		  	    ar_backup.costVector.addElement((String)cost);
		  	  }
		  	  
		  	ar_backup.statusVector = new Vector(ar.statusVector.size());
		  	for (j=0;j<ar.statusVector.size();j++)
		  	  {
		  	    t = (Tuple)ar.statusVector.elementAt(j);
		  	    
		  	    v = t.fields;
		  	    
		  	    vv = new Vector (v.size());
		  	    
		  	    for (k=0;k<v.size();k++)
		  	      {
		  	      	String param = new String((String)v.elementAt(k));
		  	      	
		  	      	vv.addElement((String)param);		  	      	
		  	      }
		  	      
		  	    tt = new Tuple(vv);
		  	    
		  	    ar_backup.statusVector.addElement((Tuple)tt);
		  	  }
		  	  
		  	requests_backup.addElement((ActionRequest)ar_backup);
		  }
      }   
    
	
	/**
	 * Initialize the devices involved in the scheduling.
	 */
	public static void initialize (SchedulingDevice schdevs [], PrintWriter logFileWriter, double [] failProbs)
 	  {
 	   	for (int i=0;i<schdevs.length;i++)
    	  {
    	  	schdevs[i] = new SchedulingDevice();
    	  	
    	  	SchedulingDevice sd = schdevs[i];
    	  	
    	  	sd.id = i+1;
    	  	
    	  	if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
    	  	  sd.camSim = new CameraSimulator(String.valueOf(sd.id));
    	    else 
    	  	  if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.WARN)
    	  	    sd.er1Sim = new ER1Simulator(String.valueOf(sd.id));
    	      else
    	        {
      	          logFileWriter.println("Unsupported action to be scheduled. Must be \"photo\" or \"warn\".");
      	          logFileWriter.close();
      	      
      	          System.err.println("Unsupported action to be scheduled. Must be \"photo\" or \"warn\".");
      	          System.exit(0);
      	        }
      	        
      	    sd.failProb = failProbs[i];
      	  }   	    	  	
 	  } 	 
	
	
	/** 
	 * Update the cost of other requests on the camera that services the currently selected {@link Photo} request.
	 */
	public static void updatePhotoCosts(Vector requests, String sel_did, SchedulingDevice [] schdevs) 
	  {
		int i, j, id;
		int pan, tilt, zoom, deltaPan, deltaTilt, deltaZoom;
		double headMovementCost, actionCost;
		Tuple t;
		ActionRequest ar;
				
		
		for (i=0;i<requests.size();i++)
		  {
		  	ar = (ActionRequest)requests.elementAt(i);
		  	
		  	for (j=0;j<ar.didVector.size();j++)
		      if (((String)ar.didVector.elementAt(j)).equals(sel_did))
		       {
		         t = (Tuple)ar.statusVector.elementAt(j);
		         
		         pan = Integer.parseInt((String)t.fields.elementAt(0));
		         
    	   	     tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	     
    	   	     zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	     
    	   	     id = Integer.parseInt(sel_did)-1;
    	   	     
    	   	     deltaPan = Math.abs(schdevs[id].camSim.pan - pan);
    	   	     
    	   	     deltaTilt = Math.abs(schdevs[id].camSim.tilt - tilt);
    	   	     
    	   	     deltaZoom = Math.abs(schdevs[id].camSim.zoom - zoom);
    	   	     
    	   	     headMovementCost = schdevs[id].camSim.computeHeadMovementCost(deltaPan, deltaTilt, deltaZoom);
		         
		         actionCost = AortaSchedulingSimulator.estimatePhotoActionCost(headMovementCost)/1000.0;  	        
		         
		         ar.costVector.removeElementAt(j);
		         
		         ar.costVector.insertElementAt(String.valueOf(actionCost),j);
		       } 
		   }	
	  }	
	  
	  
	/** 
	 * Update the cost of other requests on the robot that services the currently selected {@link Warn} request.
	 */
	public static void updateWarnCosts(Vector requests, String sel_did, SchedulingDevice schdevs [])
	  {
	  	int i, j, id;	  	
		double actionCost;
		Tuple t;
		Point loc = new Point();
		ActionRequest ar;	
		
		for (i=0;i<requests.size();i++)
		  {
		  	ar = (ActionRequest)requests.elementAt(i);
		  	for (j=0;j<ar.didVector.size();j++)
		      if (((String)ar.didVector.elementAt(j)).equals(sel_did))
		       {
		         t = (Tuple)ar.statusVector.elementAt(j);
		         
		         loc.x = Double.parseDouble((String)t.fields.elementAt(0));
		         
    	   	     loc.y = Double.parseDouble((String)t.fields.elementAt(1));    	   	     
    	   	     
    	   	     loc.z = Double.parseDouble((String)t.fields.elementAt(2));    	   	     
    	   	     
    	   	     id = Integer.parseInt(sel_did)-1;
    	   	     
    	   	     actionCost = AortaSchedulingSimulator.estimateWarnActionCost(schdevs[id].er1Sim.loc,schdevs[id].er1Sim.angle,loc)/1000.0;      	   	  	
		         
		         ar.costVector.removeElementAt(j);
		         
		         ar.costVector.insertElementAt(String.valueOf(actionCost),j);
		       } 
		   }
	  }           
    
	  
	/** 
	 * Estimate the cost of an execution of the {@link Photo} action.
	 */
	public static double estimatePhotoActionCost (double headMovementCost)
      {
      	return 2*CameraSimulator.STANDARD_POLLING_TIME + headMovementCost + CameraSimulator.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST;
      }
      
      
    /** 
     * Estimate the cost of an execution of the {@link Warn} action. 
     */
   public static double estimateWarnActionCost (Point currentLoc, double currentAngle, Point targetLoc)
     {
       double actionCost = 0;
       
       actionCost += Math.sqrt(Math.pow(targetLoc.x-currentLoc.x,2)+Math.pow(targetLoc.y-currentLoc.y,2))*100*ER1Simulator.MOVEPERCM_ESTIMATED_COST;
       
       
       double deltaAngle = computeDeltaAngle(currentLoc, targetLoc);
       
       deltaAngle -= currentAngle;
	   
	   if (deltaAngle > 180)
	     deltaAngle -= 360;
	   
	   if (deltaAngle <= -180)
	     deltaAngle += 360;
	   
	   actionCost += Math.abs(ER1Simulator.ROTATEPERDEG_ESTIMATED_COST*deltaAngle);
       
       actionCost += ER1Simulator.DEFAULT_SOUND_COST;
       
       return actionCost; 
     }     
      
      
    /**
     * Compute the angle to be rotated given the current ang target locations
     */
    public static double computeDeltaAngle(Point currentLoc, Point targetLoc)
     {
       double currentX = currentLoc.x, currentY = currentLoc.y;
       double targetX = targetLoc.x, targetY = targetLoc.y;  	    
       double deltaAngle; 
       
       if ((currentX == targetX) && (currentY == targetY))
	      deltaAngle = 0; 
	   else
	      {
	     	if (currentX == targetX)
	     	  if (currentY < targetY)
	     	    deltaAngle = 90;
	     	  else
	     	    deltaAngle = -90;
	     	else
	     	  {
	     	  	 deltaAngle = Math.atan((targetY - currentY)/(targetX-currentX))/Math.PI*180;
	     	  	  	
	     	  	 if ((deltaAngle > 0) && (targetY < currentY))
	     	  	   deltaAngle -= 180;
	     	  	  	
	     	  	 if ((deltaAngle < 0) && (targetY > currentY))
	     	  	   deltaAngle += 180;
	     	  }     		     	  	
	      }	
	      
	   return deltaAngle;	
     }
     
     
    /*public static void printRequests (Vector requests)
	  {
	  	int i;
	  	for (i=0;i<requests.size();i++)
		  {
		  	System.out.print("aid = " + ((ActionRequest)requests.elementAt(i)).aid + "   ");
		  	
		  	System.out.print("deadline = " + ((ActionRequest)requests.elementAt(i)).deadline + " ");
		  	
		  	System.out.print("startTime = " + ((ActionRequest)requests.elementAt(i)).startTime + " ");
		  	
		  	System.out.print("didVector = (");
		  	for (int j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
		  	  System.out.print(((ActionRequest)requests.elementAt(i)).didVector.elementAt(j) + ",");
		  	System.out.print(")   ");
		  	
		  	System.out.print("costVector = (");
		  	for (int j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
		  	  System.out.print(((ActionRequest)requests.elementAt(i)).costVector.elementAt(j) + ",");
		  	System.out.print(")   ");		  	
		  	
		  	System.out.println();
          }           	
	  }*/
	  
	  
	/*public static void printRequestsStatus (Vector requests)
	  {
	  	int i;
	  	for (i=0;i<requests.size();i++)
		  {
		  	System.out.println("Request #" + (i+1) +".");
		  	
		  	for (int j=0;j<((ActionRequest)requests.elementAt(i)).statusVector.size();j++)
		    {
		      Tuple t = (Tuple)((ActionRequest)requests.elementAt(i)).statusVector.elementAt(j);
		      
		      for (int k=0;k< t.fields.size();k++)
		        System.out.print(t.fields.elementAt(k) + "   ");
		      
		      System.out.println();		      
		    } 
		  }			
	  } */	  
	
	
	/** 
	 * Start the scheduling simulator.
	 */
	public static void main (String argv [])  
	  {
	  	AortaSchedulingSimulator.schedule();
	  }	
} 