/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;
import java.net.*;
import java.awt.*;

/**
 * A simulator for the AXIS 2130 PTZ network cameras.
 *
 * @see ActionOperator
 * @see AortaSchedulingSimulator
 */
public class CameraSimulator {	
	
	/** ID of this simulated camera. */
	public String id;
	
	public static final String ipPrefix = "192.168.1.";
	/** IP of this simualted camera. */
	public String ip;
	
	String username;  // authorization information about this simulated camera
	String password;
	
	public static final float MIN_LOC = 1;  // unit: meter
	public static final float MAX_LOC = 3;
	/** Location of this simulated camera. */
	public Point loc;
	
	/* The pan, tilt, zoom range of the camera. */
	public static final int MAX_PAN = 169;
	public static final int MIN_PAN = -169;
	public static final int MAX_TILT = 10;
	public static final int MIN_TILT = -90;
	public static final int MAX_ZOOM = 9999;
	public static final int MIN_ZOOM = 1;
	/** Current head position (pan, tilt, zoom values) of this simulated camera. */
	public int pan, tilt, zoom;  
	
	/* The "home" position of a camera. */
	public static final int HOME_PAN = 0;
	public static final int HOME_TILT = 0;
	public static final int HOME_ZOOM = 1;		
	
   	public static final double STANDARD_POLLING_TIME = 110;
	public static final double POLLING_TIMEOUT_LIMIT = 10*STANDARD_POLLING_TIME;
	public static final double POLLING_TIMED_OUT = Double.MAX_VALUE;  // a constant that indicates the polling to a simulated camera has timed out
	public static final double TIMEOUT_PROBABILITY = 0.0f;  // the probability that the polling to a simulated camera times out
	/** Last polling time of this simulated camera. */
	public double pollingTime;
	
	/* The estimated costs of atomic operations */
	public static final double PAN_ESTIMATED_COST = 13;
	public static final double TILT_ESTIMATED_COST = 14;
	public static final double ZOOM_ESTIMATED_COST = 0.5;
	public static final double TAKE_SMALL_SIZE_PHOTO_ESTIMATED_COST = 200;
	public static final double TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST = 190;
	public static final double TAKE_LARGE_SIZE_PHOTO_ESTIMATED_COST = 330;	
	
	/* Example photos returned by the simulator */
	public static final String SMALL_IMAGE_FILE_EXAMPLE = "../photos/lib/photo_example_small.jpg";
	public static final String MEDIUM_IMAGE_FILE_EXAMPLE = "../photos/lib/photo_example_medium.jpg";
	public static final String LARGE_IMAGE_FILE_EXAMPLE = "../photos/lib/photo_example_large.jpg";
	
	/** Maxium number of real cameras we currently have. */	
	public static final int MAX_REAL_CAMERA_NUM = 4;  
	
	Random ran;
	
	
	/**
	 * Constructor of a simulated camera.
	 * 
	 * @param id the id of the camera
	 */
	public CameraSimulator (String id) 
      {
      	this.id = id;
      	ip = ipPrefix + id;
      	
      	ran = new Random(Integer.parseInt(id)); 
      	
      	username = "root";
      	password = "";
      	
      	pan = HOME_PAN; tilt = HOME_TILT; zoom = HOME_ZOOM;  // initialize the camera head to the "home" position
      	
      	loc = new Point ();
      	if (id.equals("1"))  // Camera 1
      	  {loc.x = -2; loc.y = 0.5; loc.z = 2;}
      	else if (id.equals("2"))
      	       {loc.x = 2; loc.y = 3.5; loc.z = 2;}
      	     else if (id.equals("3"))
      	            {loc.x = -2; loc.y = 3; loc.z = 1;}
      	          else if (id.equals("4"))
      	                 {loc.x = -2; loc.y = 2; loc.z = 0;}
      	               else generateRandomLocation();       	      	
      } 
       
	
	/** 
	 * Randomly generate the location of this simulated camera. 
	 * 
	 * @see #ranPan
	 * @see #ranTilt
	 * @see #ranZoom
	 */
	public void generateRandomLocation () 
	  {
	   	loc.x = MIN_LOC + Math.abs(ran.nextDouble()) % (MAX_LOC - MIN_LOC + 1);
	   	
	   	loc.y = MIN_LOC + Math.abs(ran.nextDouble()) % (MAX_LOC - MIN_LOC + 1);
	   	
	   	loc.z = MIN_LOC + Math.abs(ran.nextDouble()) % (MAX_LOC - MIN_LOC + 1);		
	  }
	
	
	/** Connect this simulated camera. */
	public double polling()
	  {
	  	int pb = Math.abs(ran.nextInt()) % 10;
	  	  	
	  	if (pb >= 10 - (int)(10*TIMEOUT_PROBABILITY))
	  	  {
	  	  	pollingTime = POLLING_TIMED_OUT;
	  	  	try {Thread.currentThread().sleep((long)POLLING_TIMEOUT_LIMIT);}
	  	    catch (Exception e) {e.printStackTrace();}
	  	  }
	  	else
	  	  {
	  	  	pollingTime = (STANDARD_POLLING_TIME - 10) + (100*Math.abs(ran.nextDouble())) % 60;
	  	  	try {Thread.currentThread().sleep((long)pollingTime);}
	     	catch (Exception e) {e.printStackTrace();}
	      }
	  	  
	  	return pollingTime;
	  }	
	
	
	/** 
	 * @param pan the current pan value of this simulated camera
	 * @param loc the target location 
	 *
	 * @return the <code>delta</code> pan required for moving the camera head
	 *
	 * @see #targetPan
	 */
	public int deltaPan (int pan, Point loc)
	  {
	  	return Math.abs(targetPan(loc) - pan);
	  }	  
	  
    /** 
	 * @param tilt the current pan value of this simulated camera
	 * @param loc the target location 
	 *
	 * @return the <code>delta</code> tilt required for moving the camera head
	 *
	 * @see #targetTilt 
	 */
	public int deltaTilt (int tilt, Point loc)
	  {
	  	return Math.abs(targetTilt(loc) - tilt);
	  }
	  
	/** 
	 * @param zoom the current pan value of this simulated camera
	 * @param loc the target location 
	 *
	 * @return the <code>delta</code> zoom required for moving the camera head
	 *
	 * @see #targetZoom
	 */
	public int deltaZoom (int zoom, Point loc)
	  {
	  	return Math.abs(targetZoom(loc) - zoom);
	  }	 
	  
	public int targetPan (Point loc)
	  {
	  	if (loc.x == 0 && loc.y == 0 && loc.z == 0)  // Location of Sensor 1
	  	  if (id.equals("1"))  // Camera 1
	  	      return 45;  // system-tuned preset values, the same as follows
	  	    else if (id.equals("2"))  // Camera 2
	  	           return -37;
	  	         else if (id.equals("3"))  // Camera 3
	  	                return 52;
	  	              else if (id.equals("4"))  // Camera 4
	  	                      return -30;
	  	                   else  // other simualted cameras
	  	                     return ranPan();
	  	                     
	  	if (loc.x == 0 && loc.y == 6 && loc.z == 1)  // Location of Sensor 2
	  	  if (id.equals("1"))  
	  	      return -22;  
	  	    else if (id.equals("2"))  
	  	           return 117;
	  	         else if (id.equals("3"))  
	  	                return -61;
	  	              else if (id.equals("4")) 
	  	                      return Integer.MAX_VALUE; // the view range of the camera can not cover the location
	  	                   else 
	  	                     return ranPan(); 
	  	 
	  	if (loc.x == 1.5 && loc.y == 3 && loc.z == 0)  // Location of Sensor 3
	  	  if (id.equals("1"))  
	  	      return 4;  
	  	    else if (id.equals("2"))  
	  	           return -83;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return -70;
	  	                   else  
	  	                     return ranPan(); 
	  	                     
	  	if (loc.x == 1.5 && loc.y == 0.5 && loc.z == 0)  // Location of Sensor 4
	  	  if (id.equals("1"))  
	  	      return 31;  
	  	    else if (id.equals("2"))  
	  	           return Integer.MAX_VALUE;
	  	         else if (id.equals("3")) 
	  	                return 52;
	  	              else if (id.equals("4"))  
	  	                      return -45;
	  	                   else  
	  	                     return ranPan(); 
	  	                     
	  	if (loc.x == -2.5 && loc.y == 3 && loc.z == 0)  // Location of Sensor 5
	  	  if (id.equals("1"))  
	  	      return Integer.MAX_VALUE;  
	  	    else if (id.equals("2"))  
	  	           return 32;
	  	         else if (id.equals("3")) 
	  	                return 92;
	  	              else if (id.equals("4"))  
	  	                      return 9;
	  	                   else  
	  	                     return ranPan(); 
	  	                     
	  	if (loc.x == 1.5 && loc.y == 3 && loc.z == -1)  // Location of Sensor 6
	  	  if (id.equals("1"))  
	  	      return 6;  
	  	    else if (id.equals("2"))  
	  	           return -78;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return -70;
	  	                   else  
	  	                     return ranPan(); 
	  	                     
	  	if (loc.x == 1.5 && loc.y == 0.5 && loc.z == -1)  // Location of Sensor 7
	  	  if (id.equals("1"))  
	  	      return 33;  
	  	    else if (id.equals("2"))  
	  	           return Integer.MAX_VALUE;
	  	         else if (id.equals("3")) 
	  	                return 52;
	  	              else if (id.equals("4"))  
	  	                      return -45;
	  	                   else  
	  	                     return ranPan();  
	  	                     
	  	if (loc.x == -2.5 && loc.y == 3 && loc.z == -1)  // Location of Sensor 8
	  	  if (id.equals("1"))  
	  	      return Integer.MAX_VALUE;  
	  	    else if (id.equals("2"))  
	  	           return 33;
	  	         else if (id.equals("3")) 
	  	                return 90;
	  	              else if (id.equals("4"))  
	  	                      return 10;
	  	                   else  
	  	                     return ranPan(); 
	  	                     
	  	if (loc.x == -2 && loc.y == 0.5 && loc.z == 0)  // Location of Sensor 9
	  	  if (id.equals("1"))  
	  	      return -58;  
	  	    else if (id.equals("2"))  
	  	           return 4;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return Integer.MAX_VALUE;
	  	                   else  
	  	                     return ranPan(); 
	  	                     
	  	if (loc.x == 1.5 && loc.y == 5 && loc.z == 0.5)  // Location of Sensor 10
	  	  if (id.equals("1"))  
	  	      return -14;  
	  	    else if (id.equals("2"))  
	  	           return 134;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return 10;
	  	                   else  
	  	                     return ranPan(); 	  	                           
	  	                     
	  	 return Integer.MAX_VALUE;  	 	  	  
	  }
	  
	public int targetTilt (Point loc)
	  {
	  	if (loc.x == 0 && loc.y == 0 && loc.z == 0)  // Location of Sensor 1
	  	  if (id.equals("1"))  
	  	      return -19;
	  	    else if (id.equals("2")) 
	  	           return -14;
	  	         else if (id.equals("3"))  
	  	                return 4;
	  	              else if (id.equals("4"))  
	  	                      return -8;
	  	                   else  
	  	                     return ranTilt();
	  	
	  	if (loc.x == 0 && loc.y == 6 && loc.z == 1)  // Location of Sensor 2
	  	  if (id.equals("1")) 
	  	      return -8;  
	  	    else if (id.equals("2")) 
	  	           return -8;
	  	         else if (id.equals("3"))  
	  	                return 9;
	  	              else if (id.equals("4"))  
	  	                      return Integer.MAX_VALUE;
	  	                   else  
	  	                     return ranTilt(); 
	  	                     
	  	if (loc.x == 1.5 && loc.y == 3 && loc.z == 0)  // Location of Sensor 3
	  	  if (id.equals("1"))  
	  	      return -10;  
	  	    else if (id.equals("2"))  
	  	           return -49;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return -12;
	  	                   else  
	  	                     return ranTilt();
	  	                     
	  	if (loc.x == 1.5 && loc.y == 0.5 && loc.z == 0)  // Location of Sensor 4
	  	  if (id.equals("1"))  
	  	      return -17;  
	  	    else if (id.equals("2"))  
	  	           return Integer.MAX_VALUE;
	  	         else if (id.equals("3")) 
	  	                return 4;
	  	              else if (id.equals("4"))  
	  	                      return -12;
	  	                   else  
	  	                     return ranTilt();
	  	                              
	  	if (loc.x == -2.5 && loc.y == 3 && loc.z == 0)  // Location of Sensor 5
	  	  if (id.equals("1"))  
	  	      return Integer.MAX_VALUE;  
	  	    else if (id.equals("2"))  
	  	           return -17;
	  	         else if (id.equals("3")) 
	  	                return 13;
	  	              else if (id.equals("4"))  
	  	                      return -30;
	  	                   else  
	  	                     return ranTilt(); 
	  	                     
	    if (loc.x == 1.5 && loc.y == 3 && loc.z == -1)  // Location of Sensor 6
	  	  if (id.equals("1"))  
	  	      return -17;  
	  	    else if (id.equals("2"))  
	  	           return -53;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return -20;
	  	                   else  
	  	                     return ranTilt(); 
	  	                     
	  	if (loc.x == 1.5 && loc.y == 0.5 && loc.z == -1)  // Location of Sensor 7
	  	  if (id.equals("1"))  
	  	      return -21;  
	  	    else if (id.equals("2"))  
	  	           return Integer.MAX_VALUE;
	  	         else if (id.equals("3")) 
	  	                return -5;
	  	              else if (id.equals("4"))  
	  	                      return -20;
	  	                   else  
	  	                     return ranTilt();  
	  	                     
	  	if (loc.x == -2.5 && loc.y == 3 && loc.z == -1)  // Location of Sensor 8
	  	  if (id.equals("1"))  
	  	      return Integer.MAX_VALUE;  
	  	    else if (id.equals("2"))  
	  	           return -23;
	  	         else if (id.equals("3")) 
	  	                return -5;
	  	              else if (id.equals("4"))  
	  	                      return -40;
	  	                   else  
	  	                     return ranTilt(); 
	  	                     
	  	if (loc.x == -2 && loc.y == 0.5 && loc.z == 0)  // Location of Sensor 9
	  	  if (id.equals("1"))  
	  	      return -48;  
	  	    else if (id.equals("2"))  
	  	           return -14;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return Integer.MAX_VALUE;
	  	                   else  
	  	                     return ranTilt(); 
	  	                     
	  	if (loc.x == 1.5 && loc.y == 5 && loc.z == 0.5)  // Location of Sensor 10
	  	  if (id.equals("1"))  
	  	      return -5;  
	  	    else if (id.equals("2"))  
	  	           return -24;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return -5;
	  	                   else  
	  	                     return ranTilt();    	                     
	  	
	  	return Integer.MAX_VALUE;	  	                     
	  }
	  
	public int targetZoom (Point loc)
	  {
	  	if (loc.x == 0 && loc.y == 0 && loc.z == 0)  // Location of Sensor 1
	  	  if (id.equals("1"))  
	  	      return 2501;
	  	    else if (id.equals("2")) 
	  	           return 5001;
	  	         else if (id.equals("3")) 
	  	                return 5001;
	  	              else if (id.equals("4"))  
	  	                      return 7501;
	  	                   else
	  	                     return ranZoom();
	  	 
	  	 if (loc.x == 0 && loc.y == 6 && loc.z == 1)  // Location of Sensor 2
	  	  if (id.equals("1"))  
	  	      return 2501;  
	  	    else if (id.equals("2"))  
	  	           return 2501;
	  	         else if (id.equals("3")) 
	  	                return 1;
	  	              else if (id.equals("4"))  
	  	                      return Integer.MAX_VALUE;
	  	                   else  
	  	                     return ranZoom();
	  	 
	  	 if (loc.x == 1.5 && loc.y == 3 && loc.z == 0)  // Location of Sensor 3
	  	  if (id.equals("1"))  
	  	      return 5001;  
	  	    else if (id.equals("2"))  
	  	           return 1;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return 2501;
	  	                   else  
	  	                     return ranZoom();  
	  	                     
	  	 if (loc.x == 1.5 && loc.y == 0.5 && loc.z == 0)  // Location of Sensor 4
	  	  if (id.equals("1"))  
	  	      return 2501;  
	  	    else if (id.equals("2"))  
	  	           return Integer.MAX_VALUE;
	  	         else if (id.equals("3")) 
	  	                return 1;
	  	              else if (id.equals("4"))  
	  	                      return 5001;
	  	                   else  
	  	                     return ranZoom();  
	  	                     
	  	 if (loc.x == -2.5 && loc.y == 3 && loc.z == 0)  // Location of Sensor 5
	  	  if (id.equals("1"))  
	  	      return Integer.MAX_VALUE;  
	  	    else if (id.equals("2"))  
	  	           return 2501;
	  	         else if (id.equals("3")) 
	  	                return 1;
	  	              else if (id.equals("4"))  
	  	                      return 1;
	  	                   else  
	  	                     return ranZoom();
	  	                     
	  	if (loc.x == 1.5 && loc.y == 3 && loc.z == -1)  // Location of Sensor 6
	  	  if (id.equals("1"))  
	  	      return 7501;  
	  	    else if (id.equals("2"))  
	  	           return 5001;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return 5001;
	  	                   else  
	  	                     return ranZoom(); 
	  	                     
	  	if (loc.x == 1.5 && loc.y == 0.5 && loc.z == -1)  // Location of Sensor 7
	  	  if (id.equals("1"))  
	  	      return 7501;  
	  	    else if (id.equals("2"))  
	  	           return Integer.MAX_VALUE;
	  	         else if (id.equals("3")) 
	  	                return 5001;
	  	              else if (id.equals("4"))  
	  	                      return 7501;
	  	                   else  
	  	                     return ranZoom();  
	  	                     
	  	if (loc.x == -2.5 && loc.y == 3 && loc.z == -1)  // Location of Sensor 8
	  	  if (id.equals("1"))  
	  	      return Integer.MAX_VALUE;  
	  	    else if (id.equals("2"))  
	  	           return 7501;
	  	         else if (id.equals("3")) 
	  	                return 2501;
	  	              else if (id.equals("4"))  
	  	                      return 2501;
	  	                   else  
	  	                     return ranZoom(); 
	  	                     
	  	if (loc.x == -2 && loc.y == 0.5 && loc.z == 0)  // Location of Sensor 9
	  	  if (id.equals("1"))  
	  	      return 1;  
	  	    else if (id.equals("2"))  
	  	           return 5001;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return Integer.MAX_VALUE;
	  	                   else  
	  	                     return ranZoom(); 
	  	                     
	  	if (loc.x == 1.5 && loc.y == 5 && loc.z == 0.5)  // Location of Sensor 10
	  	  if (id.equals("1"))  
	  	      return 5001;  
	  	    else if (id.equals("2"))  
	  	           return 1;
	  	         else if (id.equals("3")) 
	  	                return Integer.MAX_VALUE;
	  	              else if (id.equals("4"))  
	  	                      return 1;
	  	                   else  
	  	                     return ranZoom();       
	  	                     
	  	return Integer.MAX_VALUE;  	  	                     
	  }
	  
	public int ranPan()
	  {
	  	return MIN_PAN + Math.abs(ran.nextInt()) % (MAX_PAN - MIN_PAN + 1);
	  }
	  
	public int ranTilt()
	  {
	  	return MIN_TILT + Math.abs(ran.nextInt()) % (MAX_TILT - MIN_TILT + 1);
	  }
	
	public int ranZoom ()
	  {
	  	int num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num > 1) 
	  	  num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num > 2)
	  	  num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num == 0) return 1;
	  	else if (num == 1) return 2501; 
	         else if (num == 2) return 5001;
	              else return 9999;	  	
	  }
	  
	
	/** 
	 * Move the head of this this simulated to a target position.
	 *
	 * @param pan the target pan of the camera head 
	 * @param tilt the target tilt of the camera head 
	 * @param zoom the target zoom of the camera head 
	 *
	 * @see #computeHeadMovementCost
	 */
	public void headMovement(int pan, int tilt, int zoom)
	  {
	  	int deltaPan = Math.abs(pan - this.pan);
	  	int deltaTilt = Math.abs(tilt - this.tilt);
	  	int deltaZoom = Math.abs(zoom - this.zoom);
	  	
	  	double headMovementCost = computeHeadMovementCost(deltaPan, deltaTilt, deltaZoom);
	  	
	  	try {Thread.currentThread().sleep((long)headMovementCost);}
	  	catch (Exception e) {e.printStackTrace();}
	  	
	  	this.pan = pan; this.tilt = tilt; this.zoom = zoom;
	  }
	  
   public double computeHeadMovementCost (int deltaPan, int deltaTilt, int deltaZoom)
     {
        //return Math.max(Math.max(Math.abs(deltaPan) * PAN_ESTIMATED_COST, Math.abs(deltaTilt) * TILT_ESTIMATED_COST), Math.abs(deltaZoom) * ZOOM_ESTIMATED_COST);           
        
        return Math.abs(deltaPan) * PAN_ESTIMATED_COST + Math.abs(deltaTilt) * TILT_ESTIMATED_COST + Math.abs(deltaZoom) * ZOOM_ESTIMATED_COST;
	 }
	 
	 
   /** Take a small-size photo on this simulated camera. */
   public Image takeSmallSizePhoto()
	  {
	  	long startTime, endTime;
	  	try 
	  	  {
	  	  	startTime = System.currentTimeMillis();
	  	  	Toolkit tk = Toolkit.getDefaultToolkit();
	  	  	Image image = tk.createImage(SMALL_IMAGE_FILE_EXAMPLE);
	  	  	endTime = System.currentTimeMillis();
	  	  	Thread.currentThread().sleep((long)TAKE_SMALL_SIZE_PHOTO_ESTIMATED_COST - (endTime - startTime));
	  	  	return image;	  	  	
	  	  }	  	  	
	  	catch (Exception e) {e.printStackTrace(); return null;}
	  }
	  
	/** Take a medium-size photo on this simulated camera. */
	public Image takeMediumSizePhoto()
	  {
	  	long startTime, endTime;
	  	try 
	  	  {
	  	  	startTime = System.currentTimeMillis();
	  	  	Toolkit tk = Toolkit.getDefaultToolkit();
	  	  	Image image = tk.createImage(MEDIUM_IMAGE_FILE_EXAMPLE);
	  	  	endTime = System.currentTimeMillis();
	  	  	Thread.currentThread().sleep((long)TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST - (endTime - startTime));
	  	  	return image;	  	  	
	  	  }	  	  	
	  	catch (Exception e) {e.printStackTrace(); return null;}
	  }
	  
    /** Take a large-size photo on this simulated camera. */
    public Image takeLargeSizePhoto()
	  {
	  	long startTime, endTime;
	  	try 
	  	  {
	  	  	startTime = System.currentTimeMillis();
	  	  	Toolkit tk = Toolkit.getDefaultToolkit();
	  	  	Image image = tk.createImage(LARGE_IMAGE_FILE_EXAMPLE);
	  	  	endTime = System.currentTimeMillis();
	  	  	Thread.currentThread().sleep((long)TAKE_LARGE_SIZE_PHOTO_ESTIMATED_COST - (endTime - startTime));
	  	  	return image;	  	  	
	  	  }	  	  	
	  	catch (Exception e) {e.printStackTrace(); return null;}	
	  }	 
} 