/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Implementation of the Dynamic, Heuristic Action Scheduling Algorithm in Aorta (DHASA).
 * The algorithm is developed under the list scheduling (LS) discipline with a new 
 * scheduling heuristic.
 * 
 * @see AortaSchedulingSimulator#schedule
 * @see SchedulingThread 
 */ 
public class DHASA {
 	
 	Vector requests;
	
	PrintWriter logFileWriter;
	
	double currentSystemTime = 0;	
	
	double schedulingTime = 0;
	
	double totalRequestCompletionTime = 0;
	
	double droppedRequestNum = 0;
	
	SchedulingDevice schdevs [] = new SchedulingDevice [AortaSchedulingSimulator.deviceNum];
		
	ActionRequest sel_ar = null;
	String sel_did = null;
	double sel_weight = Double.MAX_VALUE, sel_actionCost = Double.MAX_VALUE;
	Tuple sel_status = null;
 	
 	Random failRan = new Random();
 	
 	
 	/** 
	 * @param requests a set of action requests to be scheduled
	 * @param logFileWriter a {@link PrintWriter} corresponds to the log file
	 */
	 public DHASA (Vector requests, PrintWriter logFileWriter) 
 	  {
 		this.requests = requests;
		
		this.logFileWriter = logFileWriter;  
 	  }  
 	     	  
 	  
 	/** 
 	 * Assign and execute requests on devices using one of the following three heuristics:
 	 * (1) weighted sum, (2) multiplication, and (3) priority ordering of the factors involved
 	 * in the scheduling.
 	 *
 	 **/
 	public void schedule()
 	  {
 		int i,j;
 		int count = 0, id;
 		long startTime, endTime;
 		double pollingTime, angle, actionCost;
 		double minStartTime, minNextFreeTime;
 		Point loc = new Point();
 		ActionRequest ar;	
 		
 		if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
 		  {
 		  	pollingTime = 0;
 		    
 		    for (i=0;i<schdevs.length;i++)
 		      if (pollingTime < schdevs[i].camSim.polling())
 		        pollingTime = schdevs[i].camSim.pollingTime; 
 		   
 		    schedulingTime += pollingTime / 1000.0;
 		  }
 		else
 		  schedulingTime += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;		
 		
 		currentSystemTime += schedulingTime;  
 		 		
 		   		
 		while (requests.size() > 0)
 		  {
 		  	// decide the set of requests that are eligible to be scheduled currently
 		  	Vector currentRequests = new Vector (requests.size(),10);
 		  	
 		  	for (i=0;i<requests.size();i++)
 		  	  {
 		  	  	ar = (ActionRequest)requests.elementAt(i);
 		  	  	
 		  	  	if (ar.startTime <= currentSystemTime)
 		  	  	  currentRequests.addElement((ActionRequest)ar);
 		  	  }
 		  	  		  	
 		  			  	
            // filter out those requests whose deadlines have been missed on all candidate devices
            startTime = System.currentTimeMillis();
            
            Vector removedRequests = new Vector (currentRequests.size(),10);
            for (i=0;i<currentRequests.size();i++)
              {
              	ar = (ActionRequest)currentRequests.elementAt(i);
              	
              	boolean outdated = true;
              	
              	for (j=0;j<ar.costVector.size();j++)
              	  {
              	  	double cost = Double.parseDouble((String)ar.costVector.elementAt(j));
              	  	
              	  	if (currentSystemTime + cost <= ar.startTime + ar.deadline)
              	  	  {outdated = false; break;}            	  	
              	  }
              	
              	if (outdated) removedRequests.addElement((ActionRequest)ar);
              }            
           
            for (i=0;i<removedRequests.size();i++)
              {
                droppedRequestNum ++;
                
                ar = (ActionRequest)removedRequests.elementAt(i);
                
                currentRequests.removeElement(ar);
                
                requests.removeElement(ar);
                
                logFileWriter.println("#");	
                logFileWriter.println("Round #" + (++count) + ": Request #" + ar.aid +  " is dropped.");
                logFileWriter.println();
 		  	    logFileWriter.println("Request start time: " +  ar.startTime + ".");
 		  	    logFileWriter.println("Request deadline: " + ar.deadline + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println("Current system time: " + currentSystemTime + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Candidate device ids of the request: ");
 		  	    for (j=0;j<ar.didVector.size();j++)
 		  	      logFileWriter.print((String)ar.didVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Costs of the request on candidate devices: ");
 		  	    for (j=0;j<ar.costVector.size();j++)
 		  	      logFileWriter.print((String)ar.costVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
              }            
              
            endTime = System.currentTimeMillis();
 		  	
 		  	schedulingTime += (endTime - startTime) / 1000.0;
 		  	
 		  	currentSystemTime += (endTime - startTime) / 1000.0;  	
 		  
 		  	
 		  	// select one request to be scheduled
 		  	if (currentRequests.size() == 0)
 		  	  {
 		  	  	if (requests.size() == 0)
 		  	  	  break;
 		  	  	
 		  	  	minStartTime = ((ActionRequest)requests.elementAt(0)).startTime;
 		  	  	
 		  	  	if (currentSystemTime < minStartTime)
 		  	  	  {
 		  	  	  	currentSystemTime = minStartTime;
 		  	  	  	
 		  	  	  	for (i=0;i<schdevs.length;i++)
 		  	          if (schdevs[i].nextFreeTime <= currentSystemTime)
 		  	            schdevs[i].busy = false; 		  	  	  	  
 		  	  	  }
 		  	  	
 		  	  	continue; 		  	  	
 		  	  }
 		  	  
 		  	startTime = System.currentTimeMillis(); 		  	 		  	
 		  	
 		  	heuristicRequestSelection(currentRequests);
 		  	
 		  	endTime = System.currentTimeMillis();
 		  	
 		  	schedulingTime += (endTime - startTime) / 1000.0;
 		  	
 		  	currentSystemTime += (endTime - startTime) / 1000.0; 		  	
 		  	
 		  	if (sel_ar == null)  // there is no request to be scheduled now
 		  	  {
 		  	  	if (requests.size() == 0)
 		  	  	  break;
 		  	  	  
 		  	  	minNextFreeTime = Double.MAX_VALUE; 		  	
 		  	    for (i=0;i<schdevs.length;i++)
 		  	      if ((minNextFreeTime > schdevs[i].nextFreeTime)&&(schdevs[i].busy==true))
 		  	        minNextFreeTime = schdevs[i].nextFreeTime;
 		  	        
 		  	    minStartTime = Double.MAX_VALUE;
 		  	  	for (i=0;i<requests.size();i++)
 		  	  	  {
 		  	  	  	ar = (ActionRequest)requests.elementAt(i);
 		  	  	  	
 		  	  	  	if (ar.startTime > currentSystemTime)
 		  	  	  	  {minStartTime = ar.startTime; break;}
 		  	  	  }
 		  	    
 		  	    currentSystemTime = Math.min(minNextFreeTime, minStartTime);
 		  	      	  	
 		  	  	for (i=0;i<schdevs.length;i++)
 		  	      if (schdevs[i].nextFreeTime <= currentSystemTime)
 		  	        schdevs[i].busy = false;
 		  	        
 		  	    /*logFileWriter.println(minStartTime + " " + minNextFreeTime + " " + currentSystemTime);
 		  	    for (i=0;i<schdevs.length;i++)
 		  	      logFileWriter.print(schdevs[i].busy + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println(currentRequests.size());*/
 		  	      	
 		  	  	continue;
 		  	  } 		  	
 		  	
 		  	
 		  	// service the selected requests on the device
 		  	SchedulingDevice sd = schdevs [id = Integer.parseInt(sel_did)-1];
 		  	sd.busy = true;
 		  	sd.nextFreeTime = currentSystemTime + sel_actionCost;
 		  	
 		  	totalRequestCompletionTime += sel_actionCost;  		  			  	
 		  	
 		  	logFileWriter.println("#");  
 		  	if (failRan.nextDouble() > sd.failProb)
 		  	  {
 		  	  	logFileWriter.println("Round #" + (++count) + ": Request #" + sel_ar.aid +  " is serviced on Device " + sel_did + " from system time " + currentSystemTime + " to " + (currentSystemTime + sel_actionCost) + ".");
 		  	    logFileWriter.println("Next free time of the device: " + sd.nextFreeTime);
 		  	    logFileWriter.println();
 		  	    logFileWriter.println("Request start time: " +  sel_ar.startTime + ".");
 		  	    logFileWriter.println("Request deadline: " + sel_ar.deadline + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println("Current system time: " + currentSystemTime + ".");
 		  	    logFileWriter.println("Action cost: " + sel_actionCost + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Candidate device ids of the request: ");
 		  	    for (j=0;j<sel_ar.didVector.size();j++)
 		  	      logFileWriter.print((String)sel_ar.didVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Costs of the request on candidate devices: ");
 		  	    for (j=0;j<sel_ar.costVector.size();j++)
 		  	      logFileWriter.print((String)sel_ar.costVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println();
 		  	     		  	    
 		  	     		  	    
 		  	    if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
 		  	  	  {
 		  	  		schedulingTime += sd.camSim.polling() / 1000.0;
 		  	  	
 		  	  		currentSystemTime += sd.camSim.pollingTime / 1000.0;
 		  	  	
 		  	  		logFileWriter.println("Old current physical status of Device " + sd.id + ": pan = " + sd.camSim.pan + "  tilt = " + sd.camSim.tilt + "  zoom = " + sd.camSim.zoom + ".");
 		   		  	  	
 		  	  		sd.camSim.pan = Integer.parseInt((String)sel_status.fields.elementAt(0));
 		  	    
 		  	   		sd.camSim.tilt = Integer.parseInt((String)sel_status.fields.elementAt(1));
 		  	    
 		  	    	sd.camSim.zoom = Integer.parseInt((String)sel_status.fields.elementAt(2));
 		  	    
 		  	    	logFileWriter.println("New current physical status of Device " + sd.id + ": pan = " + sd.camSim.pan + "  tilt = " + sd.camSim.tilt + "  zoom = " + sd.camSim.zoom + ".");
 		  	  	  }
 		  		else
 		  	  	  {
 		  	  		schedulingTime += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;
 		  	  	
 		  	  		currentSystemTime += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;
 		  	  	
 		  	  		logFileWriter.println("Old current physical status of Device " + sd.id + ": loc = (" + sd.er1Sim.loc.x + "," + sd.er1Sim.loc.y + "," + sd.er1Sim.loc.z + ") " + "angle = " + sd.er1Sim.angle + ".");
            	 
 		  	   		loc.x = Double.parseDouble((String)sel_status.fields.elementAt(0));
    	   	    
    	   	   		loc.y = Double.parseDouble((String)sel_status.fields.elementAt(1));
    	   	    
    	   	    	loc.z = Double.parseDouble((String)sel_status.fields.elementAt(2));
    	   	    
    	   	    	angle = AortaSchedulingSimulator.computeDeltaAngle(sd.er1Sim.loc,loc);
    	   	    
    	   	    	sd.er1Sim.setLoc(loc.x,loc.y,loc.z);
    	   	    
    	   	    	sd.er1Sim.setAngle(angle);
    	   	    
    	   	    	logFileWriter.println("New current physical status of Device " + sd.id + ": loc = (" + sd.er1Sim.loc.x + "," + sd.er1Sim.loc.y + "," + sd.er1Sim.loc.z + ") " + "angle = " + sd.er1Sim.angle + "."); 
 		  	  	  }            		  	
 		  	  }
 		  	else
 		  	  {
 		  	  	/*System.out.println(sd.failProb);
 		  	  	System.exit(1);*/
 		  	  	
 		  	  	droppedRequestNum ++;
 		  	  	
 		  	  	logFileWriter.println("Round #" + (++count) + ": Request #" + sel_ar.aid +  " is scheduled on Device " + sel_did + ". The action execution lasts from system time " + currentSystemTime + " to " + (currentSystemTime + sel_actionCost) + " but fails to due to temporary device malfunction.");
 		  	    logFileWriter.println("Next free time of the device: " + sd.nextFreeTime);
 		  	    logFileWriter.println();
 		  	    logFileWriter.println("Request start time: " +  sel_ar.startTime + ".");
 		  	    logFileWriter.println("Request deadline: " + sel_ar.deadline + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println("Current system time: " + currentSystemTime + ".");
 		  	    logFileWriter.println("Action cost: " + sel_actionCost + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Candidate device ids of the request: ");
 		  	    for (j=0;j<sel_ar.didVector.size();j++)
 		  	      logFileWriter.print((String)sel_ar.didVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Costs of the request on candidate devices: ");
 		  	    for (j=0;j<sel_ar.costVector.size();j++)
 		  	      logFileWriter.print((String)sel_ar.costVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println();
 		  	    
 		  	    if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
 		  	  	  {
 		  	  		schedulingTime += sd.camSim.polling() / 1000.0;
 		  	  	
 		  	  		currentSystemTime += sd.camSim.pollingTime / 1000.0;
 		  	  	
 		  	  		logFileWriter.println("Old current physical status of Device " + sd.id + ": pan = " + sd.camSim.pan + "  tilt = " + sd.camSim.tilt + "  zoom = " + sd.camSim.zoom + ".");
 		   		  	  	
 		  	  		int originPan = sd.camSim.pan;
 		  	    	
 		  	    	int originTilt = sd.camSim.tilt;
 		  	    	
 		  	    	int originZoom = sd.camSim.zoom;
 		  	    	
 		  	    	int targetPan = Integer.parseInt((String)sel_status.fields.elementAt(0));
 		  	    
 		  	   		int targetTilt = Integer.parseInt((String)sel_status.fields.elementAt(1));
 		  	    
 		  	    	int targetZoom = Integer.parseInt((String)sel_status.fields.elementAt(2));
 		  	      	    	
 		  	    	int maxPan = Math.max(originPan, targetPan), minPan = Math.min(originPan, targetPan);
 		  	    	
 		  	    	int maxTilt = Math.max(originTilt, targetTilt), minTilt = Math.min(originTilt, targetTilt);
 		  	    	
 		  	    	int maxZoom = Math.max(originZoom, targetZoom), minZoom = Math.min(originZoom, targetZoom);
 		  	    	 		  	    	
 		  	    	sd.camSim.pan = minPan + Math.abs(failRan.nextInt())%(maxPan - minPan + 1);
 		  	    	
 		  	    	sd.camSim.tilt = minTilt + Math.abs(failRan.nextInt())%(maxTilt - minTilt + 1);
 		  	    	
 		  	    	sd.camSim.zoom = minZoom + Math.abs(failRan.nextInt())%(maxZoom - minZoom + 1);
 		  	    	
 		  	    	if (sd.camSim.zoom > 1 && sd.camSim.zoom < 2501)
 		  	    	  sd.camSim.zoom = 1;
 		  	    	else
 		  	    	  if (sd.camSim.zoom > 2501 && sd.camSim.zoom < 5001)
 		  	    	    sd.camSim.zoom = 2501;
 		  	    	  else
 		  	    	    if (sd.camSim.zoom > 5001 && sd.camSim.zoom < 7501)
 		  	    	      sd.camSim.zoom = 5001;
 		  	    	    else
 		  	    	      if (sd.camSim.zoom > 7501 && sd.camSim.zoom < 9999)
 		  	    	        sd.camSim.zoom = 7501;
 		  	    	      else
 		  	    	        sd.camSim.zoom = 9999;
 		  	    	
 		  	    	logFileWriter.println("New current physical status of Device " + sd.id + ": pan = " + sd.camSim.pan + "  tilt = " + sd.camSim.tilt + "  zoom = " + sd.camSim.zoom + ".");
 		  	  	  }
 		  		else
 		  	  	  {
 		  	  		schedulingTime += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;
 		  	  	
 		  	  		currentSystemTime += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;
 		  	  	
 		  	  		logFileWriter.println("Old current physical status of Device " + sd.id + ": loc = (" + sd.er1Sim.loc.x + "," + sd.er1Sim.loc.y + "," + sd.er1Sim.loc.z + ") " + "angle = " + sd.er1Sim.angle + ".");
            	 
 		  	   		double originLocX = sd.er1Sim.loc.x;
 		  	   		
 		  	   		double originLocY = sd.er1Sim.loc.y;

                    double originAngle = sd.er1Sim.angle;
 		  	   		
 		  	   		double targetLocX = Double.parseDouble((String)sel_status.fields.elementAt(0));
    	   	    
    	   	   		double targetLocY = Double.parseDouble((String)sel_status.fields.elementAt(1));
    	   	    
    	   	    	double targetLocZ = Double.parseDouble((String)sel_status.fields.elementAt(2));
    	   	    
    	   	    	double targetAngle = AortaSchedulingSimulator.computeDeltaAngle(sd.er1Sim.loc,loc);
    	   	    
    	   	    	double maxLocX = Math.max(originLocX, targetLocX), minLocX = Math.min(originLocX, targetLocX);
    	   	    	
    	   	    	double maxLocY = Math.max(originLocY, targetLocY), minLocY = Math.min(originLocY, targetLocY);
    	   	    	
    	   	    	double maxAngle = Math.max(originAngle, targetAngle), minAngle = Math.min(originAngle, targetAngle);
    	   	    	    	   	    	
    	   	    	loc.x = minLocX + Math.abs(failRan.nextInt())%(maxLocX - minLocX + 1);
    	   	    	
    	   	    	loc.y = minLocY + Math.abs(failRan.nextInt())%(maxLocY - minLocY + 1);
    	   	    	
    	   	    	loc.z = targetLocZ;
    	   	    	
    	   	    	angle = minAngle + Math.abs(failRan.nextInt())%(maxAngle - minAngle + 1);
    	   	    	
    	   	    	sd.er1Sim.setLoc(loc.x,loc.y,loc.z);
    	   	    
    	   	    	sd.er1Sim.setAngle(angle);
    	   	    
    	   	    	logFileWriter.println("New current physical status of Device " + sd.id + ": loc = (" + sd.er1Sim.loc.x + "," + sd.er1Sim.loc.y + "," + sd.er1Sim.loc.z + ") " + "angle = " + sd.er1Sim.angle + "."); 
 		  	  	  } 		  	  	 		 	  	  	
 		  	  }  		  	 		  	
 		  
 		  	
 		  	// update request cost
 		  	startTime = System.currentTimeMillis();
 		  	
 		  	currentRequests.removeElement((ActionRequest)sel_ar);		  	
 		  	
 		  	if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
 		  	  AortaSchedulingSimulator.updatePhotoCosts(currentRequests, sel_did, schdevs);
 		  	else
 		  	  AortaSchedulingSimulator.updateWarnCosts(currentRequests, sel_did, schdevs); 		  	  
 		  	  
 		  	requests.removeElement((ActionRequest)sel_ar);
 		  	
 		  	endTime = System.currentTimeMillis();
 		  	
 		  	schedulingTime += (endTime - startTime) / 1000.0;
 		  	
 		  	currentSystemTime += (endTime - startTime) / 1000.0; 		  	
 		  	 		  	
 		  	
 		  	// go to schedule next request
 		  	minNextFreeTime = Double.MAX_VALUE;
 		  	
 		  	for (i=0;i<schdevs.length;i++)
 		  	  if (minNextFreeTime > schdevs[i].nextFreeTime)
 		  	    minNextFreeTime = schdevs[i].nextFreeTime; 		  	
 		  
 		  	if (minNextFreeTime > currentSystemTime)
 		  	  currentSystemTime = minNextFreeTime;
 		  	  
 		  	for (i=0;i<schdevs.length;i++)
 		  	  if (schdevs[i].nextFreeTime <= currentSystemTime)
 		  	    schdevs[i].busy = false;
 		  	    
 		  	sel_ar = null; sel_did = null; sel_status = null;
 		  	sel_weight = Double.MAX_VALUE; sel_actionCost = Double.MAX_VALUE;
 		  	 		  	
 		  	logFileWriter.flush(); 		   		  			  		     	
 		  }	
 		      
 	      
 	    try
 	      {
 	      	if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.WARN)
 	          for (i=0;i<schdevs.length;i++)
 	            schdevs[i].er1Sim.serverSocket.close();
 	       }
    	catch (Exception e) {e.printStackTrace();}
 		
 		
 		logFileWriter.println("#");
 		logFileWriter.println("Scheduling time: " + schedulingTime + ".");
 		logFileWriter.println("Number of dropped requests: " + droppedRequestNum + ".");
 		logFileWriter.println("Average request completion time: " + (totalRequestCompletionTime / (AortaSchedulingSimulator.requestNum - droppedRequestNum)) + ".");
 		logFileWriter.println("#"); 		
 	  }
 	   	
 	
 	/** Select a request to be scheduled on a free device with the least weight. */
 	public void heuristicRequestSelection(Vector currentRequests) 
 	  {
		int i,j;
		ActionRequest ar;
		
		for (i=0;i<schdevs.length;i++)
		  if (schdevs[i].busy == false)
		    {
		      heuristicRequestSelectionSingleDevice(schdevs[i], currentRequests);
		      
		      break;	
		    }
		    		
		    
		ActionRequest sel_ar_bak = sel_ar;
		String sel_did_bak = sel_did;	
	    double sel_weight_bak = sel_weight, sel_actionCost_bak = sel_actionCost;
		Tuple sel_status_bak = sel_status;
		
		for (j=i+1;j<schdevs.length;j++)
		  if (schdevs[j].busy==false)
		    {
		      heuristicRequestSelectionSingleDevice(schdevs[j], currentRequests);
		      
		      if (sel_ar == null) continue;
		      
		      if (sel_ar_bak==null)
		        {
		          sel_ar_bak = sel_ar;
		          
		          sel_did_bak = sel_did;
		          
		          sel_weight_bak = sel_weight;	
		        	
		          sel_actionCost_bak = sel_actionCost;
		          
		          sel_status_bak = sel_status;
		          
		          continue;
		        }	      
		      
		      /*if (AortaSchedulingSimulator.heuristic == AortaSchedulingSimulator.PRIORITY)
		        {
		          if (sel_ar.deadline < sel_ar_bak.deadline || 
		              sel_ar.deadline == sel_ar_bak.deadline && sel_actionCost < sel_actionCost_bak ||
		              sel_ar.deadline == sel_ar_bak.deadline && sel_actionCost == sel_actionCost_bak && sel_ar.didVector.size() < sel_ar_bak.didVector.size() ||
		              sel_ar.deadline == sel_ar_bak.deadline && sel_actionCost == sel_actionCost_bak && sel_ar.didVector.size() == sel_ar_bak.didVector.size() && sel_weight < sel_weight_bak) 
		            {
		              sel_ar_bak = sel_ar;
		          
		              sel_did_bak = sel_did;
		          
		              sel_weight_bak = sel_weight;
		          
		              sel_actionCost_bak = sel_actionCost;
		          
		              sel_status_bak = sel_status;
		            }		
		        }
		      else*/
		        if (sel_weight < sel_weight_bak)
		          {
		            sel_ar_bak = sel_ar;
		          
		            sel_did_bak = sel_did;
		          
		            sel_weight_bak = sel_weight;
		          
		            sel_actionCost_bak = sel_actionCost;
		          
		            sel_status_bak = sel_status;
		          }	
		    }
		    
		sel_ar = sel_ar_bak;
		          
	    sel_did = sel_did_bak;
		          
		sel_weight = sel_weight_bak;
		          
		sel_actionCost = sel_actionCost_bak;
		          
	    sel_status = sel_status_bak;   	 
      }
      
      
    public void heuristicRequestSelectionSingleDevice(SchedulingDevice sd, Vector currentRequests)
      {
      	int i,j;
      	ActionRequest ar;
      	
      	Vector currentRequestsSingleDevice = new Vector(currentRequests.size(), 10);
      	
      	for (i=0;i<currentRequests.size();i++)
      	  {
      	  	ar = (ActionRequest)currentRequests.elementAt(i);     	  	
      	  	
      	  	int did = 0;
      	  	for (j=0;j<ar.didVector.size();j++)
      	  	  {
      	  	  	did = Integer.parseInt((String)ar.didVector.elementAt(j));
      	  	  	
      	  	  	if (did == sd.id)
      	  	  	  break;
      	  	  }
      	  	
      	  	if (did == sd.id)
      	  	  currentRequestsSingleDevice.addElement((ActionRequest)ar);      	  	
      	  }	
      	        	
      	
      	if (currentRequestsSingleDevice.size()==0)
      	  {
      	  	sel_ar = null;
      	  	
      	  	sel_did = null;
      	  	
      	  	sel_weight = Double.MAX_VALUE;
      	  	
      	  	sel_actionCost = Double.MAX_VALUE;
      	  	
      	  	sel_status = null;
      	  	
      	  	return;
      	  }
      	  
      	double f = 1.0 + (currentRequestsSingleDevice.size() * 1.0) / (currentRequests.size() * 1.0);
      	//if (AortaSchedulingSimulator.heuristic == AortaSchedulingSimulator.WEIGHTED_SUM)
      	heuristicWeightedSum(sd, currentRequestsSingleDevice, f);
      	/*Ielse
      	 if (AortaSchedulingSimulator.heuristic == AortaSchedulingSimulator.MULTIPLICATION)
      	   heuristicMultiplication(sd, currentRequestsSingleDevice, f);
      	 else
      	   heuristicPriority(sd, currentRequestsSingleDevice, f);*/
      	   
        //System.out.println(sd.id + " " + sel_weight);   	      	
      }
      
    public void heuristicWeightedSum(SchedulingDevice sd, Vector requests, double f)
      {
      	double minDeadline = Double.MAX_VALUE, maxDeadline = Double.MIN_VALUE, avgDeadline = 0.0;
      	double minCost = Double.MAX_VALUE, maxCost = Double.MIN_VALUE, avgCost = 0.0;
      	double minCDN = Double.MAX_VALUE, maxCDN = Double.MIN_VALUE, avgCDN = 0.0;
      	//double minRT = Double.MAX_VALUE, maxRT = Double.MIN_VALUE, avgRT = 0.0;     	 
      	
      	int i,j;      	
      	double cost = 0.0, deadline = 0.0, cdn = 0.0; // rt = 0.0;
      	ActionRequest ar;
      	for (i=0;i<requests.size();i++)
      	  {
      	  	ar = (ActionRequest)requests.elementAt(i);
      	  	
      	  	if (minDeadline > ar.deadline)
      	  	  minDeadline = ar.deadline;
      	  	  
      	  	if (maxDeadline < ar.deadline)
      	  	  maxDeadline = ar.deadline;
      	  	  
      	    //minRT = minDeadline - currentSystemTime;
      	    
      	    //maxRT = maxDeadline - currentSystemTime;
      	  	  
      	  	avgDeadline += ar.deadline;
      	  	
      	  	//avgRT += Math.max(0, ar.deadline - currentSystemTime);
      	  	
      	  	for (j=0;j<ar.didVector.size();j++)
      	  	  if (Integer.parseInt((String)ar.didVector.elementAt(j)) == sd.id)
      	  	    {cost = Double.parseDouble((String)ar.costVector.elementAt(j)); break;}
      	  	
      	  	if (minCost > cost)
      	  	  minCost = cost;
      	  	
      	  	if (maxCost < cost)
      	  	  maxCost = cost;
      	  	  
      	  	avgCost += cost;
      	  	  
      	  	if (minCDN > ar.didVector.size())
      	  	  minCDN = ar.didVector.size();
      	  	
      	  	if (maxCDN < ar.didVector.size())
      	  	  maxCDN = ar.didVector.size();
      	  	  
      	  	avgCDN += ar.didVector.size();     	  	      	  	
      	  }
      	
      	avgDeadline /= requests.size();
      	
      	avgCost /= requests.size();
      	
      	avgCDN /= requests.size(); 
      	
      	//avgRT /= requests.size();
      	
      	/*if (minRT < 0)
      	  minRT = 0;
      	if (maxRT < 0)
      	  maxRT = 0;*/
      	        	      	        	
        
      	double minWeight = Double.MAX_VALUE;
      	double weight = 0.0;      	  	
      	for (i=0;i<requests.size();i++)
      	  {
      	  	ar = (ActionRequest)requests.elementAt(i); 	  	
      	  	
      	  	deadline = ar.deadline;
      	  	
      	  	//rt = Math.max(0, ar.deadline - currentSystemTime);
      	  	
      	  	for (j=0;j<ar.didVector.size();j++)
      	  	  if (Integer.parseInt((String)ar.didVector.elementAt(j)) == sd.id)
      	  	    {cost = Double.parseDouble((String)ar.costVector.elementAt(j)); break;}
      	  	    
      	  	cdn = ar.didVector.size();	
      	       	  	
      	  	
      	  	if (AortaSchedulingSimulator.basis == AortaSchedulingSimulator.DEADLINE_BASIS)
      	  	  {
      	  	  	weight = deadline;
      	  	    
      	  	    weight += cost*avgDeadline/avgCost;
      	  	    
      	  	    weight += cdn*avgDeadline/avgCDN;    	  	    
      	  	  }
      	  	else
      	  	  if (AortaSchedulingSimulator.basis == AortaSchedulingSimulator.COST_BASIS)
      	  	    {
      	  	      weight = cost;
      	  	      
      	  	      weight += deadline*avgCost/avgDeadline;
      	  	      
      	  	      weight += cdn*avgCost/avgCDN;  
      	  	      
      	  	      /*weight = cost;
      	  	    
      	  		  if (maxDeadline != minDeadline)
      	  	  	    weight += (deadline - minDeadline) / (maxDeadline - minDeadline) * (maxCost - minCost) + minCost;
      	  		  else
      	  	  	    weight += maxDeadline;     	  	
      	  	
      	  		  if (maxCDN != minCDN)
      	  	  	    weight += (cdn - minCDN) / (maxCDN - minCDN) * (maxCost - minCost) + minCost;
      	  		  else
      	  	        weight += maxCDN;*/	  	      
      	  	      
      	  	      /*weight = cost;
      	  	      
      	  	      if (avgRT != 0)
      	  	        weight += rt*avgCost/avgRT;
      	  	      
      	  	      weight += cdn*avgCost/avgCDN;*/
      	  	      
      	  	      /*weight = cost;
      	  	    
      	  		  if (maxRT != minRT)
      	  	  	    weight += (rt - minRT) / (maxRT - minRT) * (maxCost - minCost) + minCost;
      	  		  else
      	  	  	    weight += maxRT;     	  	
      	  	
      	  		  if (maxCDN != minCDN)
      	  	  	    weight += (cdn - minCDN) / (maxCDN - minCDN) * (maxCost - minCost) + minCost;
      	  		  else
      	  	        weight += maxCDN;*/
      	  	        
      	  	      /*if (maxRT != minRT)
      	  	        weight += (rt - minRT)/(maxRT - minRT);
      	  	        
      	  	      if (maxCost != minCost)
      	  	        weight += (cost - minCost)/(maxCost - minCost);
      	  	        
      	  	      if (maxCDN != minCDN)
      	  	        weight += (cdn - minCDN)/(maxCDN - minCDN);*/
      	  	      
      	  	      /*if (maxDeadline != minDeadline)
      	  	        weight += (deadline - minDeadline)/(maxDeadline - minDeadline);
      	  	        
      	  	      if (maxCost != minCost)
      	  	        weight += (cost - minCost)/(maxCost - minCost);
      	  	        
      	  	      if (maxCDN != minCDN)
      	  	        weight += (cdn - minCDN)/(maxCDN - minCDN);*/
      	  	      
      	  	      /*weight += deadline/avgDeadline;
      	  	      weight += cost/avgCost;
      	  	      weight += cdn/avgCDN;*/      	  	          	  	         	  	    	
      	  	    }
      	  	  else
      	  	    if (AortaSchedulingSimulator.basis == AortaSchedulingSimulator.CDN_BASIS)
      	  	      {
      	  	        weight = cdn;
      	  	        
      	  	        weight += deadline*avgCDN/avgDeadline;
      	  	        
      	  	        weight+= cost*avgCDN/avgCost;   	             	  	            	  	    	
      	  	      } 
      	  	
      	  	weight *= f;
      	  	
      	  	if (AortaSchedulingSimulator.WRE == false)
      	  	  weight *= (1+sd.failProb);      	  	
      	  	
      	  	if (weight < minWeight)
      	  	  {
      	  	  	minWeight = weight;
      	  	  	
      	  	  	sel_ar = ar;
      	  	  	
      	  	  	sel_did = String.valueOf(sd.id);
      	  	  	
      	  	  	sel_actionCost = cost;
      	  	  	
      	  	  	sel_status = (Tuple)ar.statusVector.elementAt(j);
      	  	  	
      	  	  	sel_weight = weight;
      	  	  }      	  	      	  	
      	  }      	      	   	
      }
      
    /*public void heuristicMultiplication(SchedulingDevice sd, Vector requests, double f)
      {
      	int i, j, id;
      	int pan, tilt, zoom, deltaPan, deltaTilt, deltaZoom;
      	double cost = 0.0, laxity, maxLaxity = Double.MIN_VALUE;
      	double weight, minWeight = Double.MAX_VALUE;
      	double actionCost, headMovementCost;
      	Tuple t;      	
      	Point loc = new Point();
      	ActionRequest ar;
      	SchedulingDevice sd2;
      	
      	for (i=0;i<requests.size();i++)
      	  {
      	  	ar = (ActionRequest)requests.elementAt(i);
      	  	
      	  	for (j=0;j<ar.didVector.size();j++)
      	  	  {
      	  	  	id = Integer.parseInt((String)ar.didVector.elementAt(j));     	  	  	
      	  	  	
      	  	  	if (id == sd.id)
      	  	  	  cost = Double.parseDouble((String)ar.costVector.elementAt(j));
      	  	  	  
      	  	  	sd2 = schdevs[id-1];      	  	  	
      	  	  	
      	  	  	if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
      	  	  	  {
      	  	  	  	t = (Tuple)ar.statusVector.elementAt(j);
		         	
		         	pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	     	
    	   	     	tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	     	
    	   	     	zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	     
    	   	     	deltaPan = Math.abs(sd2.camSim.pan - pan);
    	   	     	
    	   	     	deltaTilt = Math.abs(sd2.camSim.tilt - tilt);
    	   	     	
    	   	     	deltaZoom = Math.abs(sd2.camSim.zoom - zoom);
    	   	     	
    	   	     	headMovementCost = sd2.camSim.computeHeadMovementCost(deltaPan, deltaTilt, deltaZoom);
		         	
		         	actionCost = AortaSchedulingSimulator.estimatePhotoActionCost(headMovementCost)/1000.0;     	  	  	  	
      	  	  	  }
      	  	  	else
      	  	  	  {
      	  	  	  	t = (Tuple)ar.statusVector.elementAt(j);
		         	
		         	loc.x = Double.parseDouble((String)t.fields.elementAt(0));
    	   	     	
    	   	     	loc.y = Double.parseDouble((String)t.fields.elementAt(1));
    	   	     	
    	   	     	loc.z = Double.parseDouble((String)t.fields.elementAt(2));    	   	     
    	   	     	
    	   	     	actionCost = AortaSchedulingSimulator.estimateWarnActionCost(sd2.er1Sim.loc,sd2.er1Sim.angle,loc)/1000.0;	         	
      	  	  	  }
      	  	  	  
      	  	  	if (sd2.nextFreeTime < currentSystemTime)
      	  	  	  sd2.nextFreeTime = currentSystemTime;
      	  	  	
      	  	  	laxity = ar.startTime + ar.deadline - sd2.nextFreeTime - actionCost;
      	  	  	
      	  	  	if (laxity > 0)
      	  	  	  if (maxLaxity < laxity)
      	  	  	    maxLaxity = laxity;
      	  	  }
      	  	  
      	  	if (maxLaxity == Double.MIN_VALUE)
      	  	  weight = Double.MAX_VALUE;
      	  	else
      	  	  weight = cost*maxLaxity*f*(1+sd.failProb); // * ar.didVector.size()      	  	
      	  	
      	  	if (weight < minWeight)
      	  	  {
      	  	  	minWeight = weight;
      	  	  	
      	  	  	sel_ar = ar;
      	  	  	
      	  	  	sel_did = String.valueOf(sd.id);
      	  	  	
      	  	    for (j=0;j<sel_ar.didVector.size();j++)
      	  	      if (Integer.parseInt((String)sel_ar.didVector.elementAt(j)) == sd.id)
      	  	        {
      	  	          sel_actionCost = Double.parseDouble((String)sel_ar.costVector.elementAt(j));	
      	  	          
      	  	          sel_status = (Tuple)sel_ar.statusVector.elementAt(j);
      	  	          
      	  	          break;
      	  	        }
      	  	  	
      	  	  	sel_weight = weight;
      	  	  }        	  	         	  	
      	  }    	
      }
      
      
    public void heuristicPriority(SchedulingDevice sd, Vector requests, double f)
      {
        int i,j;
        double minDeadline = Double.MAX_VALUE;
        ActionRequest ar;
        
                
        // first priority: deadline      
        for (i=0;i<requests.size();i++)
          {
          	ar = (ActionRequest)requests.elementAt(i);
          	
          	if (minDeadline > ar.deadline)
          	  minDeadline = ar.deadline;          	
          }          
        
        Vector requests_c = new Vector (requests.size(), 10);
        
        for (i=0;i<requests.size();i++)
          {
          	ar = (ActionRequest)requests.elementAt(i);
          	
          	if (ar.deadline == minDeadline)
          	  requests_c.addElement((ActionRequest)ar); 
          }
          
        if (requests_c.size() == 1)
          {
          	sel_ar = (ActionRequest)requests_c.elementAt(0);
          	
          	sel_did = String.valueOf(sd.id);
          	
          	sel_weight = sel_ar.deadline*f*(1+sd.failProb);
          	
          	for (j=0;j<sel_ar.didVector.size();j++)
          	  if (Integer.parseInt((String)sel_ar.didVector.elementAt(j))==sd.id)
          	    {
          	      sel_actionCost = Double.parseDouble((String)sel_ar.costVector.elementAt(j));
          	      sel_status = (Tuple)sel_ar.statusVector.elementAt(j);
          	      break;	
          	    } 
          	    
          	return;         	
          }          
        
        
        // second priority: cost   
        double minCost = Double.MAX_VALUE;             
            
        for (i=0;i<requests_c.size();i++)
          {
          	ar = (ActionRequest)requests_c.elementAt(i);
          	
          	for (j=0;j<ar.didVector.size();j++)
              if (Integer.parseInt((String)ar.didVector.elementAt(j)) == sd.id) 
                {
                  double cost = Double.parseDouble((String)ar.costVector.elementAt(j));
                  if (minCost > cost)
                    minCost = cost;
                }
          }
          
        Vector requests_cdn = new Vector (requests_c.size(), 10);
        
        for (i=0;i<requests_c.size();i++)
          {
          	ar = (ActionRequest)requests_c.elementAt(i);
          	
          	for (j=0;j<ar.didVector.size();j++)
              if (Integer.parseInt((String)ar.didVector.elementAt(j)) == sd.id) 
                {
                  double cost = Double.parseDouble((String)ar.costVector.elementAt(j));
                  if (cost == minCost)
                    requests_cdn.addElement((ActionRequest)ar);
                }          	
          }
          
        if (requests_cdn.size() == 1)
         {
          	sel_ar = (ActionRequest)requests_cdn.elementAt(0);
          	
          	sel_did = String.valueOf(sd.id);          	
          	
          	for (j=0;j<sel_ar.didVector.size();j++)
          	  if (Integer.parseInt((String)sel_ar.didVector.elementAt(j)) == sd.id)
          	    {
          	      sel_actionCost = Double.parseDouble((String)sel_ar.costVector.elementAt(j));
          	      sel_status = (Tuple)sel_ar.statusVector.elementAt(j);	
          	    } 
          	
          	sel_weight = sel_actionCost*f*(1+sd.failProb);
          	   
          	return;         	
          }
          
        
        // third priority: candidate device number        
        double minCDN = Double.MAX_VALUE;
        int mark = 0;
        
        for (i=0;i<requests_cdn.size();i++)
          {
          	ar = (ActionRequest)requests_cdn.elementAt(i);
          	
          	if (minCDN > ar.didVector.size())
          	  {mark = i; minCDN = ar.didVector.size();}
          }
          
        sel_ar = (ActionRequest)requests_cdn.elementAt(mark);
        
        sel_did = String.valueOf(sd.id);          	
          	
        for (j=0;j<sel_ar.didVector.size();j++)
          if (Integer.parseInt((String)sel_ar.didVector.elementAt(j)) == sd.id)
           {
          	 sel_actionCost = Double.parseDouble((String)sel_ar.costVector.elementAt(j));
          	 sel_status = (Tuple)sel_ar.statusVector.elementAt(j);	
           } 
          	
        sel_weight = minCDN*f*(1+sd.failProb);
          	   
        return;   	          	
      }	*/
 } 