/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * A simulator for the ER1 programmable robots.
 *
 * @see ActionOperator
 * @see AortaSchedulingSimulator
 */
public class ER1Simulator extends Thread {	

   /** ID of this simulated robot. */
	public String id;
	
	public static final String ipPrefix = "192.168.42.";
	/** IP of (the notebook connected to) this simualted robot. */
	public String ip;
	
	/** Location of this simulated robot. */
	public Point loc;
	public static final double HOME_LOC_X = -3.8;  // unit: meter
	public static final double HOME_LOC_Y = 0.15;
	public static final double HOME_LOC_Z = -1.2;
	public static final double MIN_LOC_X = -3.5;
	public static final double MAX_LOC_X = 0.5;
	public static final double MIN_LOC_Y = 0.5;
	public static final double MAX_LOC_Y = 11.5;
	public static final double MIN_LOC_Z = -1.2;
	public static final double MAX_LOC_Z = 1.3;	
	
	/** Orientational angle of this simulated robot. */
	public double angle;
	public static final double HOME_ANGLE = 0;  // unit: degree, range (-180, 180]	
	
	public static final int initialPort = 9000;
	/** Port of the socket server. */
	public int port;
	public ServerSocket serverSocket;
    
    
	public static final double MOVE_MAX_SPEED = 50;  // unit: cm
	public static final double ROTATE_MAX_SPEED = 90; // unit: degree
	public static final double MOVE_DEFAULT_SPEED = MOVE_MAX_SPEED;
	public static final double ROTATE_DEFAULT_SPEED = ROTATE_MAX_SPEED / 3;
	/* The estimated costs of atomic operations */
	public static final double MOVEPERCM_ESTIMATED_COST = 1000/MOVE_DEFAULT_SPEED;  
	public static final double ROTATEPERDEG_ESTIMATED_COST = 1000/ROTATE_DEFAULT_SPEED;
	public static final double PLAYPERSEC_ESTIMATED_COST = 1000;
	
	public static final String DEFAULT_PHRASE_FILE = "../messages/warning.txt";
	public static final String DEFAULT_SOUND_FILE = "../sounds/warning.wav";
	public static final double DEFAULT_SOUND_COST = 8000;
	
	PrintWriter writeToClient;
    BufferedReader readFromClient;
    
    public static final String MOVE_RESPONSE = "move done";
    public static final String PLAY_RESPONSE = "play done";
	public static final String COMMAND_RESPONSE = "ok";
	public static final String TERMINATE_STRING = "quit";
	
	private byte tag = COMMAND_TAG;
	private static final byte MOVE_TAG = 0x01;
	private static final byte PLAY_TAG = 0x02;
	private static final byte COMMAND_TAG = 0x03;
	
	/** Maximum number of real ER1 robots in the pervasive lab. */
	public static final int MAX_REAL_ROBOT_NUM = 1;	
	
    public static final double STANDARD_POLLING_TIME = 320;    
	public static final double POLLING_TIMEOUT_LIMIT = 2*STANDARD_POLLING_TIME;
	
	
    ER1Simulator (String id)
      {
      	this.id = id;
      	ip = ipPrefix + id;
      	port = initialPort + Integer.parseInt(id) - 1;  
      	
        loc = new Point();
        loc.x = HOME_LOC_X;
        loc.y = HOME_LOC_Y;
        loc.z = HOME_LOC_Z;  
        
        angle = HOME_ANGLE; 	
      	
      	try {serverSocket = new ServerSocket(port);}
	    catch (Exception e) {e.printStackTrace();}	    
	        
      }
      
    public void setLoc (double loc_x, double loc_y, double loc_z)
      {
      	loc.x = loc_x;
      	
      	loc.y = loc_y;
      	
      	loc.z = loc_z;
      }
      
    public void setAngle (double angle)
      {
      	this.angle = angle;
      }      
      
    public void run()
      {
      	try
      	  {
      	  	while (true)
      	  	  {
      	  	  	Socket clientSocket = serverSocket.accept();
	       	    
	            readFromClient = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
		        writeToClient = new PrintWriter (new DataOutputStream(clientSocket.getOutputStream()));
		    
		        while (true)
		          if (receiveCommand() == false)
		            break;
		        
		        clientSocket.close();		        
		      }	    
      	  }
      	catch (Exception e) {e.printStackTrace();}    	
      }    
    
    public boolean receiveCommand()
      {
      	try
      	  {
      	  	String cmd = readFromClient.readLine();
		    cmd.toLowerCase().trim();		    
		    
		    if (cmd.equals(TERMINATE_STRING))
		      return false;  
		    
		    int mark;
		    if ((mark = cmd.indexOf("move ")) != -1)
		      move(cmd.substring(mark+"move ".length()).trim());
		    else 
		      if ((mark = cmd.indexOf("play ")) != -1)
		        play(cmd.substring(mark+"play ".length()).trim());
		      else
		        if (cmd.equals("events"))
		          events();
		        else
		          if (cmd.equals("position"))
		            position();
		          else
		            others(); 	        
          }
      	catch (Exception e) {return false;}
      	
      	return true;
      }      
   
   public void move(String cmd)
     {
     	int mark;
     	mark = cmd.indexOf(" ");
     	if (mark == -1)
     	  mark = cmd.indexOf("\t");
     	if (mark == -1)
     	  return;          
        
        double movement = 0;
        try {movement = Double.parseDouble(cmd.substring(0,mark).trim());}
        catch (Exception e) {return;}
     	
     	cmd = cmd.substring(mark).trim();
     	if (cmd.charAt(0) == 'c')
     	  drive(movement);
     	else 
     	  if (cmd.charAt(0) == 'd')
     	    rotate(movement);
     	  else
     	    return;
     	    
       writeToClient.println(COMMAND_RESPONSE);
       writeToClient.flush();  
     	    
       tag = MOVE_TAG;     	
     }
     
   public void drive(double deltaDistance)
     {
     	/* double slopeAngle;
     	
     	if (loc.x != 0)
     	  slopeAngle = Math.atan(loc.y / loc.x)/Math.PI*180;
     	else
     	  slopeAngle = 90;
     	
     	slopeAngle += angle; 
     	
     	if (slopeAngle > 180)
     	  slopeAngle -= 360;
     	  
     	if (slopeAngle < 0)
     	  slopeAngle += 180;
     	
     	if (Math.abs(slopeAngle - 90)<0.1)
     	  slopeAngle = 90;
     	  
     	
        double c0, c1;  // y = c0 + c1*x
     	if (Math.abs(slopeAngle) == 90)
     	  {
     	  	c0 = 0;
     	  	c1 = 0;
     	  }
     	else
     	  {
     	  	c1 = Math.tan(slopeAngle/180*Math.PI);
     	  	c0 = loc.y - c1*loc.x;
     	  } 
     	
     	
     	double x1, y1, x2, y2;
     	double deltaX = deltaDistance / Math.sqrt(1+c1*c1);
     	
     	x1 = loc.x + deltaX;  x2 = loc.x - deltaX;
     	y1 = c0 + c1*x1;  y2 = c0 + c1*x2;
     	
     	if (slopeAngle == 90)
     	  {
     	  	x1 = x2 = 0;
     	  	y1 = loc.y + deltaDistance;
     	  	y2 = loc.y - deltaDistance;
     	  }
     	  
     	     	 
     	if (angle == 0)  
     	  {
     	  	if (y1 < 0)
     	  	  {loc.x = x2; loc.y = y2;}
     	  	else
     	  	  if (y2 < 0)
     	  	    {loc.x = x1; loc.y = y1;}
     	  	  else
     	  	    if (x1*x1 + y1*y1 >= loc.x*loc.x + loc.y*loc.y)
     	  	      {loc.x = x1; loc.y = y1;}
     	  	    else
     	  	      {loc.x = x2; loc.y = y2;}
     	  }
     	else
     	  if (angle == 180)
     	    {
     	  	  if (y1 < 0)
     	  	    {loc.x = x2; loc.y = y2;}
     	  	  else
     	  	    if (y2 < 0)
     	  	      {loc.x = x1; loc.y = y1;}
     	  	    else
     	  	      if (x1*x1 + y1*y1 <= loc.x*loc.x + loc.y*loc.y)
     	  	        {loc.x = x1; loc.y = y1;}
     	  	      else
     	  	        {loc.x = x2; loc.y = y2;}
     	    }
     	  else
     	    if (angle > 0)
     	      {
     	      	if (loc.x != 0)
     	      	  {
     	      	  	if (loc.x > 0)
     	      	  	  if (y1 > loc.y/loc.x*x1 )
     	      	        {loc.x = x1; loc.y = y1;}
     	               else
     	      	        {loc.x = x2; loc.y = y2;}
     	      	    else
     	      	      if (y1 < loc.y/loc.x*x1 )
     	      	        {loc.x = x1; loc.y = y1;}
     	              else
     	      	        {loc.x = x2; loc.y = y2;} 
     	      	  }    	      	  
     	      	else
     	      	  if (x1 < 0)
     	      	  	{loc.x = x1; loc.y = y1;}
     	          else
     	      	      {loc.x = x2; loc.y = y2;} 
     	      	      
     	      	if (y1 < 0)
     	      	  {loc.x = x2; loc.y = y2;} 
     	      	else
     	      	  if (y2 < 0)
     	      	    {loc.x = x1; loc.y = y1;}   	      	  
     	      }
     	    else  // angle < 0
              {
              	if (loc.x != 0)
     	      	  {
     	      	  	if (loc.x > 0)
     	      	  	  if (y1 < loc.y/loc.x*x1 )
     	      	        {loc.x = x1; loc.y = y1;}
     	               else
     	      	        {loc.x = x2; loc.y = y2;}
     	      	    else
     	      	      if (y1 > loc.y/loc.x*x1)
     	      	        {loc.x = x1; loc.y = y1;}
     	              else
     	      	        {loc.x = x2; loc.y = y2;} 
     	      	  }       	      	  
     	      	else
     	      	  if (x1 > 0)
     	      	  	{loc.x = x1; loc.y = y1;}
     	          else
     	      	      {loc.x = x2; loc.y = y2;}
     	      	      
     	      	if (y1 < 0)
     	      	  {loc.x = x2; loc.y = y2;} 
     	      	else
     	      	  if (y2 < 0)
     	      	    {loc.x = x1; loc.y = y1;}  
     	      }
     	      
                
     	angle = 0;      	        	     	
        
        System.out.println(slopeAngle);
        System.out.println(x1 + " " + y1 + " " + x2 + " " + y2);
        System.out.println(loc.x + " " + loc.y); System.exit(0); */
        
        if (angle == 90)
          loc.y += deltaDistance;
        else
          if (angle == -90)
            loc.y -= deltaDistance;
          else
            if (angle == 0)
              loc.x += deltaDistance;
            else 
              if (angle == 180)
                loc.x -= deltaDistance;
              else
                {
                  double k = Math.tan(angle/180*Math.PI);
                  double c = loc.y - k*loc.x;
              
                  double x1,x2,y1,y2;
                  x1 = Math.sqrt(deltaDistance*deltaDistance/(1+k*k));
                  x2 = -x1;
                  y1 = k*x1;
                  y2 = k*x2;
              
                  x1 += loc.x; x2 += loc.x;
                  y1 += loc.y; y2 += loc.y;
              
                  double targetX, targetY; 
                  if (angle > 0)
                    if (y1 > y2)
                      {loc.x = x1; loc.y = y1;}
                    else
                      {loc.x = x2; loc.y = y2;}
                  else
                     if (y1 < y2)
                      {loc.x = x1; loc.y = y1;}
                    else
                      {loc.x = x2; loc.y = y2;}                          	
                }     	
     	
     	long driveCost = (long)Math.abs(Math.ceil(deltaDistance*MOVEPERCM_ESTIMATED_COST));
     	
     	try {Thread.currentThread().sleep(driveCost);}
     	catch (Exception e) {e.printStackTrace();}
     	System.out.println("Robot " + id + " drives " + Math.ceil(deltaDistance) + " centimeters.");   	   	  	
     }
     
   public void rotate(double deltaAngle)
     {
     	angle += deltaAngle;
     	
     	if (angle > 180)
     	  angle -= 360;
     	  
     	if (angle <= -180)
     	  angle += 360;      	  
     	  
     	long rotateCost = (long)Math.abs(Math.ceil(deltaAngle*ROTATEPERDEG_ESTIMATED_COST));     	
     	try {Thread.currentThread().sleep(rotateCost);}
     	catch (Exception e) {e.printStackTrace();}     	
     	System.out.println("Robot " + id + " rotates " + Math.ceil(angle) + " degrees.");    	
     }
     
   public void play(String cmd)
     {
     	int mark;
     	if ((mark=cmd.indexOf("phrase ")) == -1)
     	  return;
          
        mark += "phrase ".length();
        cmd = cmd.substring(mark).trim();
        cmd = cmd.substring(1,cmd.length()-1);
     	
     	String str = "";
     	try 
     	  {
     	  	BufferedReader br = new BufferedReader (new FileReader (DEFAULT_PHRASE_FILE));
     	  	String line;
     	  	while ((line=br.readLine())!=null)
     	  	  str += line;
     	  	br.close();
     	  }
     	catch (Exception e) {return;}
     	
     	if (!cmd.equals(str))
     	  return;
     	
     	try 
     	  {
     	  	cmd = "C:/Program Files/Windows Media Player/mplayer2.exe " + "\"" + DEFAULT_SOUND_FILE + "\"";
     	    Runtime.getRuntime().exec(cmd);
     	    
     	    Thread.currentThread().sleep((long)(DEFAULT_SOUND_COST));
     	  }
     	catch (Exception e) {e.printStackTrace();} 
     	
     	writeToClient.println(COMMAND_RESPONSE);
     	writeToClient.flush();    	
     	     	
     	tag = PLAY_TAG;     	
     	
     	System.out.println("Robot " + id + " plays phrase \"" + str + "\".");      	     	
     }
     
   public void events()
     {
     	if (tag == MOVE_TAG)
     	  writeToClient.println(MOVE_RESPONSE);
     	else
     	  if (tag == PLAY_TAG)
     	    writeToClient.println(PLAY_RESPONSE);
     	  else
     	    writeToClient.println(COMMAND_RESPONSE);
     	    
       	tag = COMMAND_TAG;
     	    
     	writeToClient.flush();
     }
     
   public void position()
     {
       String response = "OK ";
       
       response += String.valueOf(loc.x) + " ";
       
       response += String.valueOf(loc.y) + " ";
       
       response += "0";
       
       /* double slopeAngle;     	
       if (loc.x != 0)
     	  slopeAngle = Math.atan(loc.y / loc.x)/Math.PI*180;
       else
     	 slopeAngle = 90;
       if (slopeAngle == 0)
         if (loc.x < 0)
           slopeAngle = 180;
       if (slopeAngle < 0)
         slopeAngle += 180;     	
       slopeAngle += angle;
       slopeAngle -= 90;      	
       if (slopeAngle > 180)
     	 slopeAngle -= 360;     	  
       if (slopeAngle <= -180)
     	   slopeAngle += 360; 
     	 
       response += String.valueOf(slopeAngle); */      	   	      
       
       writeToClient.println(response);	
       writeToClient.flush();       
     }
     
   public void others()
     {
       writeToClient.println(COMMAND_RESPONSE);
       writeToClient.flush();
     }
     
   public double deltaDistance(Point targetLoc, String sid)
     {
       if (sid.equals("1"))  // Sensor 1
	     {targetLoc.x = 0; targetLoc.y = 1.2;}
	   else
	     if (sid.equals("2"))  // Sensor 2
	       {targetLoc.x = -1.3; targetLoc.y = 10;}
	     else
	       if (sid.equals("3") || sid.equals("6"))  // Sensor 3 or Sensor 6
	         {targetLoc.x = 0; targetLoc.y = 5.2;}
	       else
	         if (sid.equals("4") || sid.equals("7"))  // Sensor 4 or Sensor 7
	           {targetLoc.x = 0; targetLoc.y = 1.3;}
	         else
	           if (sid.equals("5") || sid.equals("8"))  // Sensor 5 or Sensor 8
	             {targetLoc.x = -3.3; targetLoc.y = 4.8;}
	           else
	             if (sid.equals("9"))  // Sensor 9
	               {targetLoc.x = -3.5; targetLoc.y = 1.5;}
	             else
	               if (sid.equals("10"))  // Sensor 10
	                 {targetLoc.x = -0.5; targetLoc.y = 6.5;}
        
        return Math.sqrt(Math.pow(loc.x-targetLoc.x,2)+Math.pow(loc.y-targetLoc.y,2))*100;
     }
     
   public double deltaAngle(Point targetLoc)
     {
     	/* if ((loc.x == targetLoc.x) && (loc.y == targetLoc.y))
     	  return 0;
     	  
     	if ((loc.x == 0) && (loc.y == 0))
     	  {
     	  	double slopeAngle;
     	  	if (targetLoc.x == 0)
     	  	  slopeAngle = 90;
     	  	else
     	  	  {
     	  	  	slopeAngle = Math.atan(targetLoc.y/targetLoc.x)/Math.PI*180;
     	  	  	if (slopeAngle < 0)
     	  	  	  slopeAngle += 180;
     	  	  	slopeAngle -= 90;
     	  	  }
     	  	
     	  	if (slopeAngle - angle > 180)
     	  	  return Math.abs(deltaAngle = slopeAngle-angle-360);
     	  	else
     	  	  if (slopeAngle - angle <= -180)
     	  	    return Math.abs(deltaAngle = slopeAngle-angle+360);
     	  	  else
     	  	    return Math.abs(deltaAngle = slopeAngle-angle);
     	  }
     	  
     	double x1 = loc.x;
     	double y1 = loc.y;
     	double x2 = targetLoc.x - loc.x;
     	double y2 = targetLoc.y - loc.y; 
     	
     	double theta = Math.acos((x1*x2+y1*y2)/(Math.sqrt(x1*x1+y1*y1)*Math.sqrt(x2*x2+y2*y2)))/Math.PI*180;
     	if (x1*y2-x2*y1 < 0)
     	  theta = 360 - theta;
     	
     	if (theta > 180)
     	  theta -= 360;
     	  
     	if (theta - angle > 180)
     	  	  return Math.abs(deltaAngle = theta-angle-360);
     	  	else
     	  	  if (theta - angle <= -180)
     	  	    return Math.abs(deltaAngle = theta-angle+360);
     	  	  else
     	  	    return Math.abs(deltaAngle = theta-angle); */
     	  	    
     	double currentX = loc.x, currentY = loc.y;
     	double targetX = targetLoc.x, targetY = targetLoc.y;  	    
        
        if ((currentX == targetX) && (currentY == targetY))
	      return 0; 
	    else
	      {
	     	double deltaAngle;
	     	if (currentX == targetX)
	     	  if (currentY < targetY)
	     	    deltaAngle = 90;
	     	  else
	     	    deltaAngle = -90;
	     	else
	     	  {
	     	  	 deltaAngle = Math.atan((targetY - currentY)/(targetX-currentX))/Math.PI*180;
	     	  	  	
	     	  	 if ((deltaAngle > 0) && (targetY < currentY))
	     	  	   deltaAngle -= 180;
	     	  	  	
	     	  	 if ((deltaAngle < 0) && (targetY > currentY))
	     	  	   deltaAngle += 180;
	     	  } 
	       	deltaAngle -= angle;
	     	if (deltaAngle > 180)
	     	  deltaAngle -= 360;
	     	if (deltaAngle <= -180)
	     	  deltaAngle += 360;
	     	  
	     	return deltaAngle;
	     }	
     }     
     
   /* public static void main(String argvs [])
     {
     	ER1Simulator er1 = new ER1Simulator("1");
     	
     	er1.setLoc(1.73,1,0);
     	er1.angle = -30;
     	er1.drive(2);     	
     	
        er1.start();     	    	
     } */
}