/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.lang.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.awt.*;
import java.net.*;
import sun.misc.*;

/**
 * Implementation of the system built-in action photo() 
 * on AXIS 2130 PTZ network cameras.
 *
 * @see ActionOperator
 */
public class Photo extends Action {
	
    Aorta aortaqp;
	
	TakePhotos tp;
	
	String imageFileName;	// the pathname of the stored photo	
	
	boolean firstTime = true;		
	
    Calendar currentCalendar = Calendar.getInstance();  
		
	/** Width of the photo display window (unit: pixel). */
	public static final int DEFAULT_WIDTH = 360;
	/** Height of the photo display window (unit: pixel). */
	public static final int DEFAULT_HEIGHT = 260;  
	/* Candidate resolutions for photos taken by AXIS 2130 PTZ network cameras: (176, 112), (352,240), and (704,480). */
		
	/* Specific information about AXIS 2130 network cameras */
	static final String SMALL_PHOTO_URL = "halfsize.jpg";
	static final String MEDIUM_PHOTO_URL = "fullsize.jpg";
	static final String LARGE_PHOTO_URL = "hugesize.jpg";	
		
	/* Private variables */
	String camera_id;
	int id;
	String camera_ip;
	Point loc;
	String pathName;
	String userPassword;
	String encoding;
	CameraSimulator camSim [] = new CameraSimulator [CameraSimulator.MAX_REAL_CAMERA_NUM];
		
	
	/** 
	 * @param aortaqp the Aorta query processor that this action registers to
	 */
	public Photo(Aorta aortaqp)  
	  { 
	    this.aortaqp = aortaqp;
	    
	    name = "photo";  
	    
	    profile = "../action_profiles/system/photo.xml";
	    
	    params[0].name = "camera_ip";
	    params[0].type = "String";
	    params[1].name = "location";
	    params[1].type = "Point";
	    params[2].name = "directory_name";
	    params[2].type = "String";
	    paramNum = 3;
	    
	    flag = Action.SYSTEM_BUILT_IN;
	    
	    if (!parseProfile())
	      {
	      	System.err.println("Fail to initialize the system built-in action photo().");
	      	System.exit(0);
	      }  
	    
	    for (int i=0;i<camSim.length;i++)
	      camSim[i] = new CameraSimulator(String.valueOf(i+1));
	       
	    try {tp = new TakePhotos(DEFAULT_WIDTH,DEFAULT_HEIGHT,aortaqp);} 
	    catch (Exception e) {e.printStackTrace();}	    
	  } 
	  
	
	/**
	 * This action implements the operations involved in the action.  
	 * It uses a simulated camera to execute the action if {@link Aorta#CAMERA_SIMULATION} 
	 * is set to be <code>true</code>;
	 * uses a real camera otherwise.
	 * 
	 * <p>
	 * The action has three input parameters:
	 * <ul>
	 * <li> The ip of a camera.
	 * <li> The location for the camera to take a photo (.jpeq file) towards.
	 * <li> The directory (path) name to store the photo.
	 * </ul>
	 *
	 * @param s the string representation of the instantiated values of the action input parameters
	 * @return a {@link Image} object respresents the current photo taken. 
	 */
	public Object doAction(String [] s) 
	 {
	  
	   camera_ip = s[0];
	   loc = Point.parsePoint(s[1]);
	   pathName = s[2];
	   camera_id = s[3];
	   id = Integer.parseInt(camera_id);
	   userPassword = s[4] + ":" + s[5];
	   encoding = new sun.misc.BASE64Encoder().encode (userPassword.getBytes());   
	        
	   if (!polling())
	     {
	       q.logFileWriter.println("The selected camera suddenly becomes unavailable. Fail to execute the action.");
	       System.out.println("exec action photo(" + s[0] + "," + s[1] + "," + "\"" + s[2] + "\")");
           System.err.println("The selected camera suddenly becomes unavailable. Fail to execute the action.");
           System.out.print("Aorta>");           
	       return null;
	     }
	      
	   headMovement();
	   
	   Image image = takeMediumSizePhoto();
	      
	   System.out.println("exec action photo(" + s[0] + "," + s[1] + "," + "\"" + s[2] + "\")");           
       if (q.type == aortaqp.EVENT_DRIVEN_QUERY)
         System.out.print("Aorta>");
       if (q.type == aortaqp.TRADITIONAL_CONTINUOUS_QUERY)
         if (!firstTime)
           System.out.print("Aorta>");          
         else 
           {
           	 System.out.println();
           	 firstTime = false;
           }
         
       return image;                        
	}  
	
    
    /**
     * Save the current photo to disk.
     *
     * @param directoryName the directory to store the photo
     * @param size the scale of the photo
     */
    public void saveImageToDisk (String directoryName, String size) 
      {
          currentCalendar.setTimeInMillis(System.currentTimeMillis());
          
          int len = directoryName.length() - 1;
          if (pathName.charAt(len) != '/')
            pathName += "/";
          
          // use current system time stamp as the file name
          imageFileName = directoryName + String.valueOf(currentCalendar.get(Calendar.HOUR))
                            + "h_" + String.valueOf(currentCalendar.get(Calendar.MINUTE)) + "m_" 
                            + String.valueOf(currentCalendar.get(Calendar.SECOND)) + "s_" 
                            + String.valueOf(currentCalendar.get(Calendar.YEAR)) + "-" 
                            + String.valueOf(currentCalendar.get(Calendar.MONTH)) + "-" 
                            + String.valueOf(currentCalendar.get(Calendar.DAY_OF_MONTH)) + size + ".jpg";
                  
          if (aortaqp.CAMERA_SIMULATION)
            try 
              {
              	if (size.equals("_medium"))
              	  Runtime.getRuntime().exec("cp " + CameraSimulator.MEDIUM_IMAGE_FILE_EXAMPLE + " " + imageFileName);
              	else
              	  if (size.equals("_small"))
              	    Runtime.getRuntime().exec("cp " + CameraSimulator.SMALL_IMAGE_FILE_EXAMPLE + " " + imageFileName);              	
              }
            catch (Exception e) {e.printStackTrace();}
          else
            tp.saveCurrent(imageFileName);                   
      } 
      
      
    /**
     * Poll a real or simulated camera.
     *
     * @see #pollRealCamera
     * @see #pollSimulatedCamera
     *
     * @return <code>true</code> if the polling is successfully established;
     *         <code>false</code> otherwise.
     */
    public boolean polling()
      {
        if (aortaqp.CAMERA_SIMULATION)
          return pollSimulatedCamera();
        else
          return pollRealCamera();	
      }
      
    public boolean pollRealCamera ()
      {
         try
           {
             System.setProperty("sun.net.client.defaultConnectTimeout",String.valueOf(CameraSimulator.POLLING_TIMEOUT_LIMIT)); 
             System.setProperty("sun.net.client.defaultReadTimeout",String.valueOf(CameraSimulator.POLLING_TIMEOUT_LIMIT)); 
          
             URL url = new URL("http://" + camera_ip + "/axis-cgi/com/ptz.cgi?query=position"); 
      	     URLConnection uc = url.openConnection();
             uc.setRequestProperty  ("Authorization", "Basic " + encoding);
                
             BufferedReader br = new BufferedReader (new InputStreamReader(uc.getInputStream()));
          	         	    
             String str;
             int start;
             while ( (str = br.readLine()) != null)
               {
          	  	  if ( (start = str.indexOf("pan")) != -1)
          	  	    {
          	  	      start = str.indexOf("=",start)+1;
          	  	  	  camSim[id-1].pan = Integer.parseInt(str.substring(start).trim());
          	        }
          	      
          	      if ( (start = str.indexOf("tilt")) != -1)
          	        {
          	  	  	  start = str.indexOf("=",start)+1;
          	  	  	  camSim[id-1].tilt = Integer.parseInt(str.substring(start).trim());
          	        }
          	  	  	
          	  	  if ( (start = str.indexOf("zoom")) != -1)
          	        {
          	  	  	  start = str.indexOf("=",start)+1;
          	  	      camSim[id-1].zoom = Integer.parseInt(str.substring(start).trim());
          	        }             	      
          	    }              
            }
          catch (Exception e)
		    { 
			  // socket timed out
			  System.setProperty("sun.net.client.defaultConnectTimeout","-1"); 
              System.setProperty("sun.net.client.defaultReadTimeout","-1");
              e.printStackTrace();
              return false;               
		    }                  
        
          System.setProperty("sun.net.client.defaultConnectTimeout","-1"); 
          System.setProperty("sun.net.client.defaultReadTimeout","-1");         
          
          return true;     	  
      }
        
    public boolean pollSimulatedCamera ()
      {
        if (camSim[id-1].polling() == CameraSimulator.POLLING_TIMED_OUT)
          return false;       
            
        return true;
      }
      
        
    /** 
     * Move the head of a real or simulated camera to the target position.
     *
     * @see #realCameraHeadMovement
     * @see #simulatedCameraHeadMovement
     */
    public void headMovement ()
      {
      	 if (aortaqp.CAMERA_SIMULATION)
	       simulatedCameraHeadMovement();
	     else
	       realCameraHeadMovement();
      }       
      
    public void realCameraHeadMovement()
      {
       	int pan_t = camSim[id-1].targetPan(loc);
      	int tilt_t = camSim[id-1].targetTilt(loc);
        int zoom_t = camSim[id-1].targetZoom(loc);      	
      	       
        try
          {   
            System.setProperty("sun.net.client.defaultConnectTimeout",String.valueOf(100));
            System.setProperty("sun.net.client.defaultReadTimeout",String.valueOf(100));
            URL url = new URL("http://" + camera_ip + "/axis-cgi/com/ptz.cgi?" + "pan=" + pan_t + 
                            "&" + "tilt=" + tilt_t + "&" + "zoom=" + zoom_t); 
            URLConnection uc = url.openConnection();
      	    uc.setRequestProperty  ("Authorization", "Basic " + encoding);
      	    uc.getContent();
      	    
      	    camSim[id-1].headMovement(pan_t, tilt_t, zoom_t); 
      	    
      	    System.setProperty("sun.net.client.defaultConnectTimeout","-1"); 
            System.setProperty("sun.net.client.defaultReadTimeout","-1");       	                    	   	  
          }  
        catch (Exception e) 
         {
            System.setProperty("sun.net.client.defaultConnectTimeout","-1"); 
            System.setProperty("sun.net.client.defaultReadTimeout","-1");
            
            camSim[id-1].headMovement(pan_t, tilt_t, zoom_t);                 
         }  
      }
       
    public void simulatedCameraHeadMovement ()
      {
       	int pan_t = camSim[id-1].targetPan(loc);
      	int tilt_t = camSim[id-1].targetTilt(loc);
      	int zoom_t = camSim[id-1].targetZoom(loc);
      	 
      	camSim[id-1].headMovement(pan_t, tilt_t, zoom_t);      	 
      }
       
     
    /**
     * Take a medium-size photo from a real or simulated camera.
     *
     * @see #realCameraTakeMediumSizePhoto
     * @see #simulatedCameraTakeMediumSizePhoto
     *
     * @return the photo taken
     */
    public Image takeMediumSizePhoto ()
      {
       	 if (aortaqp.CAMERA_SIMULATION)
       	   return simulatedCameraTakeMediumSizePhoto();
       	 else
       	   return realCameraTakeMediumSizePhoto();
      }
       
    public Image realCameraTakeMediumSizePhoto()
      {
       	Image image = null;
       	String imageFileName_t = "";
       	URL url;
       	URLConnection uc;
       	
       	try 
       	  {
       	   	 if (q.actionNesting)  // take the small size photo for phone display
               {
               	  url = new URL("http://" + camera_ip + "/jpg/" + SMALL_PHOTO_URL);
	              uc = url.openConnection();
                  uc.setRequestProperty  ("Authorization", "Basic " + encoding);
                  image = tp.takeCurrentPhoto(uc);              
                  saveImageToDisk(pathName,"_small");
                  imageFileName_t = imageFileName;
               }            
             
       	   	 
       	   	 url = new URL("http://" + camera_ip + "/jpg/" + MEDIUM_PHOTO_URL);
	         uc = url.openConnection();
             uc.setRequestProperty  ("Authorization", "Basic " + encoding);
          
             image = tp.takeCurrentPhoto(uc);
              
             tp.drawCurrentPhoto();
         
             saveImageToDisk(pathName,"_medium");
             
             if (q.actionNesting)
               imageFileName = imageFileName_t;              
             
          }
        catch (Exception e) {e.printStackTrace(); return null;}   
         
        return image;
      }
       
    public Image simulatedCameraTakeMediumSizePhoto()
      {
       	Image image = null;
       	try 
       	  {
       	  	image = camSim[id-1].takeMediumSizePhoto();       	  	
       	  	tp.image = image;
       	  	tp.createCanva();
       	  	tp.drawCurrentPhoto();
       	  	
       	  	saveImageToDisk(pathName,"_medium");
       	  	
       	  	if (q.actionNesting)
       	  	  {
       	  	  	image = camSim[id-1].takeSmallSizePhoto();
       	  	  	saveImageToDisk(pathName,"_small");
       	  	  }       	  	
       	  }
       	catch (Exception e) {e.printStackTrace(); return null;}
       	 
       	return image;
      }		
} 