/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Implementation of the algorithm with random request selection.
 * The algorithm is developed under the list scheduling (LS) discipline.
 * 
 * @see AortaSchedulingSimulator#schedule
 * @see SchedulingThread
 */ 
public class RANDOM {
 	
 	Vector requests;
	
	PrintWriter logFileWriter;
	
	double currentSystemTime = 0;	
	
	double schedulingTime = 0;
	
	double totalRequestCompletionTime = 0;
	
	double droppedRequestNum = 0;
	
	SchedulingDevice schdevs [] = new SchedulingDevice [AortaSchedulingSimulator.deviceNum];
		
	ActionRequest sel_ar = null;
	String sel_did = null;
	double sel_weight = Double.MAX_VALUE, sel_actionCost = Double.MAX_VALUE;
	Tuple sel_status = null;
	
	Random failRan = new Random();
	  	
		
 	/** 
	 * @param requests a set of action requests to be scheduled
	 * @param logFileWriter a {@link PrintWriter} corresponds to the log file
	 */
	 public RANDOM (Vector requests, PrintWriter logFileWriter) 
 	  {
 		this.requests = requests;
		
		this.logFileWriter = logFileWriter;				  
 	  }  	
 	
 	 	  
 	/** Assign and execute requests on devices by random selection. */
 	public void schedule()
 	  {
 		int i,j;
 		int count = 0;
 		int id;
 		long startTime, endTime;
 		double pollingTime, angle, actionCost;
 		double minStartTime, minNextFreeTime;
 		Point loc = new Point();
 		ActionRequest ar;	
 		
 		if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
 		  {
 		  	pollingTime = 0;
 		    
 		    for (i=0;i<schdevs.length;i++)
 		      if (pollingTime < schdevs[i].camSim.polling())
 		        pollingTime = schdevs[i].camSim.pollingTime; 
 		   
 		    schedulingTime += pollingTime / 1000.0;
 		  }
 		else
 		  schedulingTime += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;		
 		
 		currentSystemTime += schedulingTime;  
 		 		
 		   		
 		while (requests.size() > 0)
 		  {
 		  	// decide the set of requests that are eligible to be scheduled currently
 		  	Vector currentRequests = new Vector (requests.size(),10);
 		  	
 		  	for (i=0;i<requests.size();i++)
 		  	  {
 		  	  	ar = (ActionRequest)requests.elementAt(i);
 		  	  	
 		  	  	if (ar.startTime <= currentSystemTime)
 		  	  	  currentRequests.addElement((ActionRequest)ar);
 		  	  }
 		  	
 		  	 		  	
            // filter out those requests whose deadlines have been missed on all candidate devices
            startTime = System.currentTimeMillis();
            
            Vector removedRequests = new Vector (currentRequests.size(),10);
            for (i=0;i<currentRequests.size();i++)
              {
              	ar = (ActionRequest)currentRequests.elementAt(i);
              	
              	boolean outdated = true;
              	
              	for (j=0;j<ar.costVector.size();j++)
              	  {
              	  	double cost = Double.parseDouble((String)ar.costVector.elementAt(j));
              	  	
              	  	if (currentSystemTime + cost <= ar.startTime + ar.deadline)
              	  	  {outdated = false; break;}            	  	
              	  }
              	
              	if (outdated) removedRequests.addElement((ActionRequest)ar);
              }
            
            for (i=0;i<removedRequests.size();i++)
              {
                droppedRequestNum ++;
                
                ar = (ActionRequest)removedRequests.elementAt(i);
                
                currentRequests.removeElement(ar);
                
                requests.removeElement(ar);
                
                logFileWriter.println("#");
                logFileWriter.println("Round #" + (++count) + ": Request #" + ar.aid +  " is dropped.");                
                logFileWriter.println();
 		  	    logFileWriter.println("Request start time: " +  ar.startTime + ".");
 		  	    logFileWriter.println("Request deadline: " + ar.deadline + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println("Current system time: " + currentSystemTime + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Candidate device ids of the request: ");
 		  	    for (j=0;j<ar.didVector.size();j++)
 		  	      logFileWriter.print((String)ar.didVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Costs of the request on candidate devices: ");
 		  	    for (j=0;j<ar.costVector.size();j++)
 		  	      logFileWriter.print((String)ar.costVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
              }
              
            endTime = System.currentTimeMillis();
 		  	
 		  	schedulingTime += (endTime - startTime) / 1000.0;
 		  	
 		  	currentSystemTime += (endTime - startTime) / 1000.0;
 		  	
 		  	
 		  	// select one request to be scheduled
 		  	if (currentRequests.size() == 0)
 		  	  {
 		  	  	if (requests.size() == 0)
 		  	  	  break;
 		  	  	
 		  	  	minStartTime = ((ActionRequest)requests.elementAt(0)).startTime;
 		  	  	
 		  	  	if (currentSystemTime < minStartTime)
 		  	  	  {
 		  	  	  	currentSystemTime = minStartTime;
 		  	  	  	
 		  	  	  	for (i=0;i<schdevs.length;i++)
 		  	          if (schdevs[i].nextFreeTime <= currentSystemTime)
 		  	            schdevs[i].busy = false; 		  	  	  	  
 		  	  	  }
 		  	  	
 		  	  	continue; 		  	  	
 		  	  }
 		  	  
 		  	startTime = System.currentTimeMillis(); 		  	 		  	
 		  	
 		  	randomRequestSelection(currentRequests);
 		  	
 		  	endTime = System.currentTimeMillis();
 		  	
 		  	schedulingTime += (endTime - startTime) / 1000.0;
 		  	
 		  	currentSystemTime += (endTime - startTime) / 1000.0; 		  	
 		  	
 		  	if (sel_ar == null)  // there is no request to be scheduled now
 		  	  {
 		  	  	if (requests.size() == 0)
 		  	  	  break;
 		  	  	  
 		  	  	minNextFreeTime = Double.MAX_VALUE; 		  	
 		  	    for (i=0;i<schdevs.length;i++)
 		  	      if ((minNextFreeTime > schdevs[i].nextFreeTime)&&(schdevs[i].busy==true))
 		  	        minNextFreeTime = schdevs[i].nextFreeTime;
 		  	        
 		  	    minStartTime = Double.MAX_VALUE;
 		  	  	for (i=0;i<requests.size();i++)
 		  	  	  {
 		  	  	  	ar = (ActionRequest)requests.elementAt(i);
 		  	  	  	
 		  	  	  	if (ar.startTime > currentSystemTime)
 		  	  	  	  {minStartTime = ar.startTime; break;}
 		  	  	  }
 		  	    
 		  	    currentSystemTime = Math.min(minNextFreeTime, minStartTime);
 		  	      	  	
 		  	  	for (i=0;i<schdevs.length;i++)
 		  	      if (schdevs[i].nextFreeTime <= currentSystemTime)
 		  	        schdevs[i].busy = false;	   
 		  	      	
 		  	  	continue;
 		  	  } 	
 		  	  	
 		  	
 		  	// service the selected requests on the device
 		  	SchedulingDevice sd = schdevs [id = Integer.parseInt(sel_did)-1];
 		  	sd.busy = true;
 		  	sd.nextFreeTime = currentSystemTime + sel_actionCost;
 		  	
 		  	totalRequestCompletionTime += sel_actionCost; 		  		
 		  	
 		  	Random ran = new Random(Integer.parseInt(sel_ar.aid));
 		  	
 		  	logFileWriter.println("#");
 		  	if (failRan.nextDouble() > sd.failProb)
 		  	  {
 		  	  	logFileWriter.println("Round #" + (++count) + ": Request #" + sel_ar.aid +  " is serviced on Device " + sel_did + " from system time " + currentSystemTime + " to " + (currentSystemTime + sel_actionCost) + ".");
 		  	    logFileWriter.println("Next free time of the device: " + sd.nextFreeTime);
 		  	    logFileWriter.println();
 		  	    logFileWriter.println("Request start time: " +  sel_ar.startTime + ".");
 		  	    logFileWriter.println("Request deadline: " + sel_ar.deadline + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println("Current system time: " + currentSystemTime + ".");
 		  	    logFileWriter.println("Action cost: " + sel_actionCost + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Candidate device ids of the request: ");
 		  	    for (j=0;j<sel_ar.didVector.size();j++)
 		  	      logFileWriter.print((String)sel_ar.didVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Costs of the request on candidate devices: ");
 		  	    for (j=0;j<sel_ar.costVector.size();j++)
 		  	      logFileWriter.print((String)sel_ar.costVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println();
 		  	     		  	    
 		  	    if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
 		  	  	  {
 		  	  		schedulingTime += sd.camSim.polling() / 1000.0;
 		  	  	
 		  	  		currentSystemTime += sd.camSim.pollingTime / 1000.0;
 		  	  	
 		  	  		logFileWriter.println("Old current physical status of Device " + sd.id + ": pan = " + sd.camSim.pan + "  tilt = " + sd.camSim.tilt + "  zoom = " + sd.camSim.zoom + ".");
 		   		  	  	
 		  	  		sd.camSim.pan = Integer.parseInt((String)sel_status.fields.elementAt(0));
 		  	    
 		  	   		sd.camSim.tilt = Integer.parseInt((String)sel_status.fields.elementAt(1));
 		  	    
 		  	    	sd.camSim.zoom = Integer.parseInt((String)sel_status.fields.elementAt(2));
 		  	    
 		  	    	logFileWriter.println("New current physical status of Device " + sd.id + ": pan = " + sd.camSim.pan + "  tilt = " + sd.camSim.tilt + "  zoom = " + sd.camSim.zoom + ".");
 		  	  	  }
 		  		else
 		  	  	  {
 		  	  		schedulingTime += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;
 		  	  	
 		  	  		currentSystemTime += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;
 		  	  	
 		  	  		logFileWriter.println("Old current physical status of Device " + sd.id + ": loc = (" + sd.er1Sim.loc.x + "," + sd.er1Sim.loc.y + "," + sd.er1Sim.loc.z + ") " + "angle = " + sd.er1Sim.angle + ".");
            	 
 		  	   		loc.x = Double.parseDouble((String)sel_status.fields.elementAt(0));
    	   	    
    	   	   		loc.y = Double.parseDouble((String)sel_status.fields.elementAt(1));
    	   	    
    	   	    	loc.z = Double.parseDouble((String)sel_status.fields.elementAt(2));
    	   	    
    	   	    	angle = AortaSchedulingSimulator.computeDeltaAngle(sd.er1Sim.loc,loc);
    	   	    
    	   	    	sd.er1Sim.setLoc(loc.x,loc.y,loc.z);
    	   	    
    	   	    	sd.er1Sim.setAngle(angle);
    	   	    
    	   	    	logFileWriter.println("New current physical status of Device " + sd.id + ": loc = (" + sd.er1Sim.loc.x + "," + sd.er1Sim.loc.y + "," + sd.er1Sim.loc.z + ") " + "angle = " + sd.er1Sim.angle + "."); 
 		  	  	  }            		  	
 		  	  }
 		  	else
 		  	  {
 		  	  	droppedRequestNum ++;
 		  	  	
 		  	  	logFileWriter.println("Round #" + (++count) + ": Request #" + sel_ar.aid +  " is scheduled on Device " + sel_did + ". The action execution lasts from system time " + currentSystemTime + " to " + (currentSystemTime + sel_actionCost) + " but fails to due to temporary device malfunction.");
 		  	    logFileWriter.println("Next free time of the device: " + sd.nextFreeTime);
 		  	    logFileWriter.println();
 		  	    logFileWriter.println("Request start time: " +  sel_ar.startTime + ".");
 		  	    logFileWriter.println("Request deadline: " + sel_ar.deadline + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println("Current system time: " + currentSystemTime + ".");
 		  	    logFileWriter.println("Action cost: " + sel_actionCost + ".");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Candidate device ids of the request: ");
 		  	    for (j=0;j<sel_ar.didVector.size();j++)
 		  	      logFileWriter.print((String)sel_ar.didVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.print("Costs of the request on candidate devices: ");
 		  	    for (j=0;j<sel_ar.costVector.size();j++)
 		  	      logFileWriter.print((String)sel_ar.costVector.elementAt(j) + " ");
 		  	    logFileWriter.println();
 		  	    logFileWriter.println();
 		  	     		  	    
 		  	    if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
 		  	  	  {
 		  	  		schedulingTime += sd.camSim.polling() / 1000.0;
 		  	  	
 		  	  		currentSystemTime += sd.camSim.pollingTime / 1000.0;
 		  	  	
 		  	  		logFileWriter.println("Old current physical status of Device " + sd.id + ": pan = " + sd.camSim.pan + "  tilt = " + sd.camSim.tilt + "  zoom = " + sd.camSim.zoom + ".");
 		   		  	  	
 		  	  		int originPan = sd.camSim.pan;
 		  	    	
 		  	    	int originTilt = sd.camSim.tilt;
 		  	    	
 		  	    	int originZoom = sd.camSim.zoom;
 		  	    	
 		  	    	int targetPan = Integer.parseInt((String)sel_status.fields.elementAt(0));
 		  	    
 		  	   		int targetTilt = Integer.parseInt((String)sel_status.fields.elementAt(1));
 		  	    
 		  	    	int targetZoom = Integer.parseInt((String)sel_status.fields.elementAt(2));
 		  	      	    	
 		  	    	int maxPan = Math.max(originPan, targetPan), minPan = Math.min(originPan, targetPan);
 		  	    	
 		  	    	int maxTilt = Math.max(originTilt, targetTilt), minTilt = Math.min(originTilt, targetTilt);
 		  	    	
 		  	    	int maxZoom = Math.max(originZoom, targetZoom), minZoom = Math.min(originZoom, targetZoom);
 		  	    	 		  	    	
 		  	    	sd.camSim.pan = minPan + Math.abs(failRan.nextInt())%(maxPan - minPan + 1);
 		  	    	
 		  	    	sd.camSim.tilt = minTilt + Math.abs(failRan.nextInt())%(maxTilt - minTilt + 1);
 		  	    	
 		  	    	sd.camSim.zoom = minZoom + Math.abs(failRan.nextInt())%(maxZoom - minZoom + 1);
 		  	    	
 		  	    	if (sd.camSim.zoom > 1 && sd.camSim.zoom < 2501)
 		  	    	  sd.camSim.zoom = 1;
 		  	    	else
 		  	    	  if (sd.camSim.zoom > 2501 && sd.camSim.zoom < 5001)
 		  	    	    sd.camSim.zoom = 2501;
 		  	    	  else
 		  	    	    if (sd.camSim.zoom > 5001 && sd.camSim.zoom < 7501)
 		  	    	      sd.camSim.zoom = 5001;
 		  	    	    else
 		  	    	      if (sd.camSim.zoom > 7501 && sd.camSim.zoom < 9999)
 		  	    	        sd.camSim.zoom = 7501;
 		  	    	      else
 		  	    	        sd.camSim.zoom = 9999;
 		  	    	
 		  	    	logFileWriter.println("New current physical status of Device " + sd.id + ": pan = " + sd.camSim.pan + "  tilt = " + sd.camSim.tilt + "  zoom = " + sd.camSim.zoom + ".");
 		  	  	  }
 		  		else
 		  	  	  {
 		  	  		schedulingTime += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;
 		  	  	
 		  	  		currentSystemTime += ER1Simulator.STANDARD_POLLING_TIME / 1000.0;
 		  	  	
 		  	  		logFileWriter.println("Old current physical status of Device " + sd.id + ": loc = (" + sd.er1Sim.loc.x + "," + sd.er1Sim.loc.y + "," + sd.er1Sim.loc.z + ") " + "angle = " + sd.er1Sim.angle + ".");
            	 
 		  	   		double originLocX = sd.er1Sim.loc.x;
 		  	   		
 		  	   		double originLocY = sd.er1Sim.loc.y;

                    double originAngle = sd.er1Sim.angle;
 		  	   		
 		  	   		double targetLocX = Double.parseDouble((String)sel_status.fields.elementAt(0));
    	   	    
    	   	   		double targetLocY = Double.parseDouble((String)sel_status.fields.elementAt(1));
    	   	    
    	   	    	double targetLocZ = Double.parseDouble((String)sel_status.fields.elementAt(2));
    	   	    
    	   	    	double targetAngle = AortaSchedulingSimulator.computeDeltaAngle(sd.er1Sim.loc,loc);
    	   	    
    	   	    	double maxLocX = Math.max(originLocX, targetLocX), minLocX = Math.min(originLocX, targetLocX);
    	   	    	
    	   	    	double maxLocY = Math.max(originLocY, targetLocY), minLocY = Math.min(originLocY, targetLocY);
    	   	    	
    	   	    	double maxAngle = Math.max(originAngle, targetAngle), minAngle = Math.min(originAngle, targetAngle);
    	   	    	    	   	    	
    	   	    	loc.x = minLocX + Math.abs(failRan.nextInt())%(maxLocX - minLocX + 1);
    	   	    	
    	   	    	loc.y = minLocY + Math.abs(failRan.nextInt())%(maxLocY - minLocY + 1);
    	   	    	
    	   	    	loc.z = targetLocZ;
    	   	    	
    	   	    	angle = minAngle + Math.abs(failRan.nextInt())%(maxAngle - minAngle + 1);
    	   	    	
    	   	    	sd.er1Sim.setLoc(loc.x,loc.y,loc.z);
    	   	    
    	   	    	sd.er1Sim.setAngle(angle);
    	   	    
    	   	    	logFileWriter.println("New current physical status of Device " + sd.id + ": loc = (" + sd.er1Sim.loc.x + "," + sd.er1Sim.loc.y + "," + sd.er1Sim.loc.z + ") " + "angle = " + sd.er1Sim.angle + "."); 
 		  	  	  } 		 	  	  	
 		  	  }  		  	 		  	
 		  
 		  	
 		  	// update request cost
 		  	startTime = System.currentTimeMillis();
 		  	
 		  	currentRequests.removeElement((ActionRequest)sel_ar);		  	
 		  	
 		  	if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
 		  	  AortaSchedulingSimulator.updatePhotoCosts(currentRequests, sel_did, schdevs);
 		  	else
 		  	  AortaSchedulingSimulator.updateWarnCosts(currentRequests, sel_did, schdevs); 		  	  
 		  	  
 		  	requests.removeElement((ActionRequest)sel_ar);
 		  	
 		  	endTime = System.currentTimeMillis();
 		  	
 		  	schedulingTime += (endTime - startTime) / 1000.0;
 		  	
 		  	currentSystemTime += (endTime - startTime) / 1000.0;	  	
 		  	
 		  	
 		  	// go to schedule next request
 		  	minNextFreeTime = Double.MAX_VALUE;
 		  	
 		  	for (i=0;i<schdevs.length;i++)
 		  	  if (minNextFreeTime > schdevs[i].nextFreeTime)
 		  	    minNextFreeTime = schdevs[i].nextFreeTime;
 		  	
 		  	if (minNextFreeTime > currentSystemTime)
 		  	  currentSystemTime = minNextFreeTime;
 		  	  
 		  	for (i=0;i<schdevs.length;i++)
 		  	  if (schdevs[i].nextFreeTime <= currentSystemTime)
 		  	    schdevs[i].busy = false;
 		  	    
 		  	sel_ar = null; sel_did = null; sel_status = null;
 		  	sel_weight = Double.MAX_VALUE; sel_actionCost = Double.MAX_VALUE;
 		  	
 		  	logFileWriter.flush();		  		     	
 		  }	
 		      
 	      
 	    try
 	      {
 	      	if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.WARN)
 	          for (i=0;i<schdevs.length;i++)
 	            schdevs[i].er1Sim.serverSocket.close();
 	       }
    	catch (Exception e) {e.printStackTrace();}
 		
 		
 		logFileWriter.println("#");
        logFileWriter.println("Scheduling time: " + schedulingTime + ".");
 		logFileWriter.println("Number of dropped requests: " + droppedRequestNum + ".");
 		logFileWriter.println("Average request completion time: " + (totalRequestCompletionTime / (AortaSchedulingSimulator.requestNum - droppedRequestNum)) + ".");
 		logFileWriter.println("#");
 		logFileWriter.close();
 	  }         
   
	
	public void randomRequestSelection (Vector currentRequests)
	  {
	  	int i,j,k;
	  	ActionRequest ar;	  
	  			
	  	Vector dids = new Vector (schdevs.length, 10);
	  		  	
	  	Random ran = new Random(0);
	  	
	  	for (i=0;i<schdevs.length;i++)
	  	  if (schdevs[i].busy == false)
	  	    {
	  	      boolean freeDev = false;
	  	      
	  	      for (j=0;j<currentRequests.size();j++)
	  	        {
	  	          ar = (ActionRequest)currentRequests.elementAt(j);
	  	          
	  	          for (k=0;k<ar.didVector.size();k++)
	  	            if (Integer.parseInt((String)ar.didVector.elementAt(k)) == schdevs[i].id)
	  	              {freeDev = true; break;}
	  	          
	  	          if (freeDev) break;	
	  	        }
	  	      
	  	      if (freeDev)
	  	        dids.addElement((String)String.valueOf(i));	  	      
	  	    }
	  	    
	  	if (dids.size() == 0) return;
	  	    
	    int mark;
	    if (dids.size() == 1)
	      mark = 0;
	    else
	      mark = Math.abs(ran.nextInt()) % dids.size();
	      
	    SchedulingDevice sd = schdevs[Integer.parseInt((String)dids.elementAt(mark))];
	      
	    Vector currentRequestsSingleDevice = new Vector(currentRequests.size(), 10);
      	
      	for (i=0;i<currentRequests.size();i++)
      	  {
      	  	ar = (ActionRequest)currentRequests.elementAt(i);
      	  	
      	  	int did;
      	  	boolean currentDev = false;
      	  	for (j=0;j<ar.didVector.size();j++)
      	  	  {
      	  	  	did = Integer.parseInt((String)ar.didVector.elementAt(j));
      	  	  	
      	  	  	if (did == sd.id)
      	  	  	  {currentDev = true; break;}
      	  	  }
      	  	
      	  	if (currentDev)
      	  	  currentRequestsSingleDevice.addElement((ActionRequest)ar);      	  	
      	  }	 
      	  
      	if (currentRequestsSingleDevice.size() == 1)
      	  mark = 0;
      	else
      	  mark = Math.abs(ran.nextInt()) % currentRequestsSingleDevice.size();
      	  
      	sel_ar = (ActionRequest)currentRequestsSingleDevice.elementAt(mark);
      	sel_did = String.valueOf(sd.id);
      	for (i=0;i<sel_ar.didVector.size();i++)
      	  if (Integer.parseInt((String)sel_ar.didVector.elementAt(i)) == sd.id)
      	    {
      	      sel_actionCost = Double.parseDouble((String)sel_ar.costVector.elementAt(i));
      	      
      	      sel_weight = sel_actionCost;
      	      
      	      sel_status = (Tuple)sel_ar.statusVector.elementAt(i);
      	      
      	      break;      	    	
      	    }      	 	
	  }
 } 