/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * Thread for action scheduling in a shared action operator.
 * It schedules the continuous, dynamic stream of action requests 
 * from multiple queries onto all devices in Aorta.
 *
 * @see ActionOperator
 * @see AortaSchedulingSimulator
 */
public class SchedulingThread extends Thread {

    /** Shared action operator that this scheduling thread corresponds to. */    
    public ActionOperator actionOp;    
    
    /** Representation of the devices involved in the scheduling. */
    SchedulingDevice schdevs [];
    
    /** Input request queue of the action operator. */
    Vector requests = new Vector (100,50);    
   
    /** Scheduling log file. */
    PrintWriter logFileWriter;
    
    /** Module implements our heuristic scheduling algorithm. */
    DHASA dhasa;
    
    public SchedulingThread (ActionOperator actionOp)
      {
      	this.actionOp = actionOp;
      	
      	String deviceBaseStation;
      	
      	if (actionOp.action.name.equals("photo"))
      	  deviceBaseStation = Aorta.CAMERA_BASESTATION;
      	else
      	  if (actionOp.action.name.equals("warn"))
      	    deviceBaseStation = Aorta.ROBOT_BASESTATION;
      	  else
      	    {
      	      deviceBaseStation = null;
      	      System.err.println("Unsupported action to be scheduled. Must be \"photo\" or \"wanr\" in current Aorta prototype.");	
      	      System.exit(1);
      	    }       	  
      	 
      	 String str;         
         Vector dids = new Vector(10,10);
         int start, stop;
      	 try 
      	   {
      	 	  BufferedReader br = new BufferedReader(new FileReader(deviceBaseStation));          
              
              while ((str = br.readLine()) != null) 
                {
                   str = str.toLowerCase().trim();
                   
                   start = str.indexOf("<id>");
                   
                   if (start == -1) continue; 
                   
                   start += "<id>".length();              
                   
                   stop = str.indexOf("</id>",start);                
                   
                   dids.addElement((String)str.substring(start,stop).trim());
                }
                
              br.close();
            }
         catch (Exception e) {e.printStackTrace();}
         
         double failProbs [] = new double [dids.size()];
         
         int i;
         for (i=0;i<failProbs.length;i++)
           failProbs[i]=0;
        
         dhasa = new DHASA(requests,logFileWriter);
         AortaSchedulingSimulator.initialize(dhasa.schdevs,logFileWriter,failProbs);
           
         schdevs = new SchedulingDevice [dids.size()];      
         
         for (i=0;i<dids.size();i++)
           {
           	 schdevs[i] = new SchedulingDevice ();
           	 
           	 schdevs[i].id = Integer.parseInt((String)dids.elementAt(i)) + 1;
           	 
           	 if (actionOp.action.devInf.type.equals("camera"))
           	   schdevs[i].camSim = new CameraSimulator(String.valueOf(schdevs[i].id));
           	 else
           	   schdevs[i].er1Sim = new ER1Simulator(String.valueOf(schdevs[i].id));
           }       
       
        dhasa.schdevs = schdevs;            		      
      }  
          
    
    /**
     * Process of action scheduling.
     */
    public void run ()
      {
      	int i;
      	int id;
      	int count = 0;
      	Point loc = new Point();
      	double angle;
      	
      	String logFile = "../logs/scheduling_logs/" + actionOp.action.name + ".log";
	   	
      	try {logFileWriter = new PrintWriter ( new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
      	
      	while (true)
      	  {
      	  	if (requests.size() == 0)
      	  	  {
      	  	  	try {Thread.currentThread().sleep(1000);}
      	  	  	catch (Exception e) {e.printStackTrace();}
      	  	  	
      	  	  	continue;      	  	  	
      	  	  }
      	  	
      	  	logFileWriter.println("#");
      	  	logFileWriter.println("A round of scheduling begins.");
      	  	      	  	
      	  	pollCandidateDevices();
      	  	
      	  	updateActionCost();     	  	
      	  	
      	  	dhasa.heuristicRequestSelection(requests);
      	  	
      	  	if (dhasa.sel_ar == null)
 		  	  {
 		  	  	double minNextFreeTime = Double.MAX_VALUE;
 		  	  	
 		  	  	for (i=0;i<schdevs.length;i++)
 		  	  	  if (minNextFreeTime > schdevs[i].nextFreeTime)
 		  	  	  	  minNextFreeTime = schdevs[i].nextFreeTime; 		  	  	  
 		  	  	
 		  	  	if (System.currentTimeMillis()/1000 < minNextFreeTime)
 		  	  	  try {Thread.currentThread().sleep((long)(minNextFreeTime - System.currentTimeMillis()/1000));}
 		  	  	  catch (Exception e) {e.printStackTrace();}
 		  	  	
 		  	  	continue;
 		  	  }
 		  	  
 		  	// service the selected requests on the device 		  	
 		  	logFileWriter.println("Round #" + (++count) + ": Request #" + dhasa.sel_ar.aid +  " is serviced on Device " + dhasa.sel_did + " from system time " + (System.currentTimeMillis()/1000) + " to " + (System.currentTimeMillis() / 1000 + dhasa.sel_actionCost) + ".");
 		  	logFileWriter.println();
 		  	logFileWriter.println("Request start time: " +  dhasa.sel_ar.startTime + ".");
 		  	logFileWriter.println("Request deadline: " + dhasa.sel_ar.deadline + ".");
 		  	logFileWriter.println();
 		  	logFileWriter.println("Current system time: " + System.currentTimeMillis()/1000 + ".");
 		  	logFileWriter.println("Action cost: " + dhasa.sel_actionCost + ".");
 		  	logFileWriter.println();
 		  	
 		  	SchedulingDevice sd = schdevs [id = Integer.parseInt(dhasa.sel_did)-1];
 		  	sd.busy = true;
 		  	sd.nextFreeTime = System.currentTimeMillis()/1000 + dhasa.sel_actionCost;
 		  			  	
 		  	  		  	 		  	
 		  	if (AortaSchedulingSimulator.actionName == AortaSchedulingSimulator.PHOTO)
 		  	  {
 		  	  	  	  	
 		  	  	logFileWriter.println("Old current physical status of Device " + sd.id + ": pan = " + sd.camSim.pan + "  tilt = " + sd.camSim.tilt + "  zoom = " + sd.camSim.zoom + ".");
 		   		  	  	
 		  	  	sd.camSim.pan = Integer.parseInt((String)dhasa.sel_status.fields.elementAt(0));
 		  	    
 		  	    sd.camSim.tilt = Integer.parseInt((String)dhasa.sel_status.fields.elementAt(1));
 		  	    
 		  	    sd.camSim.zoom = Integer.parseInt((String)dhasa.sel_status.fields.elementAt(2));
 		  	    
 		  	    logFileWriter.println("New current physical status of Device " + sd.id + ": pan = " + sd.camSim.pan + "  tilt = " + sd.camSim.tilt + "  zoom = " + sd.camSim.zoom + ".");
 		  	  }
 		  	else
 		  	  {
 		  	    logFileWriter.println("Old current physical status of Device " + sd.id + ": loc = (" + sd.er1Sim.loc.x + "," + sd.er1Sim.loc.y + "," + sd.er1Sim.loc.z + ") " + "angle = " + sd.er1Sim.angle + ".");
            	
 		  	    loc.x = Double.parseDouble((String)dhasa.sel_status.fields.elementAt(0));
    	   	    
    	   	    loc.y = Double.parseDouble((String)dhasa.sel_status.fields.elementAt(1));
    	   	    
    	   	    loc.z = Double.parseDouble((String)dhasa.sel_status.fields.elementAt(2));
    	   	    
    	   	    angle = AortaSchedulingSimulator.computeDeltaAngle(sd.er1Sim.loc,loc);
    	   	    
    	   	    sd.er1Sim.setLoc(loc.x,loc.y,loc.z);
    	   	    
    	   	    sd.er1Sim.setAngle(angle);
    	   	    
    	   	    logFileWriter.println("New current physical status of Device " + sd.id + ": loc = (" + sd.er1Sim.loc.x + "," + sd.er1Sim.loc.y + "," + sd.er1Sim.loc.z + ") " + "angle = " + sd.er1Sim.angle + "."); 
 		  	  } 		  	
            		  	
 		  	
 		  	requests.removeElement((ActionRequest)dhasa.sel_ar); 		  	 		  	
 		  	
 		  	// go to schedule next request
 		  	double minNextFreeTime = Double.MAX_VALUE;
 		  	
 		  	for (i=0;i<schdevs.length;i++)
 		  	  if (minNextFreeTime > schdevs[i].nextFreeTime)
 		  	    minNextFreeTime = schdevs[i].nextFreeTime; 		  	
 		  
 		  	if (System.currentTimeMillis()/1000 < minNextFreeTime)
 		  	  try {Thread.currentThread().sleep((long)(minNextFreeTime - System.currentTimeMillis()/1000));}
 		  	   catch (Exception e) {e.printStackTrace();}
 		  	  
 		  	for (i=0;i<schdevs.length;i++)
 		  	  if (schdevs[i].nextFreeTime <= System.currentTimeMillis()/1000)
 		  	    schdevs[i].busy = false;
 		  	    
 		  	dhasa.sel_ar = null; dhasa.sel_did = null; dhasa.sel_status = null;
 		  	dhasa.sel_weight = Double.MAX_VALUE; dhasa.sel_actionCost = Double.MAX_VALUE;   	  	
      	  	
      	  	logFileWriter.println("A round of scheduling ends.");
      	  	logFileWriter.flush(); 
      	  }	
      }	
      
      
   /** Poll the candidate devices and update their current physical status. */
   public void pollCandidateDevices()
     {
     	if (actionOp.action.name.equals("photo"))
      	  pollCameras();
      	else
      	  pollRobots(); 	
     }
     
   
   public void pollCameras()
     {
        logFileWriter.println("Polling the candidate cameras .... ");
	   	
	   	CameraPollingThread cpt [] = new CameraPollingThread[schdevs.length];
	   	
	   	long startTime = System.currentTimeMillis();
	   	int i;
	   	for (i=0;i<schdevs.length;i++)
	   	  {	   	    
	   	    cpt[i] = new CameraPollingThread(schdevs[i].camSim, schdevs[i].t, actionOp.q.aortaqp.CAMERA_SIMULATION);
	   	    
	   	    cpt[i].start();
	      }	
	         	    
	   	long endTime = System.currentTimeMillis();
	   		   	        
	   	while (endTime - startTime < CameraSimulator.POLLING_TIMEOUT_LIMIT)
	   	  {
	   	    boolean finished = true;
	   	          	
	   	    for (i=0;i<cpt.length;i++)
	   	      if (cpt[i] != null && cpt[i].connected == false)
	   	        finished = false;
	   	          	    	   	          	
	   	    if (finished) break;
	   	       
	   	    endTime = System.currentTimeMillis();	   	          	
	   	  } 	   	   	
      	
      	// remove those candidate devices the polling to which timed out
      	for (i=0;i<cpt.length;i++)
	   	  if (cpt[i] == null || cpt[i].connected == false)
	   	      {schdevs[i].busy = true;}		   	
	   	  
	   	logFileWriter.println("Finish polling the candidate cameras.");
	   	logFileWriter.println();
	   	logFileWriter.flush();	
     }
     
     
   public void pollRobots()
     {
       	logFileWriter.println("Polling the candidate robots .... ");
	   	
	   	RobotPollingThread rpt [] = new RobotPollingThread[schdevs.length];
	   	
	   	long startTime = System.currentTimeMillis();
	   	int i;
	   	for (i=0;i<schdevs.length;i++)
	   	  {
	   	    rpt[i] = new RobotPollingThread(schdevs[i].er1Sim, (String)schdevs[i].t.fields.elementAt(2), actionOp.q.aortaqp);
	   	    
	   	    rpt[i].start();
	      }	
	         	    
	   	long endTime = System.currentTimeMillis();
	   		   	        
	   	while (endTime - startTime < ER1Simulator.POLLING_TIMEOUT_LIMIT)
	   	  {
	   	    boolean finished = true;
	   	          	
	   	    for (i=0;i<schdevs.length;i++)
	   	      if (rpt[i] != null && rpt[i].connected == false)
	   	        finished = false;
	   	          	    	   	          	
	   	    if (finished) break;
	   	       
	   	    endTime = System.currentTimeMillis();	   	          	
	   	  } 	   	   	
      	
      	// remove those candidate devices the polling to which timed out
      	for (i=0;i<schdevs.length;i++)
	   	  if (rpt[i] == null || rpt[i].connected == false)
	   	      schdevs[i].busy = true;    	
	   
	   	logFileWriter.println("Finish polling the candidate robots.");
	   	logFileWriter.println();
	   	logFileWriter.flush();  	     	
     }
     
     
   /** Re-estimate the cost of each action request on each candidate device. */      	
   public void updateActionCost()
     {
       if (actionOp.action.name.equals("photo"))
      	  updatePhotoCost();
      	else
      	  updateWarnCost(); 
     }
     
     
   public void updatePhotoCost()
     {
       	int i,j;
       	int id, pan, tilt, zoom;
       	ActionRequest ar;
       	Tuple t;
       	
       	for (i=0;i<requests.size();i++)
      	  {
      	  	ar = (ActionRequest)requests.elementAt(i);
      	  	
      	  	for (j=0;j<ar.didVector.size();j++)
      	  	  {
      	  	  	t = (Tuple)ar.statusVector.elementAt(j);      	  	  	      	  	  	  	  	  	
	   	        actionOp.camera_id_c = (String)t.fields.elementAt(0);  	        
	   	        actionOp.loc_c = Point.parsePoint((String)t.fields.elementAt(4));
	   	        
	   	        ar.costVector.addElement(String.valueOf(actionOp.estimateActionCost()));
	   	        
	   	        id = Integer.parseInt(actionOp.camera_id_c) - 1;
	   	        pan = schdevs[id].camSim.targetPan(actionOp.loc_c);
	   	        tilt = schdevs[id].camSim.targetTilt(actionOp.loc_c);
	   	        zoom = schdevs[id].camSim.targetZoom(actionOp.loc_c);
	   	        
	   	        Vector vt = new Vector (10,5);
	   	        vt.addElement(String.valueOf(pan));
	   	        vt.addElement(String.valueOf(tilt));
	   	        vt.addElement(String.valueOf(zoom));
	   	        t = new Tuple(vt);	   	        
	   	        ar.statusVector.removeElementAt(j);
	   	        ar.statusVector.insertElementAt((Tuple)t,j);
	   	      }	   	      	     	  	
      	  }      		
     }
     
     
   public void updateWarnCost()
     {
        int i,j;
        ActionRequest ar;
        Tuple t;
            
      	for (i=0;i<requests.size();i++)
      	  {
      	  	ar = (ActionRequest)requests.elementAt(i);
      	  	
      	  	for (j=0;j<ar.didVector.size();j++)
      	  	  {
      	  	  	t = (Tuple)ar.statusVector.elementAt(j);      	  	  	      	  	  	  	  	  	
	   	        actionOp.robot_id_c = (String)t.fields.elementAt(0);  	        
	   	        actionOp.loc_c = Point.parsePoint((String)t.fields.elementAt(1));
	   	        
	   	        ar.costVector.addElement(String.valueOf(actionOp.estimateActionCost()));
	   	        
	   	        Vector vt = new Vector (10,5);
	   	        vt.addElement(String.valueOf(actionOp.loc_c.x));
	   	        vt.addElement(String.valueOf(actionOp.loc_c.y));
	   	        vt.addElement(String.valueOf(actionOp.loc_c.z));
	   	        t = new Tuple(vt);	   	        
	   	        ar.statusVector.removeElementAt(j);
	   	        ar.statusVector.insertElementAt((Tuple)t,j);
	   	      } 	       	     	  	
      	  }	     	
     }    
}

