/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.util.*;

/**
 * Implementation of the system built-in timer functions.
 * This class assumes that the input parameters of all functions are valid.
 * Such validity is ensured by the Aorta query parser. 
 *
 * @see Aorta#parseAQ
 * @see AQ
 * @see WhereTimerQC
 * @see StartClause
 * @see StopClause
 */
public class TimerFunctions {
	
	static Calendar currentCalendar = Calendar.getInstance();
	
	/* Timer units currently supported */
    public static final String YEAR = "year";
    public static final String MONTH = "month";
    public static final String DAY = "day";
    public static final String WEEK = "week";
    public static final String HOUR = "hour";
    public static final String MINUTE = "minute";
    public static final String SECOND = "second";
    public static final String PLURAL = "s";
    
    static String [] unitArray = {YEAR, MONTH, DAY, WEEK, HOUR, MINUTE, SECOND, null};
    
	/* System built-in timer variables */
    public static final String $ANY = "$any";
	public static final int $ANY_VAL = -1;
	public static final String $ONCE = "$once";
	public static final String $NOW = "$now";

    /* System built-in timer function names */
    public static final String EVERY = "every";
    public static final String ATTIME = "attime";
    public static final String ONDAY = "onday";
    public static final String INTIMEINTERVAL = "intimeinterval";
    public static final String INDAYINTERVAL = "indayinterval";
    public static final String ONDAYOFWEEK = "ondayofweek";
    public static final String INDAYOFWEEKINTERVAL = "indayofweekinterval";
    	
	public static final long ONE_DAY_SECONDS = 24*60*60;
	
	
	public static boolean atTime(int hour, int minute, int second) 
	  {
	    if ( (hour == currentCalendar.get(Calendar.HOUR_OF_DAY)) || (hour == $ANY_VAL ) )
          if ( (minute == currentCalendar.get(Calendar.MINUTE)) || (minute == $ANY_VAL) ) 
            if ( (second == currentCalendar.get(Calendar.SECOND)) || (second == $ANY_VAL) )
              return true;
              
        return false;         
	  }  
	  
	
	public static boolean onDay(int year, int month, int day)
	  {
	  	 if ( (year == currentCalendar.get(Calendar.YEAR)) || (year == $ANY_VAL) )
	  	   if ( (month == (currentCalendar.get(Calendar.MONTH) + 1)) || (month == $ANY_VAL) )
	  	     if ( (day == currentCalendar.get(Calendar.DAY_OF_MONTH))  || (day == $ANY_VAL) )
	  	       return true;
	  	       
	  	 return false;	  	 
	  } 
	  
	  
	/**
	 * The start and end points of the specified interval must be of the same day.
	 */
	public static boolean inTimeInterval(int hour1,int minute1,int second1, int hour2, int minute2, int second2)
	  {
         int hour,minute, second;
         
         hour = currentCalendar.get(Calendar.HOUR_OF_DAY);
         minute = currentCalendar.get(Calendar.MINUTE);
         second = currentCalendar.get(Calendar.SECOND);
         
         if ( (hour < hour1) || (hour == hour1)&&(minute < minute1) || (hour == hour1)&&(minute == minute1)&&(second < second1) )
           return false;
         if ( (hour > hour2) || (hour == hour2)&&(minute > minute2) || (hour == hour2)&&(minute == minute2)&&(second > second2) )
           return false;
         
         return true;         	  	
	  } 
	  
	
	public static boolean inDayInterval(int year1, int month1, int day1, int year2, int month2, int day2)
	  {
	  	int year, month, day;
	  	
	  	year = currentCalendar.get(Calendar.YEAR);
	  	month = currentCalendar.get(Calendar.MONTH) + 1;
	  	day = currentCalendar.get(Calendar.DAY_OF_MONTH);
	  	
	  	if ( (year < year1) || (year == year1)&&(month < month1) || (year == year1)&&(month == month1)&&(day < day1) )
	  	  return false;
	  	if ( (year > year2) || (year == year2)&&(month > month2) || (year == year2)&&(month == month2)&&(day > day2) )
	  	  return false;
	  	  
	  	
	  	return true;	  	
	  } 
	  
	  
	 public static boolean onDayOfWeek(String date)
	   {
	   	 int day_of_week = testDayOfWeek (date);
	   	  
	   	 if (day_of_week == currentCalendar.get(Calendar.DAY_OF_WEEK))
	   	   return true;
	   	 else 
	   	   return false;	   	  
	   }
	   
	 
	 public static boolean inDayOfWeekInterval(String date1, String date2)
	   {
	   	 int day_of_week, day_of_week1, day_of_week2;
	   	
	   	 day_of_week1 = testDayOfWeek(date1);
	   	
	   	 day_of_week2 = testDayOfWeek(date2);
	   	
	   	 day_of_week = currentCalendar.get(Calendar.DAY_OF_WEEK);
	   	
	   	 if (day_of_week2 >= day_of_week1)
	   	   {
	   	  	 if ( (day_of_week >= day_of_week1) && (day_of_week <= day_of_week2) )
	   	       return true;
	   	     else 
	   	       return false;
	   	   }
	     else
	   	   {
	   	  	 if ( (day_of_week >= day_of_week1) || (day_of_week <= day_of_week2) )
	   	       return true;
	   	     else 
	   	       return false;
	   	   }	   	
	   } 
	   
	   
	   public static int testDayOfWeek (String date)
	     {
	        int day_of_week = 0;
	        
	     	if (date.equals("sunday") || date.equals("sun.") || date.equals("sun") )
	   	      day_of_week = 1;
	   	    else if (date.equals("monday") || date.equals("mon.") || date.equals("mon") )
	   	         day_of_week = 2;
	   	       else if (date.equals("tuesday") || date.equals("tue.") || date.equals("tue") || date.equals("tues.") || date.equals("tues") )  
	   	              day_of_week = 3;
	   	            else if (date.equals("wednesday") || date.equals("wed.") || date.equals("wed") || date.equals("weds.")  || date.equals("webs") )  
	   	                   day_of_week = 4;
	   	                 else if (date.equals("thursday") || date.equals("thu") || date.equals("thur") )
	   	                        day_of_week = 5;
	   	                      else if (date.equals("friday") || date.equals("fri.") || date.equals("fri") )
	   	                             day_of_week = 6;
	   	                           else if (date.equals("saturday") || date.equals("sat.") || date.equals("sat") )
	   	                                  day_of_week = 7;
	   	                             
	   	    return day_of_week;	   	    
	     } 
	  
	  
	  public static long computeTimeInterval(int hour,int minute, int second)
	    {
	       int hour_n, minute_n, second_n;
	       long interval;
	       
	       hour_n = currentCalendar.get(Calendar.HOUR_OF_DAY);
	       minute_n = currentCalendar.get(Calendar.MINUTE);
	       second_n = currentCalendar.get(Calendar.SECOND);
	       
	       if ( (hour  > hour_n) || (hour == hour_n)&&(minute > minute_n) || (hour == hour_n)&&(minute == minute_n)&&(second > second_n) )
	         {
	            interval = ( (hour - hour_n)*60 + (minute - minute_n) )*60 + (second - second_n - 1);
	            interval *= 1000;
	         }
	       else 
	         {
	         	interval = ( (hour_n - hour)*60 + (minute_n - minute) )*60 + (second_n - second);
	            interval = ONE_DAY_SECONDS - interval;
	            interval *= 1000;
	         }
	            
	        return interval;	        	
	    }
	    
	    
	    public static long computeDayTimeInterval(int year,int month, int day)
	      {
	      	int year_n,month_n,day_n;
	      	long interval;
	      	
	      	year_n = currentCalendar.get(Calendar.YEAR);
	      	month_n = currentCalendar.get(Calendar.MONTH) + 1;
	      	day_n = currentCalendar.get(Calendar.DAY_OF_MONTH);
	       
	        if ( (year > year_n) || (year == year_n)&&(month > month_n) || (year == year_n)&&(month == month_n)&&(day > day_n) )
	         {}
	        else 
	         return -1;
	         
	        Calendar c = Calendar.getInstance();
	        c.set(year, month - 1, day,0,0,0);
	        interval = c.getTimeInMillis() - currentCalendar.getTimeInMillis();
	                    
	        return interval;	       
	      } 
	      
	      
	    public static long computeDayOfWeekTimeInterval(String date)
	      {
	      	int day_of_week = testDayOfWeek(date);
	      	int day_of_week_n = currentCalendar.get(Calendar.DAY_OF_WEEK);
	      	
	      	int hour = currentCalendar.get(Calendar.HOUR_OF_DAY);
	      	int minute = currentCalendar.get(Calendar.MINUTE);
	      	int second = currentCalendar.get(Calendar.SECOND);
	      	
	      	long interval;
	      	interval = (hour*60 + minute)*60 + second;
	      	interval = ONE_DAY_SECONDS - interval;
	      	
	      	if (day_of_week > day_of_week_n)
	      	  interval += (day_of_week - 1 - day_of_week_n)*ONE_DAY_SECONDS;
	        else 
	      	  interval += (day_of_week - day_of_week_n - 1 + 7)*ONE_DAY_SECONDS;
	      	   
	      	 
	      	return interval*1000;	      	
	      }             
	 
} 